@@ -1,4 +1,5 @@
-/* Copyright (C) 2000-2003 MySQL AB
+/*
+   Copyright (c) 2000, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 /*
   This file defines the NDB Cluster handler: the interface between MySQL and
@@ -43,16 +45,15 @@ class NdbIndexScanOperation;
 class NdbBlob;
 class NdbIndexStat;
 class NdbEventOperation;
+class NdbInterpretedCode;
 class ha_ndbcluster_cond;
 class Ndb_event_data;
 
 // connectstring to cluster if given by mysqld
 extern const char *ndbcluster_connectstring;
 extern ulong ndb_cache_check_time;
-#ifdef HAVE_NDB_BINLOG
 extern ulong ndb_report_thresh_binlog_epoch_slip;
 extern ulong ndb_report_thresh_binlog_mem_usage;
-#endif
 
 typedef enum ndb_index_type {
   UNDEFINED_INDEX = 0,
@@ -89,7 +90,6 @@ typedef struct ndb_index_data {
     from a key and from a row.
   */
   NdbRecord *ndb_record_key;
-  NdbRecord *ndb_record_row;
   NdbRecord *ndb_unique_record_key;
   NdbRecord *ndb_unique_record_row;
 } NDB_INDEX_DATA;
@@ -107,13 +107,17 @@ public:
 		 const NdbDictionary::Table *table) :
     dictionary(dict),
     old_table(table),
-    new_table(new NdbDictionary::Table(*table))
+    new_table(new NdbDictionary::Table(*table)),
+      table_id(table->getObjectId()),
+      old_table_version(table->getObjectVersion())
   {}
   ~NDB_ALTER_DATA()
   { delete new_table; }
   NdbDictionary::Dictionary *dictionary;
   const  NdbDictionary::Table *old_table;
   NdbDictionary::Table *new_table;
+  Uint32 table_id;
+  Uint32 old_table_version;
 };
 
 typedef union { const NdbRecAttr *rec; NdbBlob *blob; void *ptr; } NdbValue;
@@ -128,6 +132,36 @@ typedef enum {
   NSS_ALTERED 
 } NDB_SHARE_STATE;
 
+#ifdef HAVE_NDB_BINLOG
+enum enum_conflict_fn_type
+{
+  CFT_NDB_UNDEF = 0
+  ,CFT_NDB_MAX
+  ,CFT_NDB_OLD
+  ,CFT_NDB_MAX_DEL_WIN
+};
+
+/* NdbOperation custom data which points out handler and record. */
+struct Ndb_exceptions_data {
+  struct st_ndbcluster_share *share;
+  const uchar* row;
+};
+
+typedef struct st_ndbcluster_conflict_fn_share {
+  enum_conflict_fn_type m_resolve_cft;
+
+  /* info about original table */
+  uint8 m_pk_cols;
+  uint8 m_resolve_column;
+  uint8 m_resolve_size;
+  uint8 unused;
+  uint16 m_offset[16];
+
+  const NdbDictionary::Table *m_ex_tab;
+  uint32 m_count;
+} NDB_CONFLICT_FN_SHARE;
+#endif
+
 /*
   Stats that can be retrieved from ndb
 */
@@ -153,15 +187,17 @@ typedef struct st_ndbcluster_share {
   char *table_name;
   Ndb::TupleIdRange tuple_id_range;
   struct Ndb_statistics stat;
-#ifdef HAVE_NDB_BINLOG
+  bool util_thread; // if opened by util thread
   uint32 connect_count;
   uint32 flags;
+#ifdef HAVE_NDB_BINLOG
+  NDB_CONFLICT_FN_SHARE *m_cfn_share;
+#endif
   Ndb_event_data *event_data; // Place holder before NdbEventOperation is created
   NdbEventOperation *op;
   char *old_names; // for rename table
   MY_BITMAP *subscriber_bitmap;
   NdbEventOperation *new_op;
-#endif
 } NDB_SHARE;
 
 inline
@@ -197,7 +233,6 @@ struct Ndb_tuple_id_range_guard {
   Ndb::TupleIdRange& range;
 };
 
-#ifdef HAVE_NDB_BINLOG
 /* NDB_SHARE.flags */
 #define NSF_HIDDEN_PK   1u /* table has hidden primary key */
 #define NSF_BLOB_FLAG   2u /* table has blob attributes */
@@ -206,24 +241,41 @@ struct Ndb_tuple_id_range_guard {
 #define NSF_BINLOG_USE_UPDATE 16u  /* table update should be binlogged using
                                      update log event */
 inline void set_binlog_logging(NDB_SHARE *share)
-{ share->flags&= ~NSF_NO_BINLOG; }
+{
+  DBUG_PRINT("info", ("set_binlog_logging"));
+  share->flags&= ~NSF_NO_BINLOG;
+}
 inline void set_binlog_nologging(NDB_SHARE *share)
-{ share->flags|= NSF_NO_BINLOG; }
+{
+  DBUG_PRINT("info", ("set_binlog_nologging"));
+  share->flags|= NSF_NO_BINLOG;
+}
 inline my_bool get_binlog_nologging(NDB_SHARE *share)
 { return (share->flags & NSF_NO_BINLOG) != 0; }
 inline void set_binlog_updated_only(NDB_SHARE *share)
-{ share->flags&= ~NSF_BINLOG_FULL; }
+{
+  DBUG_PRINT("info", ("set_binlog_updated_only"));
+  share->flags&= ~NSF_BINLOG_FULL;
+}
 inline void set_binlog_full(NDB_SHARE *share)
-{ share->flags|= NSF_BINLOG_FULL; }
+{
+  DBUG_PRINT("info", ("set_binlog_full"));
+  share->flags|= NSF_BINLOG_FULL;
+}
 inline my_bool get_binlog_full(NDB_SHARE *share)
 { return (share->flags & NSF_BINLOG_FULL) != 0; }
 inline void set_binlog_use_write(NDB_SHARE *share)
-{ share->flags&= ~NSF_BINLOG_USE_UPDATE; }
+{
+  DBUG_PRINT("info", ("set_binlog_use_write"));
+  share->flags&= ~NSF_BINLOG_USE_UPDATE;
+}
 inline void set_binlog_use_update(NDB_SHARE *share)
-{ share->flags|= NSF_BINLOG_USE_UPDATE; }
+{
+  DBUG_PRINT("info", ("set_binlog_use_update"));
+  share->flags|= NSF_BINLOG_USE_UPDATE;
+}
 inline my_bool get_binlog_use_update(NDB_SHARE *share)
 { return (share->flags & NSF_BINLOG_USE_UPDATE) != 0; }
-#endif
 
 typedef enum ndb_query_state_bits {
   NDB_QUERY_NORMAL = 0,
@@ -236,7 +288,19 @@ typedef enum ndb_query_state_bits {
 
 enum THD_NDB_OPTIONS
 {
-  TNO_NO_LOG_SCHEMA_OP= 1 << 0
+  TNO_NO_LOG_SCHEMA_OP=  1 << 0,
+  /*
+    In participating mysqld, do not try to acquire global schema
+    lock, as one other mysqld already has the lock.
+  */
+  TNO_NO_LOCK_SCHEMA_OP= 1 << 1
+  /*
+    Skip drop of ndb table in delete_table.  Used when calling
+    mysql_rm_table_part2 in "show tables", as we do not want to
+    remove ndb tables "by mistake".  The table should not exist
+    in ndb in the first place.
+  */
+  ,TNO_NO_NDB_DROP_TABLE=    1 << 2
 };
 
 enum THD_NDB_TRANS_OPTIONS
@@ -262,12 +326,17 @@ class Thd_ndb
 
   Ndb_cluster_connection *connection;
   Ndb *ndb;
+  /* this */
+  class ha_ndbcluster *m_handler;
   ulong count;
   uint lock_count;
   uint start_stmt_count;
+  uint save_point_count;
   NdbTransaction *trans;
   bool m_error;
   bool m_slow_path;
+  bool m_force_send;
+
   int m_error_code;
   uint32 m_query_id; /* query id whn m_error_code was set */
   uint32 options;
@@ -285,8 +354,26 @@ class Thd_ndb
     we execute() to flush the rows buffered in m_batch_mem_root.
   */
   uint m_unsent_bytes;
+  uint m_batch_size;
+
+  uint m_execute_count;
+  uint m_max_violation_count;
+  uint m_old_violation_count;
+  uint m_conflict_fn_usage_count;
+
+  uint m_scan_count;
+  uint m_pruned_scan_count;
+
+  NdbTransaction *global_schema_lock_trans;
+  uint global_schema_lock_count;
+  uint global_schema_lock_error;
+
+  unsigned m_connect_count;
+  bool valid_ndb(void);
+  bool recycle_ndb(THD* thd);
 };
 
+int ndbcluster_commit(handlerton *hton, THD *thd, bool all);
 class ha_ndbcluster: public handler
 {
  public:
@@ -294,9 +381,10 @@ class ha_ndbcluster: public handler
   ~ha_ndbcluster();
 
   int ha_initialise();
-  void column_bitmaps_signal();
+  void column_bitmaps_signal(uint sig_type);
   int open(const char *name, int mode, uint test_if_locked);
   int close(void);
+  void local_close(THD *thd, bool release_metadata);
 
   int optimize(THD* thd, HA_CHECK_OPT* check_opt);
   int analyze(THD* thd, HA_CHECK_OPT* check_opt);
@@ -318,6 +406,7 @@ class ha_ndbcluster: public handler
   int rnd_next(uchar *buf);
   int rnd_pos(uchar *buf, uchar *pos);
   void position(const uchar *record);
+  virtual int cmp_ref(const uchar * ref1, const uchar * ref2);
   int read_range_first(const key_range *start_key,
                        const key_range *end_key,
                        bool eq_range, bool sorted);
@@ -346,6 +435,8 @@ class ha_ndbcluster: public handler
   int info(uint);
   void get_dynamic_partition_info(PARTITION_INFO *stat_info, uint part_id);
   uint32 calculate_key_hash_value(Field **field_array);
+  bool read_before_write_removal_possible(List<Item> *fields,
+                                          List<Item> *values);
   int extra(enum ha_extra_function operation);
   int extra_opt(enum ha_extra_function operation, ulong cache_size);
   int reset();
@@ -363,6 +454,7 @@ class ha_ndbcluster: public handler
   int final_drop_index(TABLE *table_arg);
   void set_part_info(partition_info *part_info, bool early);
   ulong index_flags(uint idx, uint part, bool all_parts) const;
+  virtual const key_map *keys_to_use_for_scanning() { return &btree_keys; }
   uint max_supported_record_length() const;
   uint max_supported_keys() const;
   uint max_supported_key_parts() const;
@@ -396,12 +488,26 @@ class ha_ndbcluster: public handler
   void start_bulk_insert(ha_rows rows);
   int end_bulk_insert();
 
+  bool start_bulk_update();
+  int bulk_update_row(const uchar *old_data, uchar *new_data,
+                      uint *dup_key_found);
+  int exec_bulk_update(uint *dup_key_found);
+  void end_bulk_update();
+  int ndb_update_row(const uchar *old_data, uchar *new_data,
+                     int is_bulk_update);
+
   static Thd_ndb* seize_thd_ndb();
   static void release_thd_ndb(Thd_ndb* thd_ndb);
  
 static void set_dbname(const char *pathname, char *dbname);
 static void set_tabname(const char *pathname, char *tabname);
 
+  /*
+    static member function as it needs to access private
+    NdbTransaction methods
+  */
+  static void release_completed_operations(NdbTransaction*);
+
   /*
     Condition pushdown
   */
@@ -446,7 +552,7 @@ static void set_tabname(const char *pathname, char *tabname);
   /*
    * Internal to ha_ndbcluster, used by C functions
    */
-  int ndb_err(NdbTransaction*);
+  int ndb_err(NdbTransaction*, bool have_lock= FALSE);
 
   my_bool register_query_cache_table(THD *thd, char *table_key,
                                      uint key_length,
@@ -470,18 +576,43 @@ static void set_tabname(const char *pathname, char *tabname);
                          HA_ALTER_INFO *alter_info,
                          HA_ALTER_FLAGS *alter_flags);
 
-  int alter_table_phase3(THD *thd, TABLE *table);
+  int alter_table_phase3(THD *thd, TABLE *table,
+                         HA_CREATE_INFO *create_info,
+                         HA_ALTER_INFO *alter_info,
+                         HA_ALTER_FLAGS *alter_flags);
 
 private:
+#ifdef HAVE_NDB_BINLOG
+  int delete_row_conflict_fn(enum_conflict_fn_type cft,
+                             const uchar *old_data,
+                             NdbInterpretedCode *);
+  int write_row_conflict_fn(enum_conflict_fn_type cft,
+                            uchar *data,
+                            NdbInterpretedCode *);
+  int update_row_conflict_fn(enum_conflict_fn_type cft,
+                             const uchar *old_data,
+                             uchar *new_data,
+                             NdbInterpretedCode *);
+  int row_conflict_fn_max(const uchar *new_data,
+                          NdbInterpretedCode *);
+  int row_conflict_fn_old(const uchar *old_data,
+                          NdbInterpretedCode *);
+#endif
+  void setup_key_ref_for_ndb_record(const NdbRecord **key_rec,
+                                    const uchar **key_row,
+                                    const uchar *record,
+                                    bool use_active_index);
   friend int ndbcluster_drop_database_impl(THD *thd, const char *path);
   friend int ndb_handle_schema_change(THD *thd, 
                                       Ndb *ndb, NdbEventOperation *pOp,
                                       NDB_SHARE *share);
 
-  static int delete_table(THD *thd, ha_ndbcluster *h, Ndb *ndb,
-			  const char *path,
-			  const char *db,
-			  const char *table_name);
+  void check_read_before_write_removal();
+  static int drop_table(THD *thd, ha_ndbcluster *h, Ndb *ndb,
+                        const char *path,
+                        const char *db,
+                        const char *table_name);
+
   int add_index_impl(THD *thd, TABLE *table_arg,
                      KEY *key_info, uint num_of_keys);
   int create_ndb_index(THD *thd, const char *name, KEY *key_info, bool unique);
@@ -495,7 +626,7 @@ private:
   void renumber_indexes(Ndb *ndb, TABLE *tab);
   int drop_indexes(Ndb *ndb, TABLE *tab);
   int add_index_handle(THD *thd, NdbDictionary::Dictionary *dict,
-                       KEY *key_info, const char *index_name, uint index_no);
+                       KEY *key_info, const char *key_name, uint index_no);
   int add_table_ndb_record(NdbDictionary::Dictionary *dict);
   int add_hidden_pk_ndb_record(NdbDictionary::Dictionary *dict);
   int add_index_ndb_record(NdbDictionary::Dictionary *dict,
@@ -525,13 +656,20 @@ private:
   int unique_index_read(const uchar *key, uint key_len, 
                         uchar *buf);
   int full_table_scan(const KEY* key_info, 
-                      const uchar *key, 
-                      uint key_len,
+                      const key_range *start_key,
+                      const key_range *end_key,
                       uchar *buf);
-  int flush_bulk_insert();
+  int flush_bulk_insert(bool allow_batch= FALSE);
   int ndb_write_row(uchar *record, bool primary_key_update,
                     bool batched_update);
-  int ndb_delete_row(const uchar *record, bool primary_key_update);
+
+  bool start_bulk_delete();
+  int bulk_delete_row(const uchar *record);
+  int end_bulk_delete();
+  int ndb_delete_row(const uchar *record, bool primary_key_update,
+                     bool is_bulk_delete= FALSE);
+
+  int ndb_optimize_table(THD* thd, uint delay);
 
   int alter_frm(THD *thd, const char *file, NDB_ALTER_DATA *alter_data);
 
@@ -543,6 +681,7 @@ private:
   int scan_handle_lock_tuple(NdbScanOperation *scanOp, NdbTransaction *trans);
   int fetch_next(NdbScanOperation* op);
   int set_auto_inc(THD *thd, Field *field);
+  int set_auto_inc_val(THD *thd, Uint64 value);
   int next_result(uchar *buf); 
   int close_scan();
   void unpack_record(uchar *dst_row, const uchar *src_row);
@@ -552,53 +691,41 @@ private:
   void set_dbname(const char *pathname);
   void set_tabname(const char *pathname);
 
-  uint offset_hidden_key() { return table->s->reclength; }
-  uint offset_user_partition_function() {
-    return table->s->reclength +
-      (table_share->primary_key == MAX_KEY ?
-           NDB_HIDDEN_PRIMARY_KEY_LENGTH : 0);
+  const NdbDictionary::Column *get_hidden_key_column() {
+    return m_table->getColumn(table_share->fields);
   }
-  uint offset_user_partition_fragment() {
-    return table->s->reclength +
-      (table_share->primary_key == MAX_KEY ?
-           NDB_HIDDEN_PRIMARY_KEY_LENGTH+4 : 4);
-  }
-  uint field_number_hidden_key() { return table->s->fields; }
-  uint field_number_user_partition_function() {
-    return table->s->fields +
-      (table_share->primary_key == MAX_KEY ? 1 : 0);
+  const NdbDictionary::Column *get_partition_id_column() {
+    Uint32 index= table_share->fields + (table_share->primary_key == MAX_KEY);
+    return m_table->getColumn(index);
   }
 
-  void set_hidden_key(uchar *row, Uint64 auto_value);
-  Uint64 get_hidden_key(const uchar *row);
-  void request_hidden_key(uchar *mask);
-  void set_partition_function_value(uchar *row, uint32 func_value);
-  uint32 get_partition_fragment(const uchar *row);
-  void request_partition_function_value(uchar *mask);
-  uchar *batch_copy_row_to_buffer(Thd_ndb *thd_ndb, const uchar *record,
-                                  bool & batch_full);
-  uchar *batch_copy_key_to_buffer(Thd_ndb *thd_ndb, const uchar *key,
-                                  uint key_len,
-                                  uint op_batch_size, bool & batch_full);
+  bool add_row_check_if_batch_full_size(Thd_ndb *thd_ndb, uint size);
+  bool add_row_check_if_batch_full(Thd_ndb *thd_ndb) {
+    return add_row_check_if_batch_full_size(thd_ndb, m_bytes_per_write);
+  }
+  uchar *get_buffer(Thd_ndb *thd_ndb, uint size);
   uchar *copy_row_to_buffer(Thd_ndb *thd_ndb, const uchar *record);
-  uchar *get_row_buffer();
-  void clear_extended_column_set(uchar *mask);
-  uchar *copy_column_set(MY_BITMAP *bitmap);
 
-  int get_blob_values(NdbOperation *ndb_op, uchar *dst_record,
+  int get_blob_values(const NdbOperation *ndb_op, uchar *dst_record,
                       const MY_BITMAP *bitmap);
-  int set_blob_values(NdbOperation *ndb_op, my_ptrdiff_t row_offset,
-                      const MY_BITMAP *bitmap, uint *set_count);
+  int set_blob_values(const NdbOperation *ndb_op, my_ptrdiff_t row_offset,
+                      const MY_BITMAP *bitmap, uint *set_count, bool batch);
   friend int g_get_ndb_blobs_value(NdbBlob *ndb_blob, void *arg);
-  void eventSetAnyValue(THD *thd, NdbOperation *op);
+  void release_blobs_buffer();
+  Uint32 setup_get_hidden_fields(NdbOperation::GetValueSpec gets[2]);
+  void get_hidden_fields_keyop(NdbOperation::OperationOptions *options,
+                               NdbOperation::GetValueSpec gets[2]);
+  void get_hidden_fields_scan(NdbScanOperation::ScanOptions *options,
+                              NdbOperation::GetValueSpec gets[2]);
+  void eventSetAnyValue(THD *thd, NdbOperation::OperationOptions *options);
   bool check_index_fields_in_write_set(uint keyno);
 
-  NdbOperation *pk_unique_index_read_key(uint idx, const uchar *key, uchar *buf,
-                                         NdbOperation::LockMode lm);
+  const NdbOperation *pk_unique_index_read_key(uint idx, 
+                                               const uchar *key, uchar *buf,
+                                               NdbOperation::LockMode lm,
+                                               Uint32 *ppartition_id);
   int read_multi_range_fetch_next();
   
-  int set_bounds(NdbIndexScanOperation*, uint inx, bool rir,
-                 const key_range *keys[2], uint= 0);
   int primary_key_cmp(const uchar * old_row, const uchar * new_row);
   void print_results();
 
@@ -608,6 +735,8 @@ private:
                                   ulonglong *nb_reserved_values);
   bool uses_blob_value(const MY_BITMAP *bitmap);
 
+  static inline bool isManualBinlogExec(THD *thd);
+
   char *update_table_comment(const char * comment);
 
   int write_ndb_file(const char *name);
@@ -620,16 +749,30 @@ private:
   void no_uncommitted_rows_update(int);
   void no_uncommitted_rows_reset(THD *);
 
-  void release_completed_operations(NdbTransaction*, bool);
 
-  friend int execute_commit(ha_ndbcluster*, NdbTransaction*);
-  friend int execute_no_commit_ignore_no_key(ha_ndbcluster*, NdbTransaction*);
-  friend int execute_no_commit(ha_ndbcluster*, NdbTransaction*, bool);
-  friend int execute_no_commit_ie(ha_ndbcluster*, NdbTransaction*, bool);
+  NdbTransaction *start_transaction_part_id(uint32 part_id, int &error);
+  inline NdbTransaction *get_transaction_part_id(uint32 part_id, int &error)
+  {
+    if (m_thd_ndb->trans)
+      return m_thd_ndb->trans;
+    return start_transaction_part_id(part_id, error);
+  }
 
-  void transaction_checks(THD *thd);
-  int start_statement(THD *thd, Thd_ndb *thd_ndb, Ndb* ndb);
-  int init_handler_for_statement(THD *thd, Thd_ndb *thd_ndb);
+  NdbTransaction *start_transaction(int &error);
+  inline NdbTransaction *get_transaction(int &error)
+  {
+    if (m_thd_ndb->trans)
+      return m_thd_ndb->trans;
+    return start_transaction(error);
+  }
+
+  friend int check_completed_operations_pre_commit(Thd_ndb*,
+                                                   NdbTransaction*,
+                                                   const NdbOperation*,
+                                                   uint *ignore_count);
+  friend int ndbcluster_commit(handlerton *hton, THD *thd, bool all);
+  int start_statement(THD *thd, Thd_ndb *thd_ndb, uint table_count);
+  int init_handler_for_statement(THD *thd);
 
   Thd_ndb *m_thd_ndb;
   NdbScanOperation *m_active_cursor;
@@ -639,23 +782,16 @@ private:
     fields (hidden primary key, user-defined partitioning function value).
   */
   NdbRecord *m_ndb_record;
-  /* As m_ndb_record, but adding the FRAGMENT pseudo-column at end of row. */
-  NdbRecord *m_ndb_record_fragment;
   /* NdbRecord for accessing tuple by hidden Uint64 primary key. */
   NdbRecord *m_ndb_hidden_key_record;
 
-  /*
-    Special NdbRecord for ndb_get_table_statistics(), reading lots of
-    pseudo-columns.
-  */
-  NdbRecord *m_ndb_statistics_record;
   /* Bitmap used for NdbRecord operation column mask. */
   MY_BITMAP m_bitmap;
   my_bitmap_map m_bitmap_buf[(NDB_MAX_ATTRIBUTES_IN_TABLE +
                               8*sizeof(my_bitmap_map) - 1) /
                              (8*sizeof(my_bitmap_map))]; // Buffer for m_bitmap
   /* Bitmap with bit set for all primary key columns. */
-  MY_BITMAP m_pk_bitmap;
+  MY_BITMAP *m_pk_bitmap_p;
   my_bitmap_map m_pk_bitmap_buf[(NDB_MAX_ATTRIBUTES_IN_TABLE +
                                  8*sizeof(my_bitmap_map) - 1) /
                                 (8*sizeof(my_bitmap_map))]; // Buffer for m_pk_bitmap
@@ -669,6 +805,8 @@ private:
   bool m_lock_tuple;
   NDB_SHARE *m_share;
   NDB_INDEX_DATA  m_index[MAX_KEY];
+  key_map btree_keys;
+
   /*
     Pointer to row returned from scan nextResult().
   */
@@ -679,17 +817,9 @@ private:
   };
   /* For read_multi_range scans, the get_range_no() of current row. */
   int m_current_range_no;
-  /*
-    A buffer of rows for when we cannot pass the mysqld record pointer directly
-    to the NDB API, either because the mysqld buffer is too small (eg. hidden
-    primary key), or because it will not remain valid until execute() (eg.
-    bulk insert).
-  */
-  char *m_row_buffer;
-  uint m_row_buffer_size;
-  char *m_row_buffer_current;
-  /* Extra bytes needed in row for hidden fields. */
-  uint m_extra_reclength;
+
+  MY_BITMAP **m_key_fields;
+  MY_BITMAP m_save_read_set;
   // NdbRecAttr has no reference to blob
   NdbValue m_value[NDB_MAX_ATTRIBUTES_IN_TABLE];
   Uint64 m_ref;
@@ -702,6 +832,10 @@ private:
   bool m_ignore_dup_key;
   bool m_has_unique_index;
   bool m_ignore_no_key;
+  bool m_read_before_write_removal_possible;
+  bool m_read_before_write_removal_used;
+  ha_rows m_rows_updated;
+  ha_rows m_rows_deleted;
   ha_rows m_rows_to_insert; // TODO: merge it with handler::estimation_rows_to_insert?
   ha_rows m_rows_inserted;
   ha_rows m_rows_changed;
@@ -714,16 +848,15 @@ private:
 
   /* State for setActiveHook() callback for reading blob data. */
   uint m_blob_counter;
-  uint m_blob_expected_count;
+  uint m_blob_expected_count_per_row;
   uchar *m_blob_destination_record;
-  Uint64 m_blob_total_size;
+  Uint64 m_blobs_row_total_size; /* Bytes needed for all blobs in current row */
   
   // memory for blobs in one tuple
   uchar *m_blobs_buffer;
   uint32 m_blobs_buffer_size;
   uint m_dupkey;
   // set from thread variables at external lock
-  bool m_force_send;
   ha_rows m_autoincrement_prefetch;
 
   ha_ndbcluster_cond *m_cond;
@@ -736,11 +869,10 @@ private:
     pieces.
   */
   KEY_MULTI_RANGE *m_multi_range_defined_end;
-  const NdbOperation *m_current_multi_operation;
   NdbIndexScanOperation *m_multi_cursor;
   Ndb *get_ndb(THD *thd);
 
-  int update_stats(THD *thd, bool do_read_stat);
+  int update_stats(THD *thd, bool do_read_stat, bool have_lock= FALSE);
 };
 
 extern SHOW_VAR ndb_status_variables[];
