@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008-2010 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef NdbOperation_H
 #define NdbOperation_H
@@ -419,6 +422,7 @@ public:
   int  setValue(const char* anAttrName, double aValue);
 #ifndef DOXYGEN_SHOULD_SKIP_INTERNAL
   int  setAnyValue(Uint32 aValue);
+  int  setOptimize(Uint32 options);
 #endif
 
 #ifndef DOXYGEN_SHOULD_SKIP_DEPRECATED
@@ -451,6 +455,8 @@ public:
    */
   virtual NdbBlob* getBlobHandle(const char* anAttrName);
   virtual NdbBlob* getBlobHandle(Uint32 anAttrId);
+  virtual NdbBlob* getBlobHandle(const char* anAttrName) const;
+  virtual NdbBlob* getBlobHandle(Uint32 anAttrId) const;
  
 #ifndef DOXYGEN_SHOULD_SKIP_INTERNAL
   /** @} *********************************************************************/
@@ -467,6 +473,9 @@ public:
    *
    * @note There are four versions of NdbOperation::incValue with
    *       slightly different parameters.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param anAttrName     Attribute name.
    * @param aValue         Value to add.
@@ -486,6 +495,9 @@ public:
    *
    * @note There are four versions of NdbOperation::subValue with
    *       slightly different parameters.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param anAttrName    Attribute name.
    * @param aValue        Value to subtract.
@@ -505,14 +517,20 @@ public:
    *       match the automatic numbering to make it easier to 
    *       debug the interpreted program.
    * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   * 
    * @param labelNumber   Label number.
-   * @return              -1 if unsuccessful.
+   * @return              Label number, -1 if unsuccessful.
    */
   int   def_label(int labelNumber);
 
   /**
    * Interpreted program instruction:
    * Add two registers into a third.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param RegSource1   First register.
    * @param RegSource2   Second register.
@@ -524,6 +542,9 @@ public:
   /**
    * Interpreted program instruction:
    * Substract RegSource2 from RegSource1 and put the result in RegDest.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param RegSource1   First register.
    * @param RegSource2   Second register.
@@ -535,6 +556,9 @@ public:
   /**
    * Interpreted program instruction:
    * Load a constant into a register.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param RegDest      Destination register.
    * @param Constant     Value to load.
@@ -546,6 +570,9 @@ public:
   /**
    * Interpreted program instruction:
    * Load NULL value into a register.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param RegDest      Destination register.
    * @return             -1 if unsuccessful.
@@ -555,6 +582,9 @@ public:
   /**
    * Interpreted program instruction:
    * Read an attribute into a register.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param anAttrName   Attribute name.
    * @param RegDest      Destination register.
@@ -565,6 +595,9 @@ public:
   /**
    * Interpreted program instruction:
    * Write an attribute from a register. 
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param anAttrName   Attribute name.
    * @param RegSource    Source register.
@@ -575,6 +608,9 @@ public:
   /**
    * Interpreted program instruction:
    * Read an attribute into a register.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param anAttrId the attribute id.
    * @param RegDest the destination register.
@@ -585,6 +621,9 @@ public:
   /**
    * Interpreted program instruction:
    * Write an attribute from a register. 
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param anAttrId the attribute id.
    * @param RegSource the source register.
@@ -605,6 +644,9 @@ public:
    * - lt RegR <  RegL
    * - eq RegR =  RegL
    * - ne RegR <> RegL
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param RegLvalue left value. 
    * @param RegRvalue right value.
@@ -621,6 +663,9 @@ public:
   /**
    * Interpreted program instruction:
    * Jump to Label if RegLvalue is not NULL.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param RegLvalue the value to check.
    * @param Label the label to jump to.
@@ -631,6 +676,9 @@ public:
   /**
    * Interpreted program instruction:
    * Jump to Label if RegLvalue is equal to NULL.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param  RegLvalue  Value to check.
    * @param  Label      Label to jump to.
@@ -641,6 +689,9 @@ public:
   /**
    * Interpreted program instruction:
    * Jump to Label.
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @param  Label  Label to jump to.
    * @return -1 if unsuccessful.
@@ -649,6 +700,10 @@ public:
 
   /**
    * Interpreted program instruction:  branch after memcmp
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   *
    * @param  ColId   Column to check
    * @param  Label   Label to jump to
    * @return -1 if unsuccessful
@@ -658,6 +713,10 @@ public:
 
   /**
    * Interpreted program instruction:  branch after memcmp
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   *
    * @param  ColId   column to check
    * @param  val     search value
    * @param  len     length of search value   
@@ -678,22 +737,60 @@ public:
   int branch_col_ge(Uint32 ColId, const void * val, Uint32 len, 
 		    bool nopad, Uint32 Label);
   /**
-   * The argument is always plain char, even if the field is varchar
+   * LIKE/NOTLIKE wildcard comparisons
+   * These instructions support SQL-style % and _ wildcards for
+   * (VAR)CHAR/BINARY columns only
+   *
+   * The argument is always plain char format, even if the field 
+   * is varchar
    * (changed in 5.0.22).
+   * 
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    */
   int branch_col_like(Uint32 ColId, const void *, Uint32 len, 
 		      bool nopad, Uint32 Label);
   int branch_col_notlike(Uint32 ColId, const void *, Uint32 len, 
 			 bool nopad, Uint32 Label);
-  
+
+  /**
+   * Bitwise logical comparisons
+   *
+   * These comparison types are only supported for the Bitfield
+   * type
+   * They can be used to test for bit patterns in bitfield columns
+   *   The value passed is a bitmask which is bitwise-ANDed with the
+   *   column data.
+   *   Bitfields are passed in/out of NdbApi as 32-bit words with
+   *   bits set from lsb to msb.
+   *   The platform's endianness controls which byte contains the ls
+   *   bits.
+   *     x86= first(0th) byte.  Sparc/PPC= last (3rd byte)
+   *
+   *   To set bit n of a bitmask to 1 from a Uint32* mask : 
+   *     mask[n >> 5] |= (1 << (n & 31))
+   *
+   *   The branch can be taken in 4 cases :
+   *     - Column data AND Mask == Mask (all masked bits are set in data)
+   *     - Column data AND Mask != Mask (not all masked bits are set in data)
+   *     - Column data AND Mask == 0    (No masked bits are set in data)
+   *     - Column data AND Mask != 0    (Some masked bits are set in data)
+   *   
+   */
+  int branch_col_and_mask_eq_mask(Uint32 ColId, const void *, Uint32 len, 
+                                  bool nopad, Uint32 Label);
+  int branch_col_and_mask_ne_mask(Uint32 ColId, const void *, Uint32 len, 
+                                  bool nopad, Uint32 Label);
+  int branch_col_and_mask_eq_zero(Uint32 ColId, const void *, Uint32 len, 
+                                  bool nopad, Uint32 Label);
+  int branch_col_and_mask_ne_zero(Uint32 ColId, const void *, Uint32 len, 
+                                  bool nopad, Uint32 Label);
+
   /**
    * Interpreted program instruction: Exit with Ok
    *
-   * For scanning transactions,
-   * end interpreted operation and return the row to the application.
-   *
-   * For non-scanning transactions,
-   * exit interpreted program.
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @return -1 if unsuccessful.
    */
@@ -702,11 +799,8 @@ public:
   /**
    * Interpreted program instruction: Exit with Not Ok
    *
-   * For scanning transactions, 
-   * continue with the next row without returning the current row.
-   *
-   * For non-scanning transactions,
-   * abort the whole transaction.
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
    *
    * @note A method also exists without the error parameter.
    * 
@@ -720,12 +814,11 @@ public:
   /**
    * Interpreted program instruction:
    *
-   * For scanning transactions, 
-   * return this row, but no more from this fragment
-   *
-   * For non-scanning transactions,
    * abort the whole transaction.
    *
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   *
    * @return            -1 if unsuccessful.
    */
   int interpret_exit_last_row();
@@ -734,6 +827,9 @@ public:
    * Interpreted program instruction:
    * Define a subroutine in an interpreted operation.
    *
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   *
    * @param SubroutineNumber the subroutine number.
    * @return -1 if unsuccessful.
    */
@@ -743,6 +839,9 @@ public:
    * Interpreted program instruction:
    * Call a subroutine.
    *
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   *
    * @param Subroutine the subroutine to call.
    * @return -1 if unsuccessful. 
    */
@@ -752,6 +851,9 @@ public:
    * Interpreted program instruction:
    * End a subroutine.
    *
+   * @note For Scans and NdbRecord operations, use the 
+   *       NdbInterpretedCode interface.
+   *
    * @return -1 if unsuccessful. 
    */
   int   ret_sub();
@@ -776,7 +878,10 @@ public:
    * 
    * @return method number where the error occured.
    */
+#ifndef DOXYGEN_SHOULD_SKIP_DEPRECATED
   int getNdbErrorLine();
+#endif
+  int getNdbErrorLine() const;
 
   /**
    * Get table name of this operation.
@@ -793,7 +898,7 @@ public:
   /**
    * Get the type of access for this operation
    */
-  const Type getType() const;
+  Type getType() const;
 
   /** @} *********************************************************************/
 
@@ -825,6 +930,11 @@ public:
    */
   AbortOption getAbortOption() const;
   int setAbortOption(AbortOption);
+
+  /**
+   * Get NdbTransaction object pointer for this operation
+   */
+  virtual NdbTransaction* getNdbTransaction() const;
   
 #ifndef DOXYGEN_SHOULD_SKIP_INTERNAL
   
@@ -834,8 +944,129 @@ public:
   void setPartitionId(Uint32 id);
   Uint32 getPartitionId() const;
 #endif
-protected:
-  int handle_distribution_key(const NdbColumnImpl*, const Uint64 *, Uint32 len);
+
+  /* Specification of an extra value to get
+   * as part of an NdbRecord operation.
+   * Inputs : 
+   *  To specify an extra value to read, the
+   *  caller must provide a column, and a 
+   *  (optionally NULL) appStorage pointer.
+   * Outputs : 
+   *  After the operation is defined, the 
+   *  recAttr member will contain a pointer
+   *  to the NdbRecAttr object for receiving
+   *  the data.
+   * 
+   * appStorage pointer
+   *  If the appStorage pointer is null, then
+   *  the received value will be stored in
+   *  memory managed by the NdbRecAttr object.
+   *
+   *  If the appStorage pointer is non-null then 
+   *  the received value will be stored at the 
+   *  location pointed to (and will still be 
+   *  accessable via the NdbRecAttr object).  
+   *  It is the caller's responsibility to 
+   *  ensure that :
+   *    - appStorage points to sufficient space 
+   *      to store any returned data.
+   *    - Memory pointed to by appStorage is not
+   *      reused/freed until after the execute()
+   *      call returns.
+   *
+   * Limitation : Blob reads cannot be specified 
+   * using GetValueSpec.
+   */
+  struct GetValueSpec
+  {
+    const NdbDictionary::Column *column;
+    void *appStorage;
+    NdbRecAttr *recAttr;
+  };
+  
+  /* Specification of an extra value to set
+   * as part of an NdbRecord operation.
+   * The value ptr must point to the value
+   * to set, or NULL if the attribute is to
+   * be set to NULL.
+   * The pointed to value is copied when the 
+   * operation is defined and need not remain
+   * in place until execution time.
+   *
+   * Limitation : Blobs cannot be set using 
+   * SetValueSpec.
+   */
+  struct SetValueSpec
+  {
+    const NdbDictionary::Column *column;
+    const void * value;
+  };
+
+  /*
+   * OperationOptions
+   *  These are options passed to the NdbRecord primary key and scan 
+   *  takeover operation methods defined in the NdbTransaction and 
+   *  NdbScanOperation classes.
+   *  
+   *  Each option type is marked as present by setting the corresponding
+   *  bit in the optionsPresent field.  Only the option types marked in the
+   *  optionsPresent structure need have sensible data.
+   *  All data is copied out of the OperationOptions structure (and any
+   *  subtended structures) at operation definition time.
+   *  If no options are required, then NULL may be passed as the 
+   *  OperationOptions pointer.
+   *
+   *  Most methods take a supplementary sizeOfOptions parameter.  This
+   *  is optional, and is intended to allow the interface implementation
+   *  to remain backwards compatible with older un-recompiled clients 
+   *  that may pass an older (smaller) version of the OperationOptions 
+   *  structure.  This effect is achieved by passing
+   *  sizeof(OperationOptions) into this parameter.
+   */
+  struct OperationOptions
+  {
+    /*
+     * Which options are present.  See below for option details
+     */
+    Uint64 optionsPresent;
+    enum Flags { OO_ABORTOPTION  = 0x01,
+                 OO_GETVALUE     = 0x02, 
+                 OO_SETVALUE     = 0x04, 
+                 OO_PARTITION_ID = 0x08, 
+                 OO_INTERPRETED  = 0x10,
+                 OO_ANYVALUE     = 0x20,
+                 OO_CUSTOMDATA   = 0x40 };
+
+    /* An operation-specific abort option.
+     * Only necessary if the default abortoption behaviour
+     * is not satisfactory 
+     */
+    AbortOption abortOption;
+
+    /* Extra column values to be read */
+    GetValueSpec *extraGetValues;
+    Uint32        numExtraGetValues;
+    
+    /* Extra column values to be set  */
+    const SetValueSpec *extraSetValues;
+    Uint32              numExtraSetValues;
+
+    /* Specific partition to execute this operation on */
+    Uint32 partitionId;
+
+    /* Interpreted code to be executed in this operation
+     * Only supported for update operations currently 
+     */
+    const NdbInterpretedCode *interpretedCode;
+
+    /* anyValue to be used for this operation */
+    Uint32 anyValue;
+
+    /* customData ptr for this operation */
+    void * customData;
+  };
+
+
 protected:
 /******************************************************************************
  * These are the methods used to create and delete the NdbOperation objects.
@@ -858,9 +1089,13 @@ protected:
 
 public:
 #ifndef DOXYGEN_SHOULD_SKIP_INTERNAL
-  NdbTransaction* getNdbTransaction();
   const NdbOperation* next() const;
   const NdbRecAttr* getFirstRecAttr() const;
+
+  void* getCustomData() const { return m_customData; }
+  void setCustomData(void* p) { m_customData = p; }
+protected:
+  void* m_customData;
 #endif
 protected:
 
@@ -952,6 +1187,15 @@ protected:
 					      	// the operations object.      
   void		    setStartIndicator();
 
+  /* Utility method to 'add' operation options to an NdbOperation
+   *
+   * @return 0 for success.  NDBAPI to set error otherwise.
+   */
+  static int        handleOperationOptions (const OperationType type,
+                                            const OperationOptions *opts,
+                                            const Uint32 sizeOfOptions,
+                                            NdbOperation *op);
+
 /******************************************************************************
  * The methods below is the execution part of the NdbOperation
  * class. This is where the NDB signals are sent and received. The
@@ -969,14 +1213,37 @@ protected:
     
   int	 prepareSendInterpreted();            // Help routine to prepare*
 
-  int    prepareSendNdbRecord(Uint32 aTC_ConnectPtr, Uint64 aTransId,
-                              AbortOption ao);
+  int initInterpretedInfo(const NdbInterpretedCode *code,
+                          Uint32*& interpretedInfo,
+                          Uint32* stackSpace,
+                          Uint32 stackSpaceEntries,
+                          Uint32*& dynamicSpace);
 
-  /* Helper routines for prepareSendNdbRecord(). */
+  void freeInterpretedInfo(Uint32*& dynamicSpace);
+
+
+  /* Method for adding signals for an interpreted program
+   * to the signal train 
+   */
+  int buildInterpretedProgramSignals(Uint32 aTC_ConnectPtr, 
+                                     Uint64 aTransId,
+                                     Uint32 **attrInfoPtr,
+                                     Uint32 *remain,
+                                     const NdbInterpretedCode *code,
+                                     Uint32 *interpretedWorkspace,
+                                     bool mainProgram,
+                                     Uint32 &wordsWritten);
+
+  // Method which prepares signals at operation definition time.
+  int    buildSignalsNdbRecord(Uint32 aTC_ConnectPtr, Uint64 aTransId);
+
+  // Method which does final preparations at execute time.
+  int    prepareSendNdbRecord(AbortOption ao);
+
+  /* Helper routines for buildSignalsNdbRecord(). */
   Uint32 fillTcKeyReqHdr(TcKeyReq *tcKeyReq,
                          Uint32 connectPtr,
-                         Uint64 transId,
-                         AbortOption ao);
+                         Uint64 transId);
   int    allocKeyInfo(Uint32 connectPtr, Uint64 transId,
                       Uint32 **dstPtr, Uint32 *remain);
   int    allocAttrInfo(Uint32 connectPtr, Uint64 transId,
@@ -1015,6 +1282,7 @@ protected:
   NdbRecAttr* getValue_NdbRecord(const NdbColumnImpl* tAttrInfo, char* aValue);
   int setValue(const NdbColumnImpl* anAttrObject, const char* aValue);
   NdbBlob* getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* anAttrObject);
+  NdbBlob* getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* anAttrObject) const;
   int incValue(const NdbColumnImpl* anAttrObject, Uint32 aValue);
   int incValue(const NdbColumnImpl* anAttrObject, Uint64 aValue);
   int subValue(const NdbColumnImpl* anAttrObject, Uint32 aValue);
@@ -1022,13 +1290,14 @@ protected:
   int read_attr(const NdbColumnImpl* anAttrObject, Uint32 RegDest);
   int write_attr(const NdbColumnImpl* anAttrObject, Uint32 RegSource);
   int branch_reg_reg(Uint32 type, Uint32, Uint32, Uint32);
-  int branch_col(Uint32 type, Uint32, const void *, Uint32, bool, Uint32 Label);
+  int branch_col(Uint32 type, Uint32, const void *, Uint32, Uint32 Label);
   int branch_col_null(Uint32 type, Uint32 col, Uint32 Label);
   NdbBlob *linkInBlobHandle(NdbTransaction *aCon,
                             const NdbColumnImpl *column,
                             NdbBlob * & lastPtr);
   int getBlobHandlesNdbRecord(NdbTransaction* aCon);
-  int getBlobHandlesDelete(NdbTransaction* aCon);  
+  int getBlobHandlesNdbRecordDelete(NdbTransaction* aCon, bool checkReadSet);
+
   // Handle ATTRINFO signals   
   int insertATTRINFO(Uint32 aData);
   int insertATTRINFOloop(const Uint32* aDataPtr, Uint32 aLength);
@@ -1038,9 +1307,10 @@ protected:
 		    Uint32 aKeyLenInByte);
   void reorderKEYINFO();
   
-  virtual void setErrorCode(int aErrorCode);
-  virtual void setErrorCodeAbort(int aErrorCode);
+  virtual void setErrorCode(int aErrorCode) const;
+  virtual void setErrorCodeAbort(int aErrorCode) const;
 
+  bool        isNdbRecordOperation();
   int	      incCheck(const NdbColumnImpl* anAttrObject);
   int	      initial_interpreterCheck();
   int	      intermediate_interpreterCheck();
@@ -1190,6 +1460,10 @@ protected:
   /* Interpreted program for NdbRecord operations. */
   const NdbInterpretedCode *m_interpreted_code;
 
+  /* Ptr to supplied SetValueSpec for NdbRecord */
+  const SetValueSpec *m_extraSetValues;
+  Uint32 m_numExtraSetValues;
+
   Uint32 m_any_value;                           // Valid if m_use_any_value!=0
 
   // Blobs in this operation
@@ -1217,6 +1491,10 @@ protected:
   friend struct Ndb_free_list_t<NdbOperation>;
 
   Uint32 repack_read(Uint32 len);
+
+private:
+  NdbOperation(const NdbOperation&); // Not impl.
+  NdbOperation&operator=(const NdbOperation&);
 };
 
 #ifdef NDB_NO_DROPPED_SIGNAL
@@ -1229,6 +1507,9 @@ inline
 int
 NdbOperation::checkMagicNumber(bool b)
 {
+#ifndef NDB_NO_DROPPED_SIGNAL
+  (void)b;  // unused param in this context
+#endif
   if (theMagicNumber != 0xABCDEF01){
 #ifdef NDB_NO_DROPPED_SIGNAL
     if(b) abort();
@@ -1248,6 +1529,15 @@ NdbOperation::setStartIndicator()
 inline
 int
 NdbOperation::getNdbErrorLine()
+{
+  // delegate to overloaded const function for same semantics
+  const NdbOperation * const cthis = this;
+  return cthis->NdbOperation::getNdbErrorLine();
+}
+
+inline
+int
+NdbOperation::getNdbErrorLine() const
 {
   return theErrorLine;
 }
@@ -1299,7 +1589,7 @@ Return Value    Return the Type.
 Remark:         Gets type of access.
 ******************************************************************************/
 inline
-const NdbOperation::Type
+NdbOperation::Type
 NdbOperation::getType() const
 {
   return m_type;
@@ -1348,8 +1638,10 @@ NdbOperation::NdbCon(NdbTransaction* aNdbCon)
 
 inline
 int
-NdbOperation::equal(const char* anAttrName, const char* aValue, Uint32 len)
+NdbOperation::equal(const char* anAttrName, const char* aValue,
+                    Uint32 len)
 {
+  (void)len;   // unused
   return equal(anAttrName, aValue);
 }
 
@@ -1383,8 +1675,10 @@ NdbOperation::equal(const char* anAttrName, Uint64 aPar)
 
 inline
 int
-NdbOperation::equal(Uint32 anAttrId, const char* aValue, Uint32 len)
+NdbOperation::equal(Uint32 anAttrId, const char* aValue,
+                    Uint32 len)
 {
+  (void)len;   // unused
   return equal(anAttrId, aValue);
 }
 
@@ -1418,8 +1712,10 @@ NdbOperation::equal(Uint32 anAttrId, Uint64 aPar)
 
 inline
 int
-NdbOperation::setValue(const char* anAttrName, const char* aValue, Uint32 len)
+NdbOperation::setValue(const char* anAttrName, const char* aValue,
+                       Uint32 len)
 {
+  (void)len;   // unused
   return setValue(anAttrName, aValue);
 }
 
@@ -1467,8 +1763,10 @@ NdbOperation::setValue(const char* anAttrName, double aPar)
 
 inline
 int
-NdbOperation::setValue(Uint32 anAttrId, const char* aValue, Uint32 len)
+NdbOperation::setValue(Uint32 anAttrId, const char* aValue,
+                       Uint32 len)
 {
+  (void)len;   // unused
   return setValue(anAttrId, aValue);
 }
 
