@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <my_pthread.h>
@@ -47,7 +49,7 @@ extern int g_ndb_shm_signum;
 #include <mgmapi/mgmapi_debug.h>
 
 #include <EventLogger.hpp>
-extern EventLogger g_eventLogger;
+extern EventLogger * g_eventLogger;
 
 struct in_addr
 TransporterRegistry::get_connect_address(NodeId node_id) const
@@ -63,7 +65,8 @@ SocketServer::Session * TransporterService::newSession(NDB_SOCKET_TYPE sockfd)
     DBUG_RETURN(0);
   }
 
-  if (!m_transporter_registry->connect_server(sockfd))
+  BaseString msg;
+  if (!m_transporter_registry->connect_server(sockfd, msg))
   {
     NDB_CLOSE_SOCKET(sockfd);
     DBUG_RETURN(0);
@@ -93,7 +96,34 @@ TransporterRegistry::TransporterRegistry(void * callback,
   theTransporters     = new Transporter     * [maxTransporters];
   performStates       = new PerformState      [maxTransporters];
   ioStates            = new IOState           [maxTransporters]; 
-  
+ 
+  m_has_extra_wakeup_socket = false;
+#if defined(HAVE_EPOLL_CREATE)
+ m_epoll_fd = -1;
+ m_epoll_events       = new struct epoll_event[maxTransporters];
+ m_epoll_fd = epoll_create(maxTransporters);
+ if (m_epoll_fd == -1 || !m_epoll_events)
+ {
+   /* Failure to allocate data or get epoll socket, abort */
+   perror("Failed to alloc epoll-array or calling epoll_create... falling back to select!");
+   if (m_epoll_fd != -1)
+   {
+     close(m_epoll_fd);
+     m_epoll_fd = -1;
+   }
+   if (m_epoll_events)
+   {
+     delete [] m_epoll_events;
+     m_epoll_events = 0;
+   }
+ }
+ else
+ {
+   memset((char*)m_epoll_events, 0,
+          maxTransporters * sizeof(struct epoll_event));
+ }
+
+#endif
   // Initialize member variables
   nTransporters    = 0;
   nTCPTransporters = 0;
@@ -149,9 +179,19 @@ TransporterRegistry::~TransporterRegistry()
   delete[] performStates;
   delete[] ioStates;
 
+#if defined(HAVE_EPOLL_CREATE)
+  if (m_epoll_events) delete [] m_epoll_events;
+  if (m_epoll_fd != -1) close(m_epoll_fd);
+#endif
   if (m_mgm_handle)
     ndb_mgm_destroy_handle(&m_mgm_handle);
 
+  if (m_has_extra_wakeup_socket)
+  {
+    close(m_extra_wakeup_sockets[0]);
+    close(m_extra_wakeup_sockets[1]);
+  }
+
   DBUG_VOID_RETURN;
 }
 
@@ -178,12 +218,16 @@ TransporterRegistry::init(NodeId nodeId) {
   localNodeId = nodeId;
   
   DEBUG("TransporterRegistry started node: " << localNodeId);
-  
+
+  if (!m_socket_poller.set_max_count(maxTransporters +
+				     1 /* wakeup socket */))
+    DBUG_RETURN(false);
+
   DBUG_RETURN(true);
 }
 
 bool
-TransporterRegistry::connect_server(NDB_SOCKET_TYPE sockfd)
+TransporterRegistry::connect_server(NDB_SOCKET_TYPE sockfd, BaseString& msg)
 {
   DBUG_ENTER("TransporterRegistry::connect_server");
 
@@ -192,7 +236,10 @@ TransporterRegistry::connect_server(NDB_SOCKET_TYPE sockfd)
   int nodeId, remote_transporter_type= -1;
   SocketInputStream s_input(sockfd);
   char buf[256];
-  if (s_input.gets(buf, 256) == 0) {
+  if (s_input.gets(buf, 256) == 0)
+  {
+    msg.assfmt("line: %u : Failed to get nodeid from client",
+               __LINE__);
     DBUG_PRINT("error", ("Could not get node id from client"));
     DBUG_RETURN(false);
   }
@@ -205,6 +252,8 @@ TransporterRegistry::connect_server(NDB_SOCKET_TYPE sockfd)
     // ok, but with no checks on transporter configuration compatability
     break;
   default:
+    msg.assfmt("line: %u : Incorrect reply from client: >%s<",
+               __LINE__, buf);
     DBUG_PRINT("error", ("Error in node id from client"));
     DBUG_RETURN(false);
   }
@@ -214,16 +263,25 @@ TransporterRegistry::connect_server(NDB_SOCKET_TYPE sockfd)
 
   //check that nodeid is valid and that there is an allocated transporter
   if ( nodeId < 0 || nodeId >= (int)maxTransporters) {
+    msg.assfmt("line: %u : Incorrect reply from client: >%s<",
+               __LINE__, buf);
     DBUG_PRINT("error", ("Node id out of range from client"));
     DBUG_RETURN(false);
   }
-  if (theTransporters[nodeId] == 0) {
-      DBUG_PRINT("error", ("No transporter for this node id from client"));
-      DBUG_RETURN(false);
+  if (theTransporters[nodeId] == 0)
+  {
+    msg.assfmt("line: %u : Incorrect reply from client: >%s<, node: %u",
+               __LINE__, buf, nodeId);
+    DBUG_PRINT("error", ("No transporter for this node id from client"));
+    DBUG_RETURN(false);
   }
 
   //check that the transporter should be connected
   if (performStates[nodeId] != TransporterRegistry::CONNECTING) {
+    msg.assfmt("line: %u : Incorrect state for node %u state: %s (%u)",
+               __LINE__, nodeId,
+               getPerformStateString(performStates[nodeId]),
+               performStates[nodeId]);
     DBUG_PRINT("error", ("Transporter in wrong state for this node id from client"));
     DBUG_RETURN(false);
   }
@@ -241,30 +299,30 @@ TransporterRegistry::connect_server(NDB_SOCKET_TYPE sockfd)
     {
       DBUG_PRINT("error", ("Transporter types mismatch this=%d remote=%d",
 			   t->m_type, remote_transporter_type));
-      g_eventLogger.error("Incompatible configuration: Transporter type "
-			  "mismatch with node %d", nodeId);
+      g_eventLogger->error("Incompatible configuration: Transporter type "
+                           "mismatch with node %d", nodeId);
 
-      // wait for socket close for 1 second to let message arrive at client
-      {
-	fd_set a_set;
-	FD_ZERO(&a_set);
-	FD_SET(sockfd, &a_set);
-	struct timeval timeout;
-	timeout.tv_sec  = 1; timeout.tv_usec = 0;
-	select(sockfd+1, &a_set, 0, 0, &timeout);
-      }
       DBUG_RETURN(false);
     }
   }
   else if (t->m_type == tt_SHM_TRANSPORTER)
   {
-    g_eventLogger.warning("Unable to verify transporter compatability with node %d", nodeId);
+    g_eventLogger->warning("Unable to verify transporter compatability with node %d", nodeId);
   }
 
   // setup transporter (transporter responsible for closing sockfd)
-  t->connect_server(sockfd);
+  bool res = t->connect_server(sockfd, msg);
 
-  DBUG_RETURN(true);
+  if (res && performStates[nodeId] != TransporterRegistry::CONNECTING)
+  {
+    msg.assfmt("line: %u : Incorrect state for node %u state: %s (%u)",
+               __LINE__, nodeId,
+               getPerformStateString(performStates[nodeId]),
+               performStates[nodeId]);
+    DBUG_RETURN(false);
+  }
+
+  DBUG_RETURN(res);
 }
 
 bool
@@ -518,7 +576,7 @@ TransporterRegistry::prepareSend(const SignalHeader * const signalHeader,
 	 
     if(t->isConnected()){
       Uint32 lenBytes = t->m_packer.getMessageLength(signalHeader, ptr);
-      if(lenBytes <= MAX_MESSAGE_SIZE){
+      if(lenBytes <= MAX_SEND_MESSAGE_BYTESIZE){
 	Uint32 * insertPtr = t->getWritePtr(lenBytes, prio);
 	if(insertPtr != 0){
 	  t->m_packer.pack(insertPtr, prio, signalHeader, signalData, ptr);
@@ -590,7 +648,7 @@ TransporterRegistry::prepareSend(const SignalHeader * const signalHeader,
     
     if(t->isConnected()){
       Uint32 lenBytes = t->m_packer.getMessageLength(signalHeader, ptr);
-      if(lenBytes <= MAX_MESSAGE_SIZE){
+      if(lenBytes <= MAX_SEND_MESSAGE_BYTESIZE){
 	Uint32 * insertPtr = t->getWritePtr(lenBytes, prio);
 	if(insertPtr != 0){
 	  t->m_packer.pack(insertPtr, prio, signalHeader, signalData, thePool, ptr);
@@ -659,6 +717,66 @@ TransporterRegistry::external_IO(Uint32 timeOutMillis) {
   performSend();
 }
 
+bool
+TransporterRegistry::setup_wakeup_socket()
+{
+  if (m_has_extra_wakeup_socket)
+  {
+    return true;
+  }
+
+  if (socketpair(AF_UNIX, SOCK_STREAM, 0, m_extra_wakeup_sockets))
+  {
+    perror("socketpair failed!");
+    return false;
+  }
+
+  if (!TCP_Transporter::setSocketNonBlocking(m_extra_wakeup_sockets[0]) ||
+      !TCP_Transporter::setSocketNonBlocking(m_extra_wakeup_sockets[1]))
+  {
+    goto err;
+  }
+
+#if defined(HAVE_EPOLL_CREATE)
+  if (m_epoll_fd != -1)
+  {
+    int sock = m_extra_wakeup_sockets[0];
+    struct epoll_event event_poll;
+    bzero(&event_poll, sizeof(event_poll));
+    event_poll.data.u32 = 0;
+    event_poll.events = EPOLLIN;
+    int ret_val = epoll_ctl(m_epoll_fd, EPOLL_CTL_ADD, sock, &event_poll);
+    if (ret_val != 0)
+    {
+      int error= errno;
+      fprintf(stderr, "Failed to add extra sock %u to epoll-set: %u\n",
+              sock, error);
+      fflush(stderr);
+      goto err;
+    }
+  }
+#endif
+  m_has_extra_wakeup_socket = true;
+  return true;
+
+err:
+  close(m_extra_wakeup_sockets[0]);
+  close(m_extra_wakeup_sockets[1]);
+  m_extra_wakeup_sockets[0] = NDB_INVALID_SOCKET;
+  m_extra_wakeup_sockets[1] = NDB_INVALID_SOCKET;
+  return false;
+}
+
+void
+TransporterRegistry::wakeup()
+{
+  if (m_has_extra_wakeup_socket)
+  {
+    static char c = 37; // not looked at on receiver side
+    send(m_extra_wakeup_sockets[1], &c, 1, 0);
+  }
+}
+
 Uint32
 TransporterRegistry::pollReceive(Uint32 timeOutMillis){
   Uint32 retVal = 0;
@@ -681,12 +799,38 @@ TransporterRegistry::pollReceive(Uint32 timeOutMillis){
 #endif
 
 #ifdef NDB_TCP_TRANSPORTER
-  if(nTCPTransporters > 0 || retVal == 0)
+#if defined(HAVE_EPOLL_CREATE)
+  if (likely(m_epoll_fd != -1))
   {
-    retVal |= poll_TCP(timeOutMillis);
+    Uint32 num_trps = nTCPTransporters;
+    /**
+     * If any transporters have left-over data that was not fully executed in
+     * last loop, don't wait and return 'data available' even if nothing new
+     * from epoll.
+     */
+    if (!m_has_data_transporters.isclear())
+    {
+      timeOutMillis = 0;
+      retVal = 1;
+    }
+    
+    if (num_trps)
+    {
+      tcpReadSelectReply = epoll_wait(m_epoll_fd, m_epoll_events,
+                                      num_trps, timeOutMillis);
+      retVal |= tcpReadSelectReply;
+    }
   }
   else
-    tcpReadSelectReply = 0;
+#endif
+  {
+    if(nTCPTransporters > 0 || m_has_extra_wakeup_socket || retVal == 0)
+    {
+      retVal |= poll_TCP(timeOutMillis);
+    }
+    else
+      tcpReadSelectReply = 0;
+  }
 #endif
 #ifdef NDB_SCI_TRANSPORTER
   if(nSCITransporters > 0)
@@ -709,7 +853,8 @@ TransporterRegistry::poll_SCI(Uint32 timeOutMillis)
 {
   for (int i=0; i<nSCITransporters; i++) {
     SCI_Transporter * t = theSCITransporters[i];
-    if (t->isConnected()) {
+    Uint32 node_id= t->getRemoteNodeId();
+    if (t->isConnected() && is_connected(node_id)) {
       if(t->hasDataToRead())
 	return 1;
     }
@@ -728,7 +873,8 @@ TransporterRegistry::poll_SHM(Uint32 timeOutMillis)
   {
     for (int i=0; i<nSHMTransporters; i++) {
       SHM_Transporter * t = theSHMTransporters[i];
-      if (t->isConnected()) {
+      Uint32 node_id= t->getRemoteNodeId();
+      if (t->isConnected() && is_connected(node_id)) {
 	if(t->hasDataToRead()) {
 	  return 1;
 	}
@@ -744,90 +890,193 @@ Uint32
 TransporterRegistry::poll_TCP(Uint32 timeOutMillis)
 {
   bool hasdata = false;
-  if (false && nTCPTransporters == 0)
-  {
-    tcpReadSelectReply = 0;
-    return 0;
-  }
-  
-  NDB_SOCKET_TYPE maxSocketValue = -1;
-  
-  // Needed for TCP/IP connections
-  // The read- and writeset are used by select
-  
-  FD_ZERO(&tcpReadset);
 
-  // Prepare for sending and receiving
-  for (int i = 0; i < nTCPTransporters; i++) {
+  m_socket_poller.clear();
+
+  if (m_has_extra_wakeup_socket)
+  {
+    const NDB_SOCKET_TYPE socket = m_extra_wakeup_sockets[0];
+
+    // Poll the wakup-socket for read
+    m_socket_poller.add(socket, true, false, false);
+  }
+
+  for (int i = 0; i < nTCPTransporters; i++)
+  {
+    unsigned index = ~0;
     TCP_Transporter * t = theTCPTransporters[i];
-    
-    // If the transporter is connected
-    NodeId nodeId = t->getRemoteNodeId();
-    if (is_connected(nodeId) && t->isConnected()) {
-      
-      const NDB_SOCKET_TYPE socket = t->getSocket();
-      // Find the highest socket value. It will be used by select
-      if (socket > maxSocketValue)
-	maxSocketValue = socket;
-      
-      // Put the connected transporters in the socket read-set 
-      FD_SET(socket, &tcpReadset);
+    const Uint32 node_id = t->getRemoteNodeId();
+    const NDB_SOCKET_TYPE socket = t->getSocket();
+
+    if (is_connected(node_id) && t->isConnected() &&
+        my_socket_valid(socket))
+    {
+      // Poll the connected transporter for read
+      index = m_socket_poller.add(socket, true, false, false);
     }
+    // Remember the index into poll list
+    t->set_poll_index(index);
+
     hasdata |= t->hasReceiveData();
   }
-  
-  timeOutMillis = hasdata ? 0 : timeOutMillis;
-  
-  struct timeval timeout;
-  timeout.tv_sec  = timeOutMillis / 1000;
-  timeout.tv_usec = (timeOutMillis % 1000) * 1000;
 
-  // The highest socket value plus one
-  maxSocketValue++; 
-  
-  tcpReadSelectReply = select(maxSocketValue, &tcpReadset, 0, 0, &timeout);  
-  if(false && tcpReadSelectReply == -1 && errno == EINTR)
-    g_eventLogger.info("woke-up by signal");
+  tcpReadSelectReply = m_socket_poller.poll_unsafe(hasdata ? 0 : timeOutMillis);
 
-#ifdef NDB_WIN32
-  if(tcpReadSelectReply == SOCKET_ERROR)
-  {
-    NdbSleep_MilliSleep(timeOutMillis);
-  }
-#endif
-  
   return tcpReadSelectReply || hasdata;
 }
 #endif
 
+#if defined(HAVE_EPOLL_CREATE)
+bool
+TransporterRegistry::change_epoll(TCP_Transporter *t, bool add)
+{
+  struct epoll_event event_poll;
+  bzero(&event_poll, sizeof(event_poll));
+  int sock_fd = t->getSocket();
+  int node_id = t->getRemoteNodeId();
+  int op = add ? EPOLL_CTL_ADD : EPOLL_CTL_DEL;
+  int ret_val, error;
+
+  if (sock_fd == NDB_INVALID_SOCKET)
+    return FALSE;
+
+  event_poll.data.u32 = t->getRemoteNodeId();
+  event_poll.events = EPOLLIN;
+  ret_val = epoll_ctl(m_epoll_fd, op, sock_fd, &event_poll);
+  if (!ret_val)
+    goto ok;
+  error= errno;
+  if (error == ENOENT && !add)
+  {
+    /*
+     * Could be that socket was closed premature to this call.
+     * Not a problem that this occurs.
+     */
+    goto ok;
+  }
+  if (!add || (add && (error != ENOMEM)))
+  {
+    /*
+     * Serious problems, we are either using wrong parameters,
+     * have permission problems or the socket doesn't support
+     * epoll!!
+     */
+    ndbout_c("Failed to %s epollfd: %u fd %u node %u to epoll-set,"
+             " errno: %u %s",
+             add ? "ADD" : "DEL",
+             m_epoll_fd,
+             sock_fd,
+             node_id,
+             error,
+             strerror(error));
+    abort();
+  }
+  ndbout << "We lacked memory to add the socket for node id ";
+  ndbout << node_id << endl;
+  return TRUE;
+
+ok:
+  return FALSE;
+}
+
+void
+TransporterRegistry::get_tcp_data(TCP_Transporter *t)
+{
+  const NodeId node_id = t->getRemoteNodeId();
+  bool hasdata = false;
+  checkJobBuffer();
+  if (is_connected(node_id) && t->isConnected())
+  {
+    t->doReceive();
+    
+    Uint32 *ptr;
+    Uint32 sz = t->getReceiveData(&ptr);
+    transporter_recv_from(callbackObj, node_id);
+    Uint32 szUsed = unpack(ptr, sz, node_id, ioStates[node_id]);
+    t->updateReceiveDataPtr(szUsed);
+    hasdata = t->hasReceiveData();
+  }
+  m_has_data_transporters.set(node_id, hasdata);
+}
+
+#endif
 
 void
 TransporterRegistry::performReceive()
 {
 #ifdef NDB_TCP_TRANSPORTER
-  for (int i=0; i<nTCPTransporters; i++) 
+#if defined(HAVE_EPOLL_CREATE)
+  if (likely(m_epoll_fd != -1))
   {
-    checkJobBuffer();
-    TCP_Transporter *t = theTCPTransporters[i];
-    const NodeId nodeId = t->getRemoteNodeId();
-    const NDB_SOCKET_TYPE socket    = t->getSocket();
-    if(is_connected(nodeId)){
-      if(t->isConnected())
+    int num_socket_events = tcpReadSelectReply;
+    int i;
+    
+    if (num_socket_events > 0)
+    {
+      for (i = 0; i < num_socket_events; i++)
       {
-        if (FD_ISSET(socket, &tcpReadset))
-	{
-	  t->doReceive();
-        }
+        m_has_data_transporters.set(m_epoll_events[i].data.u32);
+      }
+    }
+    else if (num_socket_events < 0)
+    {
+      assert(errno == EINTR);
+    }
 
-        if (t->hasReceiveData())
+    if (m_has_data_transporters.get(0))
+    {
+      m_has_data_transporters.clear(Uint32(0));
+      consume_extra_sockets();
+    }
+    
+    Uint32 id = 0;
+    while ((id = m_has_data_transporters.find(id + 1)) != BitmaskImpl::NotFound)
+    {
+      get_tcp_data((TCP_Transporter*)theTransporters[id]);
+    }
+  }
+  else
+#endif
+  {
+    if (m_has_extra_wakeup_socket)
+    {
+      // The wakeupsocket is always added first => use index 0
+      if (m_socket_poller.has_read(0))
+      {
+        assert(m_socket_poller.is_socket_equal(0, m_extra_wakeup_sockets[0]));
+        consume_extra_sockets();
+      }
+    }
+
+    for (int i=0; i<nTCPTransporters; i++) 
+    {
+      checkJobBuffer();
+      TCP_Transporter *t = theTCPTransporters[i];
+      const NodeId nodeId = t->getRemoteNodeId();
+
+      if(is_connected(nodeId)){
+        if(t->isConnected())
         {
-          Uint32 * ptr;
-          Uint32 sz = t->getReceiveData(&ptr);
-          transporter_recv_from(callbackObj, nodeId);
-          Uint32 szUsed = unpack(ptr, sz, nodeId, ioStates[nodeId]);
-          t->updateReceiveDataPtr(szUsed);
-	}
-      } 
+          const unsigned index = t->get_poll_index();
+          if (index != (unsigned)~0 &&
+              m_socket_poller.has_read(index))
+          {
+            assert(m_socket_poller.is_socket_equal(index, t->getSocket()));
+            t->doReceive();
+          }
+	  // Reset the index into poll list
+	  t->set_poll_index(~0);
+          
+          if (t->hasReceiveData())
+          {
+            Uint32 * ptr;
+            Uint32 sz = t->getReceiveData(&ptr);
+            transporter_recv_from(callbackObj, nodeId);
+            Uint32 szUsed = unpack(ptr, sz, nodeId, ioStates[nodeId]);
+            t->updateReceiveDataPtr(szUsed);
+          }
+        } 
+      }
     }
   }
 #endif
@@ -873,6 +1122,20 @@ TransporterRegistry::performReceive()
 #endif
 }
 
+void
+TransporterRegistry::consume_extra_sockets()
+{
+  char buf[4096];
+  int ret;
+  int err;
+  NDB_SOCKET_TYPE sock = m_extra_wakeup_sockets[0];
+  do
+  {
+    ret = recv(sock, buf, sizeof(buf), 0);
+    err = InetErrno;
+  } while (ret == sizeof(buf) || (ret == -1 && err == EINTR));
+}
+
 void
 TransporterRegistry::performSend()
 {
@@ -1025,6 +1288,17 @@ TransporterRegistry::report_connect(NodeId node_id)
   DBUG_ENTER("TransporterRegistry::report_connect");
   DBUG_PRINT("info",("performStates[%d]=CONNECTED",node_id));
   performStates[node_id] = CONNECTED;
+#if defined(HAVE_EPOLL_CREATE)
+  if (likely(m_epoll_fd != -1))
+  {
+    if (change_epoll((TCP_Transporter*)theTransporters[node_id],
+                     TRUE))
+    {
+      performStates[node_id] = DISCONNECTING;
+      DBUG_VOID_RETURN;
+    }
+  }
+#endif
   reportConnect(callbackObj, node_id);
   DBUG_VOID_RETURN;
 }
@@ -1035,6 +1309,9 @@ TransporterRegistry::report_disconnect(NodeId node_id, int errnum)
   DBUG_ENTER("TransporterRegistry::report_disconnect");
   DBUG_PRINT("info",("performStates[%d]=DISCONNECTED",node_id));
   performStates[node_id] = DISCONNECTED;
+#ifdef HAVE_EPOLL_CREATE
+  m_has_data_transporters.clear(node_id);
+#endif
   reportDisconnect(callbackObj, node_id, errnum);
   DBUG_VOID_RETURN;
 }
@@ -1093,8 +1370,24 @@ TransporterRegistry::start_clients_thread()
 	  /**
 	   * First, we try to connect (if we have a port number).
 	   */
+
 	  if (t->get_s_port())
-	    connected= t->connect_client();
+          {
+            // When ndbd is starting up, it won't allow
+            // ndbapi clients to connect until it's started
+            // The transporter will detect this case and
+            // limit rapid reconnect attempts
+#if 0
+            // disable this "fix" as it cause some 50 tests to fail
+	    // in autotest...
+	    // suggest that we send back "blocked" instead
+	    // and ping-pong this back-and-forth until connection
+	    // gets un-blocked
+            if (t->is_connect_blocked())
+              continue; // Too many refused connections
+#endif
+            connected= t->connect_client();
+          }
 
 	  /**
 	   * If dynamic, get the port for connecting from the management server
@@ -1129,19 +1422,20 @@ TransporterRegistry::start_clients_thread()
 	      }
 	      else if(ndb_mgm_is_connected(m_mgm_handle))
 	      {
-		g_eventLogger.info("Failed to get dynamic port to connect to: %d", res);
+                g_eventLogger->info("Failed to get dynamic port to connect to: %d", res);
 		ndb_mgm_disconnect(m_mgm_handle);
 	      }
 	      else
 	      {
-		g_eventLogger.info("Management server closed connection early. "
-			 "It is probably being shut down (or has problems). "
-			 "We will retry the connection. %d %s %s line: %d",
-                                   ndb_mgm_get_latest_error(m_mgm_handle),
-                                   ndb_mgm_get_latest_error_desc(m_mgm_handle),
-                                   ndb_mgm_get_latest_error_msg(m_mgm_handle),
-                                   ndb_mgm_get_latest_error_line(m_mgm_handle)
-                                   );
+                g_eventLogger->info
+                  ("Management server closed connection early. "
+                   "It is probably being shut down (or has problems). "
+                   "We will retry the connection. %d %s %s line: %d",
+                   ndb_mgm_get_latest_error(m_mgm_handle),
+                   ndb_mgm_get_latest_error_desc(m_mgm_handle),
+                   ndb_mgm_get_latest_error_msg(m_mgm_handle),
+                   ndb_mgm_get_latest_error_line(m_mgm_handle)
+                   );
 	      }
 	    }
 	    /** else
@@ -1159,6 +1453,17 @@ TransporterRegistry::start_clients_thread()
 	if(t->isConnected())
 	  t->doDisconnect();
 	break;
+      case DISCONNECTED:
+      {
+        if (t->isConnected())
+        {
+          g_eventLogger->warning("Found connection to %u in state DISCONNECTED "
+                                 " while being connected, disconnecting!",
+                                 t->getRemoteNodeId());
+          t->doDisconnect();
+        }
+        break;
+      }
       default:
 	break;
       }
@@ -1170,12 +1475,22 @@ TransporterRegistry::start_clients_thread()
 bool
 TransporterRegistry::start_clients()
 {
+  char thread_object[THREAD_CONTAINER_SIZE];
+  uint len;
+
   m_run_start_clients_thread= true;
-  m_start_clients_thread= NdbThread_Create(run_start_clients_C,
+  ndb_thread_fill_thread_object((void*)thread_object, &len, FALSE);
+  m_start_clients_thread= NdbThread_CreateWithFunc(run_start_clients_C,
 					   (void**)this,
-					   32768,
+                                           0, // default stack size
 					   "ndb_start_clients",
-					   NDB_THREAD_PRIO_LOW);
+					   NDB_THREAD_PRIO_LOW,
+                                           ndb_thread_add_thread_id,
+                                           thread_object,
+                                           len,
+                                           ndb_thread_remove_thread_id,
+                                           thread_object,
+                                           len);
   if (m_start_clients_thread == 0) {
     m_run_start_clients_thread= false;
     return false;
@@ -1235,7 +1550,7 @@ TransporterRegistry::start_service(SocketServer& socket_server)
   DBUG_ENTER("TransporterRegistry::start_service");
   if (m_transporter_interface.size() > 0 && !nodeIdSpecified)
   {
-    g_eventLogger.error("TransporterRegistry::startReceiving: localNodeId not specified");
+    g_eventLogger->error("TransporterRegistry::startReceiving: localNodeId not specified");
     DBUG_RETURN(false);
   }
 
@@ -1261,10 +1576,10 @@ TransporterRegistry::start_service(SocketServer& socket_server)
 	 * If it wasn't a dynamically allocated port, or
 	 * our attempts at getting a new dynamic port failed
 	 */
-	g_eventLogger.error("Unable to setup transporter service port: %s:%d!\n"
-		 "Please check if the port is already used,\n"
-		 "(perhaps the node is already running)",
-		 t.m_interface ? t.m_interface : "*", t.m_s_service_port);
+        g_eventLogger->error("Unable to setup transporter service port: %s:%d!\n"
+                             "Please check if the port is already used,\n"
+                             "(perhaps the node is already running)",
+                             t.m_interface ? t.m_interface : "*", t.m_s_service_port);
 	delete transporter_service;
 	DBUG_RETURN(false);
       }
@@ -1306,9 +1621,9 @@ TransporterRegistry::startReceiving()
     if(ret != 0)
     {
       DBUG_PRINT("error",("Install failed"));
-      g_eventLogger.error("Failed to install signal handler for"
-			  " SHM transporter, signum %d, errno: %d (%s)",
-			  g_ndb_shm_signum, errno, strerror(errno));
+      g_eventLogger->error("Failed to install signal handler for"
+                           " SHM transporter, signum %d, errno: %d (%s)",
+                           g_ndb_shm_signum, errno, strerror(errno));
     }
   }
 #endif // NDB_SHM_TRANSPORTER
@@ -1358,16 +1673,22 @@ bool TransporterRegistry::connect_client(NdbMgmHandle *h)
 
   if(!mgm_nodeid)
   {
-    g_eventLogger.error("%s: %d", __FILE__, __LINE__);
+    g_eventLogger->error("%s: %d", __FILE__, __LINE__);
     return false;
   }
   Transporter * t = theTransporters[mgm_nodeid];
   if (!t)
   {
-    g_eventLogger.error("%s: %d", __FILE__, __LINE__);
+    g_eventLogger->error("%s: %d", __FILE__, __LINE__);
     return false;
   }
-  DBUG_RETURN(t->connect_client(connect_ndb_mgmd(h)));
+
+  bool res = t->connect_client(connect_ndb_mgmd(h));
+  if (res == true)
+  {
+    performStates[mgm_nodeid] = TransporterRegistry::CONNECTING;
+  }
+  DBUG_RETURN(res);
 }
 
 /**
@@ -1380,7 +1701,7 @@ NDB_SOCKET_TYPE TransporterRegistry::connect_ndb_mgmd(NdbMgmHandle *h)
 
   if ( h==NULL || *h == NULL )
   {
-    g_eventLogger.error("%s: %d", __FILE__, __LINE__);
+    g_eventLogger->error("%s: %d", __FILE__, __LINE__);
     return NDB_INVALID_SOCKET;
   }
 
@@ -1393,10 +1714,10 @@ NDB_SOCKET_TYPE TransporterRegistry::connect_ndb_mgmd(NdbMgmHandle *h)
 				   m_transporter_interface[i].m_s_service_port,
 				   &mgm_reply) < 0)
     {
-      g_eventLogger.error("Error: %s: %d",
-	       ndb_mgm_get_latest_error_desc(*h),
-	       ndb_mgm_get_latest_error(*h));
-      g_eventLogger.error("%s: %d", __FILE__, __LINE__);
+      g_eventLogger->error("Error: %s: %d",
+                           ndb_mgm_get_latest_error_desc(*h),
+                           ndb_mgm_get_latest_error(*h));
+      g_eventLogger->error("%s: %d", __FILE__, __LINE__);
       ndb_mgm_destroy_handle(h);
       return NDB_INVALID_SOCKET;
     }
@@ -1408,10 +1729,10 @@ NDB_SOCKET_TYPE TransporterRegistry::connect_ndb_mgmd(NdbMgmHandle *h)
   NDB_SOCKET_TYPE sockfd= ndb_mgm_convert_to_transporter(h);
   if ( sockfd == NDB_INVALID_SOCKET)
   {
-    g_eventLogger.error("Error: %s: %d",
-	     ndb_mgm_get_latest_error_desc(*h),
-	     ndb_mgm_get_latest_error(*h));
-    g_eventLogger.error("%s: %d", __FILE__, __LINE__);
+    g_eventLogger->error("Error: %s: %d",
+                         ndb_mgm_get_latest_error_desc(*h),
+                         ndb_mgm_get_latest_error(*h));
+    g_eventLogger->error("%s: %d", __FILE__, __LINE__);
     ndb_mgm_destroy_handle(h);
   }
   return sockfd;
