@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef DBDICT_H
 #define DBDICT_H
@@ -41,6 +43,7 @@
 #include <signaldata/DropTable.hpp>
 #include <signaldata/AlterTable.hpp>
 #include <signaldata/AlterTab.hpp>
+#include <signaldata/ListTables.hpp>
 #include <signaldata/CreateIndx.hpp>
 #include <signaldata/DropIndx.hpp>
 #include <signaldata/AlterIndx.hpp>
@@ -60,6 +63,7 @@
 #include <signaldata/DictObjOp.hpp>
 #include <signaldata/DropFilegroupImpl.hpp>
 #include <SLList.hpp>
+#include <LockQueue.hpp>
 
 #ifdef DBDICT_C
 // Debug Macros
@@ -69,8 +73,7 @@
 /*--------------------------------------------------------------*/
 #define ZPACK_TABLE_INTO_PAGES 0
 #define ZSEND_GET_TAB_RESPONSE 3
-#define ZDICT_LOCK_POLL 4
-
+#define ZDROP_TAB_WAIT_GCI     4
 
 /*--------------------------------------------------------------*/
 // Other constants in alphabetical order
@@ -197,6 +200,7 @@ public:
   typedef Ptr<AttributeRecord> AttributeRecordPtr;
   ArrayPool<AttributeRecord> c_attributeRecordPool;
   DLHashTable<AttributeRecord> c_attributeRecordHash;
+  RSS_AP_SNAPSHOT(c_attributeRecordPool);
 
   /**
    * Shared table / index record.  Most of this is permanent data stored
@@ -383,6 +387,7 @@ public:
 
   typedef Ptr<TableRecord> TableRecordPtr;
   ArrayPool<TableRecord> c_tableRecordPool;
+  RSS_AP_SNAPSHOT(c_tableRecordPool);
 
   /**  Node Group and Tablespace id+version + range or list data.
     *  This is only stored temporarily in DBDICT during an ongoing
@@ -462,6 +467,7 @@ public:
   Uint32 c_maxNoOfTriggers;
   typedef Ptr<TriggerRecord> TriggerRecordPtr;
   ArrayPool<TriggerRecord> c_triggerRecordPool;
+  RSS_AP_SNAPSHOT(c_triggerRecordPool);
 
   /**
    * Information for each FS connection.
@@ -532,6 +538,7 @@ public:
 
   CArray<SchemaPageRecord> c_schemaPageRecordArray;
 
+  unsigned g_trace;
   DictTabInfo::Table c_tableDesc;
 
   /**
@@ -551,6 +558,8 @@ public:
     Uint64 m_file_size;
     Uint64 m_file_free;
     RopeHandle m_path;
+
+    Uint32 m_warningFlags; // move to op in 7.0
     
     Uint32 nextList;
     union {
@@ -590,6 +599,8 @@ public:
 	File_list::HeadPOD m_files;
       } m_logfilegroup;
     };
+
+    Uint32 m_warningFlags; // move to op in 7.0
     
     union {
       Uint32 nextPool;
@@ -611,6 +622,7 @@ public:
   Filegroup_hash c_filegroup_hash;
   
   RopePool c_rope_pool;
+  RSS_AP_SNAPSHOT(c_rope_pool);
 
   struct DictObject {
     DictObject() {}
@@ -642,6 +654,7 @@ public:
   
   DLHashTable<DictObject> c_obj_hash; // Name
   ArrayPool<DictObject> c_obj_pool;
+  RSS_AP_SNAPSHOT(c_obj_pool);
   
   DictObject * get_object(const char * name){
     return get_object(name, strlen(name) + 1);
@@ -694,6 +707,8 @@ private:
   void execFSREADCONF(Signal* signal);
   void execFSREADREF(Signal* signal);
   void execFSWRITECONF(Signal* signal);
+  void execFSREMOVEREF(Signal*);
+  void execFSREMOVECONF(Signal*);
   void execNDB_STTOR(Signal* signal);
   void execREAD_CONFIG_REQ(Signal* signal);
   void execSTTOR(Signal* signal);
@@ -1057,46 +1072,6 @@ private:
   /* ----------------------------------------------------------------------- */
   // State variables
   /* ----------------------------------------------------------------------- */
-  
-#ifndef ndb_dbdict_log_block_state
-  enum BlockState {
-    BS_IDLE = 0,
-    BS_CREATE_TAB = 1,
-    BS_BUSY = 2,
-    BS_NODE_FAILURE = 3,
-    BS_NODE_RESTART = 4
-  };
-#else // quick hack to log changes
-  enum {
-    BS_IDLE = 0,
-    BS_CREATE_TAB = 1,
-    BS_BUSY = 2,
-    BS_NODE_FAILURE = 3,
-    BS_NODE_RESTART = 4
-  };
-  struct BlockState;
-  friend struct BlockState;
-  struct BlockState {
-    BlockState() :
-      m_value(BS_IDLE) {
-    }
-    BlockState(int value) :
-      m_value(value) {
-    }
-    operator int() const {
-      return m_value;
-    }
-    BlockState& operator=(const BlockState& bs) {
-      Dbdict* dict = (Dbdict*)globalData.getBlock(DBDICT);
-      dict->infoEvent("DICT: bs %d->%d", m_value, bs.m_value);
-      globalSignalLoggers.log(DBDICT, "bs %d->%d", m_value, bs.m_value);
-      m_value = bs.m_value;
-      return *this;
-    }
-    int m_value;
-  };
-#endif
-  BlockState c_blockState;
 
   struct PackTable {
     
@@ -1119,6 +1094,7 @@ private:
    * Temporary structure used when parsing table info
    */
   struct ParseDictTabInfoRecord {
+    ParseDictTabInfoRecord() { tablePtr.setNull();}
     DictTabInfo::RequestType requestType;
     Uint32 errorCode;
     Uint32 errorLine;
@@ -2076,73 +2052,12 @@ private:
   // Unique key for operation  XXX move to some system table
   Uint32 c_opRecordSequence;
 
-  /*
-   * Master DICT can be locked in 2 mutually exclusive ways:
-   *
-   * 1) for schema ops, via operation records
-   * 2) against schema ops, via a lock queue
-   *
-   * Current use of 2) is by a starting node, to prevent schema ops
-   * until started.  The ops are refused (BlockState != BS_IDLE),
-   * not queued.
-   *
-   * Master failure is not handled, in node start case the starting
-   * node will crash too anyway.  Use lock table in future..
-   *
-   * The lock queue is "serial" but other behaviour is possible
-   * by checking lock types e.g. to allow parallel node starts.
-   *
-   * Checking release of last op record is not convenient with
-   * current structure (5.0).  Instead we poll via continueB.
-   *
-   * XXX only table ops check BlockState
-   */
-  struct DictLockType;
-  friend struct DictLockType;
-
-  struct DictLockType {
-    DictLockReq::LockType lockType;
-    BlockState blockState;
-    const char* text;
-  };
-
-  struct DictLockRecord;
-  friend struct DictLockRecord;
-
-  struct DictLockRecord {
-    DictLockReq req;
-    const DictLockType* lt;
-    bool locked;
-  union {
-    Uint32 nextPool;
-    Uint32 nextList;
-    };
-    Uint32 prevList;
-  };
-
-  typedef Ptr<DictLockRecord> DictLockPtr;
-  ArrayPool<DictLockRecord> c_dictLockPool;
-  DLFifoList<DictLockRecord> c_dictLockQueue;
-  bool c_dictLockPoll;
-
-  static const DictLockType* getDictLockType(Uint32 lockType);
-  void sendDictLockInfoEvent(Uint32 pollCount);
-  void sendDictLockInfoEvent(DictLockPtr lockPtr, const char* text);
-
-  // check if any schema op exists (conflicting with dict lock)
-  bool hasDictLockSchemaOp();
-
-  void checkDictLockQueue(Signal* signal, bool poll);
-  void sendDictLockConf(Signal* signal, DictLockPtr lockPtr);
-  void sendDictLockRef(Signal* signal, DictLockReq req, Uint32 errorCode);
-
-  // control polling i.e. continueB loop
-  void setDictLockPoll(Signal* signal, bool on, Uint32 pollCount);
-
-  // NF handling
-  void removeStaleDictLocks(Signal* signal, const Uint32* theFailedNodes);
-
-
+  void handleNdbdFailureCallback(Signal* signal, 
+                                 Uint32 failedNodeId,
+                                 Uint32 ignoredRc);
+  void handleApiFailureCallback(Signal* signal,
+                                Uint32 failedNodeId,
+                                Uint32 ignoredRc);
   // Statement blocks
 
   /* ------------------------------------------------------------ */
@@ -2395,10 +2310,12 @@ private:
   void dropEvent_sendReply(Signal* signal,
 			   OpDropEventPtr evntRecPtr);
 
-  void createEvent_RT_USER_CREATE(Signal* signal, OpCreateEventPtr evntRecPtr);
+  void createEvent_RT_USER_CREATE(Signal* signal,
+				  OpCreateEventPtr evntRecPtr,
+				  SectionHandle& handle);
   void createEventComplete_RT_USER_CREATE(Signal* signal,
 					  OpCreateEventPtr evntRecPtr);
-  void createEvent_RT_USER_GET(Signal* signal, OpCreateEventPtr evntRecPtr);
+  void createEvent_RT_USER_GET(Signal*, OpCreateEventPtr, SectionHandle&);
   void createEventComplete_RT_USER_GET(Signal* signal, OpCreateEventPtr evntRecPtr);
 
   void createEvent_RT_DICT_AFTER_GET(Signal* signal, OpCreateEventPtr evntRecPtr);
@@ -2437,6 +2354,7 @@ private:
 			  LinearSectionPtr dataPtr);
 
   void parseReadEventSys(Signal *signal, sysTab_NDBEVENTS_0& m_eventRec);
+  bool upgrade_suma_NotStarted(Uint32 err, Uint32 ref) const;
 
   // create trigger
   void createTrigger_recvReply(Signal* signal, const CreateTrigConf* conf,
@@ -2507,6 +2425,8 @@ private:
 
   void dropTable_backup_mutex_locked(Signal* signal, Uint32, Uint32);
   void dropTableRef(Signal * signal, DropTableReq *, DropTableRef::ErrorCode);
+  void dropTableWaitGci(Signal*);
+
   void printTables(); // For debugging only
   int handleAlterTab(AlterTabReq * req,
 		     CreateTableRecord * regAlterTabPtr,
@@ -2573,11 +2493,14 @@ private:
   void restartDropTab_complete(Signal*, Uint32 callback, Uint32);
 
   void restartDropObj(Signal*, Uint32, const SchemaFile::TableEntry *);
+  void restartDropObj_readObjConf(Signal*, Uint32, Uint32);
   void restartDropObj_prepare_start_done(Signal*, Uint32, Uint32);
   void restartDropObj_prepare_complete_done(Signal*, Uint32, Uint32);
   void restartDropObj_commit_start_done(Signal*, Uint32, Uint32);
   void restartDropObj_commit_complete_done(Signal*, Uint32, Uint32);
-  
+  void restartDropObj_updateSchemaFile(Signal*, DropObjRecordPtr);
+  void restartDropObj_fail(DropObjRecordPtr);
+
   void restart_checkSchemaStatusComplete(Signal*, Uint32 callback, Uint32);
   void restart_writeSchemaConf(Signal*, Uint32 callbackData, Uint32);
   void masterRestart_checkSchemaStatusComplete(Signal*, Uint32, Uint32);
@@ -2630,6 +2553,7 @@ private:
   void restartCreateObj_prepare_complete_done(Signal*, Uint32, Uint32);
   void restartCreateObj_commit_start_done(Signal*, Uint32, Uint32);
   void restartCreateObj_commit_complete_done(Signal*, Uint32, Uint32);
+  void restartCreateObj_fail(CreateObjRecordPtr);
 
   void execDICT_COMMIT_REQ(Signal*);
   void execDICT_COMMIT_REF(Signal*);
@@ -2673,6 +2597,39 @@ public:
   
   int checkSingleUserMode(Uint32 senderRef);
 
+  /**
+   * Dict lock queue does currently uniformly handle
+   *
+   * - starting node
+   * - schema op
+   *
+   * The impl. is based on DbUtil lock's (LockQueue)
+   *
+   * It would be very nice to use this *fully*
+   * But instead of introducing extra break in schema-op
+   *   a lock queue in instantiated in Dict, for easy trylock-handling
+   */
+  struct DictLockType;
+  friend struct DictLockType;
+  
+  struct DictLockType {
+    DictLockReq::LockType lockType;
+    const char* text;
+  };
+  static const DictLockType* getDictLockType(Uint32 lockType);
+  void sendDictLockInfoEvent(Signal*, const UtilLockReq*, const char* text);
+  void removeStaleDictLocks(Signal* signal, const Uint32* theFailedNodes);
+
+
+  Uint32 dict_lock_trylock(const DictLockReq* req);
+  Uint32 dict_lock_unlock(Signal* signal, const DictLockReq* req);
+  
+  LockQueue::Pool m_dict_lock_pool;
+  LockQueue m_dict_lock;
+
+  void sendOLD_LIST_TABLES_CONF(Signal *signal, ListTablesReq*);
+  void sendLIST_TABLES_CONF(Signal *signal, ListTablesReq*);
+
   Uint32 c_outstanding_sub_startstop;
   NdbNodeBitmask c_sub_startstop_lock;
 
