@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "Trix.hpp"
 
@@ -34,6 +36,17 @@
 
 #define CONSTRAINT_VIOLATION 893
 
+static
+bool
+check_timeout(Uint32 errCode)
+{
+  switch(errCode){
+  case 266:
+    return true;
+  }
+  return false;
+}
+
 #define DEBUG(x) { ndbout << "TRIX::" << x << endl; }
 
 /**
@@ -435,6 +448,27 @@ Trix::execDUMP_STATE_ORD(Signal* signal)
     // Ignore
   }
   }
+
+  if (signal->theData[0] == DumpStateOrd::SchemaResourceSnapshot)
+  {
+    RSS_AP_SNAPSHOT_SAVE(c_theSubscriptionRecPool);
+    return;
+  }
+
+  if (signal->theData[0] == DumpStateOrd::SchemaResourceCheckLeak)
+  {
+    RSS_AP_SNAPSHOT_CHECK(c_theSubscriptionRecPool);
+    return;
+  }
+  
+  if (signal->theData[0] == 8004)
+  {
+    infoEvent("TRIX: c_theSubscriptionRecPool size: %u free: %u",
+              c_theSubscriptionRecPool.getSize(),
+              c_theSubscriptionRecPool.getNoOfFree());
+    return;
+  }
+
 }
 
 // Build index
@@ -451,13 +485,14 @@ void Trix:: execBUILDINDXREQ(Signal* signal)
   // Seize a subscription record
   SubscriptionRecPtr subRecPtr;
   SubscriptionRecord* subRec;
-  
+  SectionHandle handle(this, signal);
+
   if (!c_theSubscriptions.seizeId(subRecPtr, buildIndxReq->getBuildId())) {
     // Failed to allocate subscription record
     BuildIndxRef * buildIndxRef = (BuildIndxRef *)signal->getDataPtrSend();
 
     buildIndxRef->setErrorCode(BuildIndxRef::AllocationFailure);
-    releaseSections(signal);
+    releaseSections(handle);
     sendSignal(buildIndxReq->getUserRef(), 
 	       GSN_BUILDINDXREF, signal, BuildIndxRef::SignalLength, JBB);
     DBUG_VOID_RETURN;
@@ -476,18 +511,23 @@ void Trix:: execBUILDINDXREQ(Signal* signal)
   subRec->subscriptionCreated = false;
   subRec->pendingSubSyncContinueConf = false;
   subRec->prepareId = RNIL;
+  subRec->m_flags = 0;
+  if (buildIndxReq->getRequestFlag() & BuildIndxReq::RF_NO_DISK)
+  {
+    subRec->m_flags |= SubscriptionRecord::RF_NO_DISK;
+  }
 
   // Get column order segments
-  Uint32 noOfSections = signal->getNoOfSections();
+  Uint32 noOfSections = handle.m_cnt;
   if(noOfSections > 0) {
     SegmentedSectionPtr ptr;
-    signal->getSection(ptr, BuildIndxReq::INDEX_COLUMNS);
+    handle.getSection(ptr, BuildIndxReq::INDEX_COLUMNS);
     append(subRec->attributeOrder, ptr, getSectionSegmentPool());
     subRec->noOfIndexColumns = ptr.sz;
   }
   if(noOfSections > 1) {
     SegmentedSectionPtr ptr;
-    signal->getSection(ptr, BuildIndxReq::KEY_COLUMNS);
+    handle.getSection(ptr, BuildIndxReq::KEY_COLUMNS);
     append(subRec->attributeOrder, ptr, getSectionSegmentPool());
     subRec->noOfKeyColumns = ptr.sz;
   }
@@ -496,7 +536,7 @@ void Trix:: execBUILDINDXREQ(Signal* signal)
   printf("Trix:: execBUILDINDXREQ: Attribute order:\n");
   subRec->attributeOrder.print(stdout);
 #endif
-  releaseSections(signal);
+  releaseSections(handle);
   prepareInsertTransactions(signal, subRecPtr);
   DBUG_VOID_RETURN;
 }
@@ -541,7 +581,11 @@ void Trix::execUTIL_PREPARE_REF(Signal* signal)
     return;
   }
   subRecPtr.p = subRec;
-  subRec->errorCode = BuildIndxRef::InternalError;
+  subRec->errorCode = (BuildIndxRef::ErrorCode)utilPrepareRef->errorCode;
+
+  UtilReleaseConf* conf = (UtilReleaseConf*)signal->getDataPtrSend();
+  conf->senderData = subRecPtr.i;
+  execUTIL_RELEASE_CONF(signal);
 }
 
 void Trix::execUTIL_EXECUTE_CONF(Signal* signal)
@@ -578,9 +622,20 @@ void Trix::execUTIL_EXECUTE_REF(Signal* signal)
   subRecPtr.p = subRec;
   ndbrequire(utilExecuteRef->errorCode == UtilExecuteRef::TCError);
   if(utilExecuteRef->TCErrorCode == CONSTRAINT_VIOLATION)
+  {
+    jam();
     buildFailed(signal, subRecPtr, BuildIndxRef::IndexNotUnique);
+  }
+  else if (check_timeout(utilExecuteRef->TCErrorCode))
+  {
+    jam();
+    buildFailed(signal, subRecPtr, BuildIndxRef::DeadlockError);
+  }
   else
+  {
+    jam();
     buildFailed(signal, subRecPtr, BuildIndxRef::InternalError);
+  }
 }
 
 void Trix::execSUB_CREATE_CONF(Signal* signal)
@@ -611,20 +666,27 @@ void Trix::execSUB_CREATE_REF(Signal* signal)
 {
   jamEntry();
   DBUG_ENTER("Trix::execSUB_CREATE_REF");
-  // THIS SIGNAL IS NEVER SENT FROM SUMA?
-  /*
+
   SubCreateRef * subCreateRef = (SubCreateRef *)signal->getDataPtr();
   SubscriptionRecPtr subRecPtr;
   SubscriptionRecord* subRec;
 
-  subRecPtr.i = subCreateRef->subscriberData;
-  if ((subRec = c_theSubscriptions.getPtr(subRecPtr.i)) == NULL) {
+  subRecPtr.i = subCreateRef->senderData;
+  if ((subRec = c_theSubscriptions.getPtr(subRecPtr.i)) == NULL)
+  {
     printf("Trix::execSUB_CREATE_REF: Failed to find subscription data %u\n", subRecPtr.i);
     return;
   }
   subRecPtr.p = subRec;
-  buildFailed(signal, subRecPtr, BuildIndxRef::InternalError);
-  */
+  subRecPtr.p->errorCode = (BuildIndxRef::ErrorCode)subCreateRef->errorCode;
+
+  UtilReleaseReq * const req = (UtilReleaseReq*)signal->getDataPtrSend();
+  req->prepareId = subRecPtr.p->prepareId;
+  req->senderData = subRecPtr.i;
+
+  sendSignal(DBUTIL_REF, GSN_UTIL_RELEASE_REQ, signal,
+	     UtilReleaseReq::SignalLength, JBB);
+
   DBUG_VOID_RETURN;
 }
 
@@ -700,14 +762,14 @@ void Trix::execSUB_TABLE_DATA(Signal* signal)
     DBUG_VOID_RETURN;
   }
   subRecPtr.p = subRec;
+  SectionHandle handle(this, signal);
   SegmentedSectionPtr headerPtr, dataPtr;
-  if (!signal->getSection(headerPtr, 0)) {
-    printf("Trix::execSUB_TABLE_DATA: Failed to get header section\n");
-  }
-  if (!signal->getSection(dataPtr, 1)) {
-    printf("Trix::execSUB_TABLE_DATA: Failed to get data section\n");
-  }
+
+  handle.getSection(headerPtr, 0);
+  handle.getSection(dataPtr, 1);
   executeInsertTransaction(signal, subRecPtr, headerPtr, dataPtr);
+
+  releaseSections(handle);
   DBUG_VOID_RETURN;
 }
 
@@ -731,6 +793,7 @@ void Trix::setupSubscription(Signal* signal, SubscriptionRecPtr subRecPtr)
 
   sendSignal(SUMA_REF, GSN_SUB_CREATE_REQ, 
 	     signal, SubCreateReq::SignalLength, JBB);
+
   DBUG_VOID_RETURN;
 }
 
@@ -759,6 +822,12 @@ void Trix::startTableScan(Signal* signal, SubscriptionRecPtr subRecPtr)
   subSyncReq->subscriptionId = subRec->subscriptionId;
   subSyncReq->subscriptionKey = subRec->subscriptionKey;
   subSyncReq->part = SubscriptionData::TableData;
+  subSyncReq->requestInfo = 0;
+  if (subRecPtr.p->m_flags & SubscriptionRecord::RF_NO_DISK)
+  {
+    jam();
+    subSyncReq->requestInfo |= SubSyncReq::NoDisk;
+  }
 
   subRecPtr.p->expectedConf = 1;
 
