@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <my_pthread.h>
@@ -44,7 +46,7 @@
 #include <sys/processor.h> // For system informatio
 #endif
 
-extern EventLogger g_eventLogger;
+extern EventLogger * g_eventLogger;
 extern NdbMutex * theShutdownMutex;
 
 void catchsigs(bool ignore); // for process signal handling
@@ -90,7 +92,13 @@ void childExit(int code, Uint32 currentStartPhase)
   fprintf(child_info_file_w, "\n");
   fclose(child_info_file_r);
   fclose(child_info_file_w);
+
+// gcov will not produce results when using _exit
+#ifdef HAVE_gcov
   exit(code);
+#else
+  _exit(code);
+#endif
 }
 
 void childAbort(int code, Uint32 currentStartPhase)
@@ -152,7 +160,7 @@ int reportShutdown(class Configuration *config, int error_exit, int restart)
   get_int_property(info, "sphase", &sphase);
 
   Uint32 length, theData[25];
-  EventReport *rep = (EventReport *)theData;
+  EventReport *rep = CAST_PTR(EventReport, &theData[0]);
 
   rep->setNodeId(globalData.ownId);
   if (restart)
@@ -180,8 +188,8 @@ int reportShutdown(class Configuration *config, int error_exit, int restart)
 
   { // Log event
     const EventReport * const eventReport = (EventReport *)&theData[0];
-    g_eventLogger.log(eventReport->getEventType(), theData, length,
-		      eventReport->getNodeId(), 0);
+    g_eventLogger->log(eventReport->getEventType(), theData, length,
+                       eventReport->getNodeId(), 0);
   }
 
   for (unsigned n = 0; n < config->m_mgmds.size(); n++)
@@ -208,13 +216,13 @@ handle_error:
       BaseString tmp(ndb_mgm_get_latest_error_msg(h));
       tmp.append(" : ");
       tmp.append(ndb_mgm_get_latest_error_desc(h));
-      g_eventLogger.warning("Unable to report shutdown reason to %s: %s",
-			    config->m_mgmds[n].c_str(), tmp.c_str());
+      g_eventLogger->warning("Unable to report shutdown reason to %s: %s",
+                             config->m_mgmds[n].c_str(), tmp.c_str());
     }
     else
     {
-      g_eventLogger.error("Unable to report shutdown reason to %s",
-			  config->m_mgmds[n].c_str());
+      g_eventLogger->error("Unable to report shutdown reason to %s",
+                           config->m_mgmds[n].c_str());
     }
 do_next:
     if (h)
@@ -226,18 +234,52 @@ do_next:
   return 0;
 }
 
+
+/*
+  retry failed fork after sleep until fork suceeds or
+  max number of retries occurs
+*/
+
+static pid_t
+retry_fork(void)
+{
+  const unsigned max_retries = 10;
+  unsigned retry_counter = 0;
+  while(true)
+  {
+    pid_t pid = fork();
+    if (pid == -1)
+    {
+      g_eventLogger->warning("Angel failed to fork, errno: %d", errno);
+
+      if (retry_counter++ == max_retries)
+      {
+        g_eventLogger->error("Angel failed to fork %d times, giving up",
+                             retry_counter);
+        exit(-1);
+      }
+      NdbSleep_SecSleep(1);
+      continue;
+    }
+    return pid;
+  }
+}
+
+extern int g_ndb_init_need_monotonic;
+
 int main(int argc, char** argv)
 {
+  g_ndb_init_need_monotonic = 1;
   NDB_INIT(argv[0]);
   // Print to stdout/console
-  g_eventLogger.createConsoleHandler();
-  g_eventLogger.setCategory("ndbd");
-  g_eventLogger.enable(Logger::LL_ON, Logger::LL_INFO);
-  g_eventLogger.enable(Logger::LL_ON, Logger::LL_CRITICAL);
-  g_eventLogger.enable(Logger::LL_ON, Logger::LL_ERROR);
-  g_eventLogger.enable(Logger::LL_ON, Logger::LL_WARNING);
+  g_eventLogger->createConsoleHandler();
+  g_eventLogger->setCategory("ndbd");
+  g_eventLogger->enable(Logger::LL_ON, Logger::LL_INFO);
+  g_eventLogger->enable(Logger::LL_ON, Logger::LL_CRITICAL);
+  g_eventLogger->enable(Logger::LL_ON, Logger::LL_ERROR);
+  g_eventLogger->enable(Logger::LL_ON, Logger::LL_WARNING);
 
-  g_eventLogger.m_logLevel.setLogLevel(LogLevel::llStartUp, 15);
+  g_eventLogger->m_logLevel.setLogLevel(LogLevel::llStartUp, 15);
 
   globalEmulatorData.create();
 
@@ -278,26 +320,28 @@ int main(int argc, char** argv)
     int filedes[2];
     if (pipe(filedes))
     {
-      g_eventLogger.error("pipe() failed with errno=%d (%s)",
-			  errno, strerror(errno));
+      g_eventLogger->error("pipe() failed with errno=%d (%s)",
+                           errno, strerror(errno));
       return 1;
     }
     else
     {
       if (!(child_info_file_w= fdopen(filedes[1],"w")))
       {
-	g_eventLogger.error("fdopen() failed with errno=%d (%s)",
-			    errno, strerror(errno));
+        g_eventLogger->error("fdopen() failed with errno=%d (%s)",
+                             errno, strerror(errno));
       }
       if (!(child_info_file_r= fdopen(filedes[0],"r")))
       {
-	g_eventLogger.error("fdopen() failed with errno=%d (%s)",
-			    errno, strerror(errno));
+        g_eventLogger->error("fdopen() failed with errno=%d (%s)",
+                             errno, strerror(errno));
       }
     }
 
-    if ((child = fork()) <= 0)
-      break; // child or error
+    child = retry_fork();
+    if (child == 0)
+      break; // child
+    assert(child > 0);
 
     /**
      * Parent
@@ -319,7 +363,7 @@ int main(int argc, char** argv)
     if(WIFEXITED(status)){
       switch(WEXITSTATUS(status)){
       case NRT_Default:
-	g_eventLogger.info("Angel shutting down");
+        g_eventLogger->info("Angel shutting down");
 	reportShutdown(theConfig, 0, 0);
 	exit(0);
 	break;
@@ -360,7 +404,7 @@ int main(int argc, char** argv)
       else
       {
 	signum = 127;
-	g_eventLogger.info("Unknown exit reason. Stopped.");
+        g_eventLogger->info("Unknown exit reason. Stopped.");
       }
       if(theConfig->stopOnError()){
 	/**
@@ -371,35 +415,59 @@ int main(int argc, char** argv)
       }
     }
 
+    Uint32 restartDelaySecs = 0;
+
     if (!failed_startup_flag)
     {
       // Reset the counter for consecutive failed startups
       failed_startups = 0;
     }
-    else if (failed_startups >= MAX_FAILED_STARTUPS && !theConfig->stopOnError())
+    else if (!theConfig->stopOnError())
     {
-      /**
-       * Error shutdown && stopOnError()
+      if (theConfig->getMaxStartFailRetries() &&     // 0 == infinite
+          failed_startups >= theConfig->getMaxStartFailRetries())
+      {
+        /**
+         * Error shutdown && stopOnError()
+         */
+        g_eventLogger->alert("Ndbd has failed %u consecutive startups. "
+                             "Not restarting", failed_startups);
+        reportShutdown(theConfig, error_exit, 0);
+        exit(0);
+      }
+      
+      g_eventLogger->info("Ndbd has failed %u consecutive startup attempts.  "
+                          "Will attempt restart again.", failed_startups);
+      
+      /* Failed in Startup, StopOnError == false and not too many
+       * retry attempts yet.  Determine from config whether we 
+       * should wait at all before restarting.
        */
-      g_eventLogger.alert("Ndbd has failed %u consecutive startups. "
-			  "Not restarting", failed_startups);
-      reportShutdown(theConfig, error_exit, 0);
-      exit(0);
+      restartDelaySecs = theConfig->getStartFailRetryDelaySecs();
     }
+
     failed_startup_flag = false;
     reportShutdown(theConfig, error_exit, 1);
-    g_eventLogger.info("Ndb has terminated (pid %d) restarting", child);
+    
+    if (restartDelaySecs)
+    {
+      g_eventLogger->info("Delaying Ndb restart for %u seconds.", 
+                          restartDelaySecs);
+      NdbSleep_SecSleep(restartDelaySecs);
+    }
+    
+    g_eventLogger->info("Ndb has terminated (pid %d) restarting", child);
     theConfig->fetch_configuration();
   }
 
   if (child >= 0)
-    g_eventLogger.info("Angel pid: %d ndb pid: %d", getppid(), getpid());
+    g_eventLogger->info("Angel pid: %d ndb pid: %d", getppid(), getpid());
   else if (child > 0)
-    g_eventLogger.info("Ndb pid: %d", getpid());
+    g_eventLogger->info("Ndb pid: %d", getpid());
   else
-    g_eventLogger.info("Ndb started in foreground");
+    g_eventLogger->info("Ndb started in foreground");
 #else
-  g_eventLogger.info("Ndb started");
+  g_eventLogger->info("Ndb started");
 #endif
   theConfig->setupConfiguration();
   systemInfo(* theConfig, * theConfig->m_logLevel); 
@@ -469,9 +537,11 @@ int main(int argc, char** argv)
   globalEmulatorData.m_socket_server->startServer();
 
   //  theConfig->closeConfiguration();
-
-  globalEmulatorData.theThreadConfig->ipControlLoop();
-  
+  {
+    Uint32 inx = globalEmulatorData.theConfiguration->addThreadId(MainThread);
+    globalEmulatorData.theThreadConfig->ipControlLoop(inx);
+    globalEmulatorData.theConfiguration->removeThreadId(inx);
+  }
   NdbShutdown(NST_Normal);
 
   return NRT_Default;
@@ -511,19 +581,19 @@ systemInfo(const Configuration & config, const LogLevel & logLevel){
 #endif
   
   if(logLevel.getLogLevel(LogLevel::llStartUp) > 0){
-    g_eventLogger.info("NDB Cluster -- DB node %d", globalData.ownId);
-    g_eventLogger.info("%s --", NDB_VERSION_STRING);
+    g_eventLogger->info("NDB Cluster -- DB node %d", globalData.ownId);
+    g_eventLogger->info("%s --", NDB_VERSION_STRING);
     if (config.get_mgmd_host())
-      g_eventLogger.info("Configuration fetched at %s port %d",
-			 config.get_mgmd_host(), config.get_mgmd_port());
+      g_eventLogger->info("Configuration fetched at %s port %d",
+                          config.get_mgmd_host(), config.get_mgmd_port());
 #ifdef NDB_SOLARIS // ok
-    g_eventLogger.info("NDB is running on a machine with %d processor(s) at %d MHz",
-		       processor, speed);
+    g_eventLogger->info("NDB is running on a machine with %d processor(s) at %d MHz",
+                        processor, speed);
 #endif
   }
   if(logLevel.getLogLevel(LogLevel::llStartUp) > 3){
     Uint32 t = config.timeBetweenWatchDogCheck();
-    g_eventLogger.info("WatchDog timer is set to %d ms", t);
+    g_eventLogger->info("WatchDog timer is set to %d ms", t);
   }
 
 }
@@ -601,7 +671,7 @@ catchsigs(bool ignore){
 extern "C"
 void 
 handler_shutdown(int signum){
-  g_eventLogger.info("Received signal %d. Performing stop.", signum);
+  g_eventLogger->info("Received signal %d. Performing stop.", signum);
   childReportError(0);
   childReportSignal(signum);
   globalData.theRestartFlag = perform_stop;
@@ -611,9 +681,11 @@ extern "C"
 void 
 handler_error(int signum){
   // only let one thread run shutdown
-  static long thread_id= 0;
+  static bool handling_error = false;
+  static pthread_t thread_id; // Valid when handling_error is true
 
-  if (thread_id != 0 && thread_id == my_thread_id())
+  if (handling_error &&
+      pthread_equal(thread_id, pthread_self()))
   {
     // Shutdown thread received signal
 #ifndef NDB_WIN32
@@ -626,8 +698,11 @@ handler_error(int signum){
   if(theShutdownMutex && NdbMutex_Trylock(theShutdownMutex) != 0)
     while(true)
       NdbSleep_MilliSleep(10);
-  thread_id= my_thread_id();
-  g_eventLogger.info("Received signal %d. Running error handler.", signum);
+
+  thread_id = pthread_self();
+  handling_error = true;
+
+  g_eventLogger->info("Received signal %d. Running error handler.", signum);
   childReportSignal(signum);
   // restart the system
   char errorData[64], *info= 0;
@@ -648,5 +723,5 @@ handler_sigusr1(int signum)
     failed_startups++;
     failed_startup_flag = true;
   }
-  g_eventLogger.info("Angel received ndbd startup failure count %u.", failed_startups);
+  g_eventLogger->info("Angel received ndbd startup failure count %u.", failed_startups);
 }
