@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef NdbDictionaryImpl_H
 #define NdbDictionaryImpl_H
@@ -31,6 +33,8 @@
 #include "NdbWaiter.hpp"
 #include "DictCache.hpp"
 
+class ListTablesReq;
+
 bool
 is_ndb_blob_table(const char* name, Uint32* ptab_id = 0, Uint32* pcol_no = 0);
 bool
@@ -91,6 +95,12 @@ public:
   bool m_autoIncrement;
   Uint64 m_autoIncrementInitialValue;
   BaseString m_defaultValue;
+  /*
+   * Table holding the blob parts.
+   *
+   * Note that if getPartSize() is 0, there is no parts table, so
+   * m_blobTable==NULL.
+   */
   NdbTableImpl * m_blobTable;
 
   /**
@@ -104,6 +114,9 @@ public:
                                 // for blob, storage type of NDB$DATA
   bool m_dynamic;
 
+  bool m_indexSourced;          // Computed when column defined, or when
+                                // table schema read.  Not stored in kernel 
+
   /*
    * NdbTableImpl: if m_pk, 0-based index of key in m_attrId order
    * NdbIndexImpl: m_column_no of primary table column
@@ -128,10 +141,17 @@ public:
   static const NdbColumnImpl & getImpl(const NdbDictionary::Column & t);
   NdbDictionary::Column * m_facade;
 
+  static void create_pseudo_columns();
+  static void destory_pseudo_columns();
   static NdbDictionary::Column * create_pseudo(const char *);
 
   // Get total length in bytes, used by NdbOperation
   bool get_var_length(const void* value, Uint32& len) const;
+
+  // Get significant and send length in bytes, accounting for bug39645 
+  bool get_var_length_bug39645(const void* value, 
+                               Uint32& sigLen, 
+                               Uint32& sendLen) const;
 };
 
 class NdbTableImpl : public NdbDictionary::Table, public NdbDictObjectImpl {
@@ -161,9 +181,15 @@ public:
   const void * getRangeListData() const;
   Uint32 getRangeListDataLen() const;
 
+  Uint32 getFragmentNodes(Uint32 fragmentId, 
+                          Uint32* nodeIdArrayPtr,
+                          Uint32 arraySize) const;
+
   const char * getMysqlName() const;
   int updateMysqlName();
 
+  bool matchDb(const char * name, size_t len) const;
+
   int aggregate(NdbError& error);
   int validate(NdbError& error);
 
@@ -190,6 +216,7 @@ public:
     if iterating over columns.
   */
   Vector<NdbColumnImpl *> m_columns;
+  Uint32 m_noOfAutoIncColumns;
   void computeAggregates();
   int buildColumnHash(); 
 
@@ -238,6 +265,18 @@ public:
   Uint8 m_noOfDiskColumns;
   Uint8 m_replicaCount;
 
+  /**
+   * Default NdbRecord for this table or index
+   * Currently used by old-Api scans to use NdbRecord API internally.
+   */
+  NdbRecord *m_ndbrecord;
+
+  /**
+   * Bitmask describing PK column positions for this table
+   * Currently used by old-Api scans to use NdbRecord API internally.
+   */
+  const unsigned char * m_pkMask;
+
   /**
    * Equality/assign
    */
@@ -297,6 +336,66 @@ public:
   NdbDictionary::Index * m_facade;
 };
 
+class NdbOptimizeTableHandleImpl : public NdbDictionary::OptimizeTableHandle
+{
+  enum State { CREATED, INITIALIZED, FINISHED, ABORTED, CLOSED };
+private:
+  int start();
+
+  State m_state;
+  Ndb *m_ndb;
+  struct fifo_element_st {
+    fifo_element_st(const NdbTableImpl *tab, fifo_element_st *prev)
+      { 
+        table = tab; 
+        previous = prev; 
+        next = NULL;
+        if (prev) 
+          prev->next = this; 
+      } 
+    const NdbTableImpl * table;
+    fifo_element_st * previous;
+    fifo_element_st * next;
+  };
+  const NdbTableImpl * m_table;
+  fifo_element_st * m_table_queue;
+  fifo_element_st * m_table_queue_first;
+  fifo_element_st * m_table_queue_end;
+  NdbTransaction * m_trans;
+  NdbScanOperation * m_scan_op;
+
+  NdbDictionary::OptimizeTableHandle * m_facade;
+public:
+  NdbOptimizeTableHandleImpl(NdbDictionary::OptimizeTableHandle &);
+  ~NdbOptimizeTableHandleImpl();
+  static NdbOptimizeTableHandleImpl & getImpl(NdbDictionary::OptimizeTableHandle &h)
+    { return h.m_impl; }
+
+  int init(Ndb *ndb, const NdbTableImpl &table);
+  int next();
+  int close();
+};
+
+class NdbOptimizeIndexHandleImpl : public NdbDictionary::OptimizeIndexHandle
+{
+  enum State { CREATED, INITIALIZED, FINISHED, ABORTED, CLOSED };
+private:
+  State m_state;
+  Ndb *m_ndb;
+  const NdbIndexImpl *m_index;
+  class NdbDictionary::OptimizeTableHandle m_optimize_table_handle;
+  NdbDictionary::OptimizeIndexHandle * m_facade;
+public:
+  NdbOptimizeIndexHandleImpl(NdbDictionary::OptimizeIndexHandle &);
+  ~NdbOptimizeIndexHandleImpl();
+  static NdbOptimizeIndexHandleImpl & getImpl(NdbDictionary::OptimizeIndexHandle &h)
+    { return h.m_impl; }
+
+  int init(Ndb *ndb, const NdbIndexImpl &index);
+  int next();
+  int close();
+};
+
 class NdbEventImpl : public NdbDictionary::Event, public NdbDictObjectImpl {
   friend class NdbDictInterface;
   friend class NdbDictionaryImpl;
@@ -440,7 +539,8 @@ public:
 
 class NdbDictInterface {
 public:
-  NdbDictInterface(NdbError& err) : m_error(err) {
+  NdbDictInterface(NdbError& err, int& warn) :
+      m_error(err), m_warn(warn) {
     m_reference = 0;
     m_masterNodeId = 0;
     m_transporter= NULL;
@@ -449,6 +549,7 @@ public:
   
   bool setTransporter(class Ndb * ndb, class TransporterFacade * tf);
   bool setTransporter(class TransporterFacade * tf);
+  class TransporterFacade *getTransporter();
   
   // To abstract the stuff thats made in all create/drop/lists below
   int dictSignal(NdbApiSignal* signal, LinearSectionPtr ptr[3], int secs,
@@ -475,7 +576,8 @@ public:
 
   int dropTable(const NdbTableImpl &);
 
-  int createIndex(class Ndb & ndb, const NdbIndexImpl &, const NdbTableImpl &);
+  int createIndex(class Ndb & ndb, const NdbIndexImpl &, const NdbTableImpl &,
+                  bool offline);
   int dropIndex(const NdbIndexImpl &, const NdbTableImpl &);
   
   int createEvent(class Ndb & ndb, NdbEventImpl &, int getFlag);
@@ -485,8 +587,14 @@ public:
   int executeSubscribeEvent(class Ndb & ndb, NdbEventOperationImpl &);
   int stopSubscribeEvent(class Ndb & ndb, NdbEventOperationImpl &);
   
-  int listObjects(NdbDictionary::Dictionary::List& list, Uint32 requestData, bool fullyQualifiedNames);
-  int listObjects(NdbApiSignal* signal);
+  int listObjects(NdbDictionary::Dictionary::List& list,
+                  ListTablesReq& ltreq, bool fullyQualifiedNames);
+  int listObjects(NdbApiSignal* signal, bool& listTablesLongSignal);
+
+  int unpackListTables(NdbDictionary::Dictionary::List& list,
+                       bool fullyQualifiedNames);
+  int unpackOldListTables(NdbDictionary::Dictionary::List& list,
+                          bool fullyQualifiedNames);
   
   NdbTableImpl * getTable(int tableId, bool fullyQualifiedNames);
   NdbTableImpl * getTable(const BaseString& name, bool fullyQualifiedNames);
@@ -494,7 +602,7 @@ public:
 			  LinearSectionPtr ptr[3],
 			  Uint32 noOfSections, bool fullyQualifiedNames);
 
-  int forceGCPWait();
+  int forceGCPWait(int type);
 
   static int parseTableInfo(NdbTableImpl ** dst, 
 			    const Uint32 * data, Uint32 len,
@@ -524,6 +632,7 @@ public:
   
   const NdbError &getNdbError() const;  
   NdbError & m_error;
+  int & m_warn;
 private:
   Uint32 m_reference;
   Uint32 m_masterNodeId;
@@ -537,8 +646,7 @@ private:
 			 class NdbApiSignal* signal, 
 			 struct LinearSectionPtr ptr[3]);
   
-  static void execNodeStatus(void* dictImpl, Uint32, 
-			     bool alive, bool nfCompleted);  
+  static void execNodeStatus(void* dictImpl, Uint32, Uint32);
   
   void execGET_TABINFO_REF(NdbApiSignal *, LinearSectionPtr ptr[3]);
   void execGET_TABINFO_CONF(NdbApiSignal *, LinearSectionPtr ptr[3]);
@@ -563,6 +671,7 @@ private:
 
   void execDROP_TABLE_REF(NdbApiSignal *, LinearSectionPtr ptr[3]);
   void execDROP_TABLE_CONF(NdbApiSignal *, LinearSectionPtr ptr[3]);
+  void execOLD_LIST_TABLES_CONF(NdbApiSignal *, LinearSectionPtr ptr[3]);
   void execLIST_TABLES_CONF(NdbApiSignal *, LinearSectionPtr ptr[3]);
 
   void execCREATE_FILE_REF(NdbApiSignal *, LinearSectionPtr ptr[3]);
@@ -582,21 +691,22 @@ private:
 
   Uint32 m_fragmentId;
   UtilBuffer m_buffer;
+
+  Uint32 m_noOfTables;
+  UtilBuffer m_tableData;
+  UtilBuffer m_tableNames;
 };
 
 class NdbDictionaryImpl;
 class GlobalCacheInitObject
 {
 public:
-  NdbDictionaryImpl *m_dict;
   const BaseString &m_name;
-  GlobalCacheInitObject(NdbDictionaryImpl *dict,
-                        const BaseString &name) :
-    m_dict(dict),
+  GlobalCacheInitObject(const BaseString &name) :
     m_name(name)
   {}
   virtual ~GlobalCacheInitObject() {}
-  virtual int init(NdbTableImpl &tab) const = 0;
+  virtual int init(NdbDictionaryImpl *dict, NdbTableImpl &tab) const = 0;
 };
 
 class NdbDictionaryImpl : public NdbDictionary::Dictionary {
@@ -609,6 +719,10 @@ public:
   bool setTransporter(class TransporterFacade * tf);
 
   int createTable(NdbTableImpl &t);
+  int optimizeTable(const NdbTableImpl &t,
+                    NdbOptimizeTableHandleImpl &h);
+  int optimizeIndex(const NdbIndexImpl &index,
+                    NdbOptimizeIndexHandleImpl &h);
   int createBlobTables(const NdbTableImpl& t);
   bool supportedAlterTable(NdbTableImpl &old_impl, NdbTableImpl &impl);
   int alterTable(NdbTableImpl &old_impl, NdbTableImpl &impl);
@@ -618,8 +732,8 @@ public:
   int invalidateObject(NdbTableImpl &);
   int removeCachedObject(NdbTableImpl &);
 
-  int createIndex(NdbIndexImpl &ix);
-  int createIndex(NdbIndexImpl &ix, NdbTableImpl & tab);
+  int createIndex(NdbIndexImpl &ix, bool offline);
+  int createIndex(NdbIndexImpl &ix, NdbTableImpl & tab, bool offline);
   int dropIndex(const char * indexName, 
 		const char * tableName);
   int dropIndex(NdbIndexImpl &, const char * tableName);
@@ -636,19 +750,22 @@ public:
   int executeSubscribeEvent(NdbEventOperationImpl &);
   int stopSubscribeEvent(NdbEventOperationImpl &);
 
-  int forceGCPWait();
+  int forceGCPWait(int type);
 
-  int listObjects(List& list, NdbDictionary::Object::Type type);
+  int listObjects(List& list, NdbDictionary::Object::Type type, 
+                  bool fullyQualified);
   int listIndexes(List& list, Uint32 indexId);
 
   NdbTableImpl * getTableGlobal(const char * tableName);
   NdbIndexImpl * getIndexGlobal(const char * indexName,
                                 NdbTableImpl &ndbtab);
+  NdbIndexImpl * getIndexGlobal(const char * indexName,
+                                const char * tableName);
   int alterTableGlobal(NdbTableImpl &orig_impl, NdbTableImpl &impl);
   int dropTableGlobal(NdbTableImpl &);
   int dropIndexGlobal(NdbIndexImpl & impl);
-  int releaseTableGlobal(NdbTableImpl & impl, int invalidate);
-  int releaseIndexGlobal(NdbIndexImpl & impl, int invalidate);
+  int releaseTableGlobal(const NdbTableImpl & impl, int invalidate);
+  int releaseIndexGlobal(const NdbIndexImpl & impl, int invalidate);
 
   NdbTableImpl * getTable(const char * tableName, void **data= 0);
   NdbTableImpl * getBlobTable(const NdbTableImpl&, uint col_no);
@@ -677,6 +794,7 @@ public:
   
   const NdbError & getNdbError() const;
   NdbError m_error;
+  int m_warn;
   Uint32 m_local_table_data_size;
 
   LocalDictCache m_localHash;
@@ -685,6 +803,11 @@ public:
   static NdbDictionaryImpl & getImpl(NdbDictionary::Dictionary & t);
   static const NdbDictionaryImpl & getImpl(const NdbDictionary::Dictionary &t);
   NdbDictionary::Dictionary * m_facade;
+  int initialiseColumnData(bool isIndex,
+                           Uint32 flags,
+                           const NdbDictionary::RecordSpecification *recSpec,
+                           Uint32 colNum,
+                           NdbRecord *rec);
 
   NdbDictInterface m_receiver;
   Ndb & m_ndb;
@@ -697,20 +820,53 @@ public:
                               const BaseString& internalName);
 
 
+  int createDefaultNdbRecord(NdbTableImpl* tableOrIndex,
+                             const NdbTableImpl* baseTableForIndex);
+
+  bool validateRecordSpec(const NdbDictionary::RecordSpecification *recSpec,
+                          Uint32 length,
+                          Uint32 flags);
+
   NdbRecord *createRecord(const NdbTableImpl *table,
                           const NdbDictionary::RecordSpecification *recSpec,
                           Uint32 length,
                           Uint32 elemSize,
                           Uint32 flags,
-                          const NdbTableImpl *base_table= 0);
-  NdbRecord *createRecord(const NdbIndexImpl *index,
-                          const NdbTableImpl *base_table,
-                          const NdbDictionary::RecordSpecification *recSpec,
-                          Uint32 length,
-                          Uint32 elemSize,
-                          Uint32 flags);
+                          bool defaultRecord);
   void releaseRecord_impl(NdbRecord *rec);
 
+  static NdbDictionary::RecordType 
+  getRecordType(const NdbRecord* record);
+  static const char* getRecordTableName(const NdbRecord* record);
+  static const char* getRecordIndexName(const NdbRecord* record);
+  static bool getNextAttrIdFrom(const NdbRecord* record,
+                                Uint32 startAttrId,
+                                Uint32& nextAttrId);
+  static bool getOffset(const NdbRecord* record,
+                        Uint32 attrId,
+                        Uint32& offset);
+  static bool getNullBitOffset(const NdbRecord* record,
+                               Uint32 attrId,
+                               Uint32& nullbit_byte_offset,
+                               Uint32& nullbit_bit_in_byte);
+  static const char* getValuePtr(const NdbRecord* record,
+                                 const char* row,
+                                 Uint32 attrId);
+  static char* getValuePtr(const NdbRecord* record,
+                           char* row,
+                           Uint32 attrId);
+  static bool isNull(const NdbRecord* record,
+                     const char* row,
+                     Uint32 attrId);
+  static int setNull(const NdbRecord* record,
+                     char* row,
+                     Uint32 attrId,
+                     bool value);
+  static Uint32 getRecordRowLength(const NdbRecord* record);
+
+  /* Empty NdbRecord column mask for user convenience */
+  static const Uint32 m_emptyMask[MAXNROFATTRIBUTESINWORDS];
+
 private:
   NdbTableImpl * fetchGlobalTableImplRef(const GlobalCacheInitObject &obj);
 };
@@ -798,6 +954,7 @@ NdbColumnImpl::get_var_length(const void* value, Uint32& len) const
 {
   DBUG_ENTER("NdbColumnImpl::get_var_length");
   Uint32 max_len = m_attrSize * m_arraySize;
+
   switch (m_arrayType) {
   case NDB_ARRAYTYPE_SHORT_VAR:
     len = 1 + *((Uint8*)value);
@@ -815,6 +972,70 @@ NdbColumnImpl::get_var_length(const void* value, Uint32& len) const
   DBUG_RETURN(len <= max_len);
 }
 
+// Useful until 6.4 where WL4499 removes the need.
+inline
+bool
+NdbColumnImpl::get_var_length_bug39645(const void* value, 
+                                       Uint32& sigLen,
+                                       Uint32& sendLen) const
+{
+  DBUG_ENTER("NdbColumnImpl::get_var_length_bug39645");
+  Uint32 max_len = m_attrSize * m_arraySize;
+  
+  sendLen= max_len;
+
+  /* Special code for bug 39645 to help avoid copying too much
+   * from user space (and potentially reading bad memory
+   * Not needed when true VARCHAR to disk is implemented (WL4499)
+   */
+  if (unlikely(m_storageType == NDB_STORAGETYPE_DISK))
+  {
+    /* Disk VARCHAR/BINARY have arrayType of NDB_ARRAYTYPE_FIXED,
+     * but we should treat as VAR types w.r.t. determining 
+     * actual data length
+     */
+    DBUG_PRINT("info", ("DISK based : type=%u", m_type));
+
+    switch (m_type) {
+    case NDB_TYPE_VARCHAR: // Fall through
+    case NDB_TYPE_VARBINARY:
+      sigLen = 1 + *((Uint8*)value);
+      DBUG_PRINT("info", ("SHORT_VAR_DISK: sigLen=%u max_len=%u", sigLen, max_len));
+      DBUG_RETURN(sigLen <= max_len);
+
+    case NDB_TYPE_LONGVARCHAR: // Fall through
+    case NDB_TYPE_LONGVARBINARY:
+      sigLen = 2 + uint2korr((char*)value);
+      DBUG_PRINT("info", ("MEDIUM_VAR: sigLen=%u max_len=%u", sigLen, max_len));
+      DBUG_RETURN(sigLen <= max_len);
+    default:
+      // Fixed size disk type - fall through 
+      // to normal processing below
+      ;
+    }
+  }
+
+  /* Normal path */
+  switch (m_arrayType) {
+  case NDB_ARRAYTYPE_SHORT_VAR:
+    sigLen = 1 + *((Uint8*)value);
+    DBUG_PRINT("info", ("SHORT_VAR: sigLen=%u max_len=%u", sigLen, max_len));
+    break;
+  case NDB_ARRAYTYPE_MEDIUM_VAR:
+    sigLen = 2 + uint2korr((char*)value);
+    DBUG_PRINT("info", ("MEDIUM_VAR: sigLen=%u max_len=%u", sigLen, max_len));
+    break;
+  default:
+    sigLen = max_len;
+    DBUG_PRINT("info", ("FIXED: sigLen=%u max_len=%u", sigLen, max_len));
+    DBUG_RETURN(true);
+  }
+
+  sendLen= sigLen;
+  DBUG_RETURN(sigLen <= max_len);
+}
+
+
 inline
 NdbTableImpl &
 NdbTableImpl::getImpl(NdbDictionary::Table & t){
@@ -843,6 +1064,15 @@ NdbTableImpl::getMysqlName() const
   return m_mysqlName.c_str();
 }
 
+inline
+bool
+NdbTableImpl::matchDb(const char * name, size_t len) const
+{
+  return 
+    len < m_internalName.length() &&
+    memcmp(name, m_internalName.c_str(), len) == 0;
+}
+
 inline
 Uint32
 Hash( const char* str ){
@@ -971,13 +1201,16 @@ NdbDictionaryImpl::getImpl(const NdbDictionary::Dictionary & t){
 class InitTable : public GlobalCacheInitObject
 {
 public:
-  InitTable(NdbDictionaryImpl *dict,
-            const BaseString &name) :
-    GlobalCacheInitObject(dict, name)
+  InitTable(const BaseString &name) :
+    GlobalCacheInitObject(name)
   {}
-  int init(NdbTableImpl &tab) const
+  int init(NdbDictionaryImpl *dict, NdbTableImpl &tab) const
   {
-    return m_dict->getBlobTables(tab);
+    int res= dict->getBlobTables(tab);
+    if (res == 0)
+      res= dict->createDefaultNdbRecord(&tab, NULL);
+    
+    return res;
   }
 };
 
@@ -985,8 +1218,23 @@ inline
 NdbTableImpl *
 NdbDictionaryImpl::getTableGlobal(const char * table_name)
 {
+  if (unlikely(strchr(table_name, '$') != 0)) {
+    if (is_ndb_blob_table(table_name)) 
+    {
+      /* Could attempt to get the Blob table here, but
+       * instead we will generate an error.
+       * The non-global getTable() calls can fetch Blob
+       * tables correctly if necessary.
+       *
+       * 4307 Invalid Table name
+       */
+      m_error.code = 4307;
+      return NULL;
+    }
+  }
+
   const BaseString internal_tabname(m_ndb.internalize_table_name(table_name));
-  return fetchGlobalTableImplRef(InitTable(this, internal_tabname));
+  return fetchGlobalTableImplRef(InitTable(internal_tabname));
 }
 
 inline
@@ -1025,7 +1273,7 @@ NdbDictionaryImpl::get_local_table_info(const BaseString& internalTableName)
   if (info == 0)
   {
     NdbTableImpl *tab=
-      fetchGlobalTableImplRef(InitTable(this, internalTableName));
+      fetchGlobalTableImplRef(InitTable(internalTableName));
     if (tab)
     {
       info= Ndb_local_table_info::create(tab, m_local_table_data_size);
@@ -1047,12 +1295,12 @@ public:
   InitIndex(const BaseString &internal_indexname,
 	    const char *index_name,
 	    const NdbTableImpl &prim) :
-    GlobalCacheInitObject(0, internal_indexname),
+    GlobalCacheInitObject(internal_indexname),
     m_index_name(index_name),
     m_prim(prim)
     {}
   
-  int init(NdbTableImpl &tab) const {
+  int init(NdbDictionaryImpl *dict, NdbTableImpl &tab) const {
     DBUG_ENTER("InitIndex::init");
     DBUG_ASSERT(tab.m_indexType != NdbDictionary::Object::TypeUndefined);
     /**
@@ -1066,7 +1314,9 @@ public:
           !idx->m_internalName.assign(m_name))
         DBUG_RETURN(4000);
       tab.m_index = idx;
-      DBUG_RETURN(0);
+
+      /* Finally, create default NdbRecord for this index */
+      DBUG_RETURN(dict->createDefaultNdbRecord(&tab, &m_prim));
     }
     DBUG_RETURN(1);
   }
@@ -1133,8 +1383,20 @@ NdbDictionaryImpl::getIndexGlobal(const char * index_name,
   DBUG_RETURN(0);
 }
 
+inline
+NdbIndexImpl * 
+NdbDictionaryImpl::getIndexGlobal(const char * indexName,
+                                  const char * tableName)
+{
+  DBUG_ENTER("NdbDictionaryImpl::getIndexGlobal");
+  NdbTableImpl * t = getTableGlobal(tableName);
+  if(t == NULL)
+    DBUG_RETURN(0);
+  DBUG_RETURN(getIndexGlobal(indexName, *t));
+}
+
 inline int
-NdbDictionaryImpl::releaseTableGlobal(NdbTableImpl & impl, int invalidate)
+NdbDictionaryImpl::releaseTableGlobal(const NdbTableImpl & impl, int invalidate)
 {
   DBUG_ENTER("NdbDictionaryImpl::releaseTableGlobal");
   DBUG_PRINT("enter", ("internal_name: %s", impl.m_internalName.c_str()));
@@ -1145,7 +1407,7 @@ NdbDictionaryImpl::releaseTableGlobal(NdbTableImpl & impl, int invalidate)
 }
 
 inline int
-NdbDictionaryImpl::releaseIndexGlobal(NdbIndexImpl & impl, int invalidate)
+NdbDictionaryImpl::releaseIndexGlobal(const NdbIndexImpl & impl, int invalidate)
 {
   DBUG_ENTER("NdbDictionaryImpl::releaseIndexGlobal");
   DBUG_PRINT("enter", ("internal_name: %s", impl.m_internalName.c_str()));
