@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008-2010 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 #include <ndb_global.h>
@@ -44,7 +47,7 @@
 #include "ndb_internal.hpp"
 
 #include <EventLogger.hpp>
-extern EventLogger g_eventLogger;
+extern EventLogger * g_eventLogger;
 
 static Gci_container_pod g_empty_gci_container;
 
@@ -182,7 +185,7 @@ NdbEventOperationImpl::init(NdbEventImpl& evnt)
 
   m_state= EO_CREATED;
 
-  m_node_bit_mask.clear();
+  m_stop_gci = 0;
 #ifdef ndb_event_stores_merge_events_flag
   m_mergeEvents = m_eventImpl->m_mergeEvents;
 #else
@@ -606,7 +609,7 @@ NdbEventOperationImpl::execute_nolock()
   // add kernel reference
   // removed on TE_STOP, TE_CLUSTER_FAILURE, or error below
   m_ref_count++;
-  m_node_bit_mask.set(0u);
+  m_stop_gci= ~(Uint64)0;
   DBUG_PRINT("info", ("m_ref_count: %u for op: %p", m_ref_count, this));
   int r= NdbDictionaryImpl::getImpl(*myDict).executeSubscribeEvent(*this);
   if (r == 0) {
@@ -623,10 +626,6 @@ NdbEventOperationImpl::execute_nolock()
           m_error.code= myDict->getNdbError().code;
           DBUG_RETURN(r);
         }
-        // add blob reference to main op
-        // removed by TE_STOP or TE_CLUSTER_FAILURE
-        m_ref_count++;
-        DBUG_PRINT("info", ("m_ref_count: %u for op: %p", m_ref_count, this));
         blob_op = blob_op->m_next;
       }
     }
@@ -639,7 +638,7 @@ NdbEventOperationImpl::execute_nolock()
   // remove kernel reference
   // added above
   m_ref_count--;
-  m_node_bit_mask.clear(0u);
+  m_stop_gci = 0;
   DBUG_PRINT("info", ("m_ref_count: %u for op: %p", m_ref_count, this));
   m_state= EO_ERROR;
   mi_type= 0;
@@ -691,6 +690,23 @@ NdbEventOperationImpl::stop()
   m_state= EO_DROPPED;
   mi_type= 0;
   if (r == 0) {
+    if (m_stop_gci == 0)
+    {
+      // response from old kernel
+      Uint64 gci= m_ndb->theEventBuffer->m_highest_sub_gcp_complete_GCI;
+      if (gci)
+      {
+        // calculate a "safe" gci in the future to remove event op.
+        gci += Uint64(3) << 32;
+      }
+      else
+      {
+        // set highest value to ensure that operation does not get dropped
+        // too early. Note '-1' as ~Uint64(0) indicates active event
+        gci = ~Uint64(0)-1;
+      }
+      m_stop_gci = gci;
+    }
     m_ndb->theEventBuffer->add_drop_unlock();
     DBUG_RETURN(0);
   }
@@ -701,22 +717,22 @@ NdbEventOperationImpl::stop()
   DBUG_RETURN(r);
 }
 
-const bool NdbEventOperationImpl::tableNameChanged() const
+bool NdbEventOperationImpl::tableNameChanged() const
 {
   return (bool)AlterTableReq::getNameFlag(m_change_mask);
 }
 
-const bool NdbEventOperationImpl::tableFrmChanged() const
+bool NdbEventOperationImpl::tableFrmChanged() const
 {
   return (bool)AlterTableReq::getFrmFlag(m_change_mask);
 }
 
-const bool NdbEventOperationImpl::tableFragmentationChanged() const
+bool NdbEventOperationImpl::tableFragmentationChanged() const
 {
   return (bool)AlterTableReq::getFragDataFlag(m_change_mask);
 }
 
-const bool NdbEventOperationImpl::tableRangeListChanged() const
+bool NdbEventOperationImpl::tableRangeListChanged() const
 {
   return (bool)AlterTableReq::getRangeListFlag(m_change_mask);
 }
@@ -785,7 +801,8 @@ NdbEventOperationImpl::receive_event()
       // Parse the new table definition and
       // create a table object
       NdbError error;
-      NdbDictInterface dif(error);
+      int warn;
+      NdbDictInterface dif(error, warn);
       NdbTableImpl *at;
       m_change_mask = m_data_item->sdata->changeMask;
       error.code = dif.parseTableInfo(&at,
@@ -860,7 +877,7 @@ NdbEventOperationImpl::receive_event()
   DBUG_PRINT_EVENT("info",("sdata->operation %u  this: %p", operation, this));
   // now move the data into the RecAttrs
     
-  int is_update= operation == NdbDictionary::Event::_TE_UPDATE;
+  int is_insert= operation == NdbDictionary::Event::_TE_INSERT;
 
   Uint32 *aAttrPtr = m_data_item->ptr[0].p;
   Uint32 *aAttrEndPtr = aAttrPtr + m_data_item->ptr[0].sz;
@@ -883,7 +900,7 @@ NdbEventOperationImpl::receive_event()
       assert(tAttr->attrId() ==
 	     AttributeHeader(*aAttrPtr).getAttributeId());
       receive_data(tAttr, aDataPtr, tDataSz);
-      if (is_update)
+      if (!is_insert)
 	receive_data(tAttr1, aDataPtr, tDataSz);
       else
         tAttr1->setUNDEFINED(); // do not leave unspecified
@@ -900,7 +917,7 @@ NdbEventOperationImpl::receive_event()
   Uint32 tRecAttrId;
   Uint32 tAttrId;
   Uint32 tDataSz;
-  int hasSomeData=0;
+  int hasSomeData= (operation != NdbDictionary::Event::_TE_UPDATE);
   while ((aAttrPtr < aAttrEndPtr) && (tWorkingRecAttr != NULL)) {
     tRecAttrId = tWorkingRecAttr->attrId();
     tAttrId = AttributeHeader(*aAttrPtr).getAttributeId();
@@ -919,7 +936,7 @@ NdbEventOperationImpl::receive_event()
       break;
     
     if (tAttrId == tRecAttrId) {
-      hasSomeData++;
+      hasSomeData=1;
       
       DBUG_PRINT_EVENT("info",("set [%u] %u 0x%x [%u] 0x%x",
                                tAttrId, tDataSz, *aDataPtr, tRecAttrId, aDataPtr));
@@ -958,7 +975,7 @@ NdbEventOperationImpl::receive_event()
       break;
     if (tAttrId == tRecAttrId) {
       assert(!m_eventImpl->m_tableImpl->getColumn(tRecAttrId)->getPrimaryKey());
-      hasSomeData++;
+      hasSomeData=1;
       
       receive_data(tWorkingRecAttr, aDataPtr, tDataSz);
       tWorkingRecAttr = tWorkingRecAttr->next();
@@ -970,7 +987,7 @@ NdbEventOperationImpl::receive_event()
     tWorkingRecAttr = tWorkingRecAttr->next();
   }
   
-  if (hasSomeData || !is_update)
+  if (hasSomeData)
   {
     DBUG_RETURN_EVENT(1);
   }
@@ -1041,15 +1058,17 @@ NdbEventOperationImpl::printAll()
 
 NdbEventBuffer::NdbEventBuffer(Ndb *ndb) :
   m_system_nodes(ndb->theImpl->theNoOfDBnodes),
-  m_ndb(ndb),
-  m_latestGCI(0), m_latest_complete_GCI(0),
   m_min_gci_index(0),
   m_max_gci_index(0),
+  m_ndb(ndb),
+  m_latestGCI(0), m_latest_complete_GCI(0),
+  m_highest_sub_gcp_complete_GCI(0),
+  m_latest_poll_GCI(0),
   m_total_alloc(0),
-  m_free_thresh(10),
-  m_min_free_thresh(10),
-  m_max_free_thresh(100),
-  m_gci_slip_thresh(3),
+  m_free_thresh(0),
+  m_min_free_thresh(0),
+  m_max_free_thresh(0),
+  m_gci_slip_thresh(0),
   m_dropped_ev_op(0),
   m_active_op_count(0),
   m_add_drop_mutex(0)
@@ -1080,6 +1099,8 @@ NdbEventBuffer::NdbEventBuffer(Ndb *ndb) :
   // initialize lists
   bzero(&g_empty_gci_container, sizeof(Gci_container));
   init_gci_containers();
+
+  m_alive_node_bit_mask.clear();
 }
 
 NdbEventBuffer::~NdbEventBuffer()
@@ -1137,7 +1158,7 @@ NdbEventBuffer::init_gci_containers()
 {
   m_startup_hack = true;
   bzero(&m_complete_data, sizeof(m_complete_data));
-  m_latest_complete_GCI = m_latestGCI = 0;
+  m_latest_complete_GCI = m_latestGCI = m_latest_poll_GCI = 0;
   m_active_gci.clear();
   m_active_gci.fill(3, g_empty_gci_container);
   m_min_gci_index = m_max_gci_index = 1;
@@ -1193,8 +1214,7 @@ NdbEventBuffer::pollEvents(int aMillisecondNumber, Uint64 *latestGCI)
     if (unlikely(ev_op == 0))
       ret= 0;
   }
-  if (latestGCI)
-    *latestGCI= m_latestGCI;
+  m_latest_poll_GCI= m_latestGCI;
 #ifdef VM_TRACE
   if (ev_op)
   {
@@ -1205,7 +1225,19 @@ NdbEventBuffer::pollEvents(int aMillisecondNumber, Uint64 *latestGCI)
   }
   m_latest_command= m_latest_command_save;
 #endif
+  if (unlikely(ev_op == 0))
+  {
+    /*
+      gci's consumed up until m_latest_poll_GCI, so we can free all
+      dropped event operations stopped up until that gci
+    */
+    deleteUsedEventOperations(m_latest_poll_GCI);
+  }
   NdbMutex_Unlock(m_mutex); // we have moved the data
+
+  if (latestGCI)
+    *latestGCI= m_latest_poll_GCI;
+
   return ret;
 }
 
@@ -1220,8 +1252,8 @@ NdbEventBuffer::flushIncompleteEvents(Uint64 gci)
   Uint32 minpos = m_min_gci_index;
   Uint32 maxpos = m_max_gci_index;
 
-  g_eventLogger.info("Flushing incomplete GCI:s < %u/%u",
-                     Uint32(gci >> 32), Uint32(gci));
+  g_eventLogger->info("Flushing incomplete GCI:s < %u/%u",
+                      Uint32(gci >> 32), Uint32(gci));
   while (minpos != maxpos && array[minpos] < gci)
   {
     Gci_container* tmp = find_bucket(array[minpos]);
@@ -1324,9 +1356,6 @@ NdbEventBuffer::nextEvent()
          EventBufData_list::Gci_ops *gci_ops = m_available_data.first_gci_ops();
          while (gci_ops && op->getGCI() > gci_ops->m_gci)
          {
-           // moved to next gci, check if any references have been
-           // released when completing the last gci
-           deleteUsedEventOperations();
            gci_ops = m_available_data.delete_next_gci_ops();
          }
          if (!gci_ops->m_consistent)
@@ -1354,15 +1383,21 @@ NdbEventBuffer::nextEvent()
   m_latest_command= m_latest_command_save;
 #endif
 
-  // free all "per gci unique" collected operations
-  // completed gci, check if any references have been
-  // released when completing the gci
   EventBufData_list::Gci_ops *gci_ops = m_available_data.first_gci_ops();
   while (gci_ops)
   {
-    deleteUsedEventOperations();
     gci_ops = m_available_data.delete_next_gci_ops();
   }
+  /*
+    gci's consumed up until m_latest_poll_GCI, so we can free all
+    dropped event operations stopped up until that gci
+  */
+  if (m_dropped_ev_op)
+  {
+    NdbMutex_Lock(m_mutex);
+    deleteUsedEventOperations(m_latest_poll_GCI);
+    NdbMutex_Unlock(m_mutex);
+  }
   DBUG_RETURN_EVENT(0);
 }
 
@@ -1419,30 +1454,33 @@ NdbEventBuffer::getGCIEventOperations(Uint32* iter, Uint32* event_types)
 }
 
 void
-NdbEventBuffer::deleteUsedEventOperations()
+NdbEventBuffer::deleteUsedEventOperations(Uint64 last_consumed_gci)
 {
-  Uint32 iter= 0;
-  const NdbEventOperation *op_f;
-  while ((op_f= getGCIEventOperations(&iter, NULL)) != NULL)
+  NdbEventOperationImpl *op= m_dropped_ev_op;
+  while (op && op->m_stop_gci)
   {
-    NdbEventOperationImpl *op = &op_f->m_impl;
-    DBUG_ASSERT(op->m_ref_count > 0);
-    // remove gci reference
-    // added in inserDataL
-    op->m_ref_count--;
-    DBUG_PRINT("info", ("m_ref_count: %u for op: %p", op->m_ref_count, op));
-    if (op->m_ref_count == 0)
+    if (last_consumed_gci > op->m_stop_gci)
     {
-      DBUG_PRINT("info", ("deleting op: %p", op));
-      DBUG_ASSERT(op->m_node_bit_mask.isclear());
-      if (op->m_next)
-        op->m_next->m_prev = op->m_prev;
-      if (op->m_prev)
-        op->m_prev->m_next = op->m_next;
-      else
-        m_dropped_ev_op = op->m_next;
-      delete op->m_facade;
+      while (op)
+      {
+        NdbEventOperationImpl *next_op= op->m_next;
+        op->m_stop_gci= 0;
+        op->m_ref_count--;
+        if (op->m_ref_count == 0)
+        {
+          if (op->m_next)
+            op->m_next->m_prev = op->m_prev;
+          if (op->m_prev)
+            op->m_prev->m_next = op->m_next;
+          else
+            m_dropped_ev_op = op->m_next;
+          delete op->m_facade;
+        }
+        op = next_op;
+      }
+      break;
     }
+    op = op->m_next;
   }
 }
 
@@ -1832,15 +1870,8 @@ NdbEventBuffer::complete_bucket(Gci_container* bucket)
 
 void
 NdbEventBuffer::execSUB_GCP_COMPLETE_REP(const SubGcpCompleteRep * const rep,
-                                         Uint32 len)
+                                         Uint32 len, int complete_cluster_failure)
 {
-  if (unlikely(m_active_op_count == 0))
-  {
-    return;
-  }
-  
-  DBUG_ENTER_EVENT("NdbEventBuffer::execSUB_GCP_COMPLETE_REP");
-
   Uint32 gci_hi = rep->gci_hi;
   Uint32 gci_lo = rep->gci_lo;
 
@@ -1850,6 +1881,21 @@ NdbEventBuffer::execSUB_GCP_COMPLETE_REP(const SubGcpCompleteRep * const rep,
   }
 
   const Uint64 gci= gci_lo | (Uint64(gci_hi) << 32);
+  if (gci > m_highest_sub_gcp_complete_GCI)
+    m_highest_sub_gcp_complete_GCI = gci;
+
+  if (!complete_cluster_failure)
+  {
+    m_alive_node_bit_mask.set(refToNode(rep->senderRef));
+
+    if (unlikely(m_active_op_count == 0))
+    {
+      return;
+    }
+  }
+  
+  DBUG_ENTER_EVENT("NdbEventBuffer::execSUB_GCP_COMPLETE_REP");
+
   const Uint32 cnt= rep->gcp_complete_rep_count;
 
   Gci_container *bucket = find_bucket(gci);
@@ -1929,11 +1975,11 @@ NdbEventBuffer::execSUB_GCP_COMPLETE_REP(const SubGcpCompleteRep * const rep,
         goto do_complete;
       }
       /** out of order something */
-      g_eventLogger.info("out of order bucket: %d gci: %u/%u minGCI: %u/%u m_latestGCI: %u/%u",
-                         (int)(bucket-(Gci_container*)m_active_gci.getBase()),
-                         Uint32(gci >> 32), Uint32(gci),
-                         Uint32(minGCI >> 32), Uint32(minGCI),
-                         Uint32(m_latestGCI >> 32), Uint32(m_latestGCI));
+      g_eventLogger->info("out of order bucket: %d gci: %u/%u minGCI: %u/%u m_latestGCI: %u/%u",
+                          (int)(bucket-(Gci_container*)m_active_gci.getBase()),
+                          Uint32(gci >> 32), Uint32(gci),
+                          Uint32(minGCI >> 32), Uint32(minGCI),
+                          Uint32(m_latestGCI >> 32), Uint32(m_latestGCI));
       bucket->m_state = Gci_container::GC_COMPLETE;
       bucket->m_gcp_complete_rep_count = 1; // Prevent from being reused
       m_latest_complete_GCI = gci;
@@ -1957,14 +2003,14 @@ NdbEventBuffer::complete_outof_order_gcis()
   Uint64 stop_gci = m_latest_complete_GCI;
 
   Uint64 start_gci = array[minpos];
-  g_eventLogger.info("complete_outof_order_gcis from: %u/%u to: %u/%u",
-                     Uint32(start_gci >> 32), Uint32(start_gci),
-                     Uint32(stop_gci >> 32), Uint32(stop_gci));
+  g_eventLogger->info("complete_outof_order_gcis from: %u/%u(%u) to: %u/%u(%u)",
+                      Uint32(start_gci >> 32), Uint32(start_gci), minpos,
+                      Uint32(stop_gci >> 32), Uint32(stop_gci), maxpos);
 
   assert(start_gci <= stop_gci);
   do
   {
-    Uint64 start_gci = array[minpos];
+    start_gci = array[minpos];
     Gci_container* bucket = find_bucket(start_gci);
     assert(bucket);
     assert(maxpos == m_max_gci_index);
@@ -1978,10 +2024,10 @@ NdbEventBuffer::complete_outof_order_gcis()
 
 #ifdef VM_TRACE
     ndbout_c("complete_outof_order_gcis - completing %u/%u rows: %u",
-             Uint32(start_gci), Uint32(start_gci), bucket->m_data.m_count);
+             Uint32(start_gci >> 32), Uint32(start_gci), bucket->m_data.m_count);
 #else
     ndbout_c("complete_outof_order_gcis - completing %u/%u",
-             Uint32(start_gci), Uint32(start_gci));
+             Uint32(start_gci >> 32), Uint32(start_gci));
 #endif
     
     complete_bucket(bucket);
@@ -2000,31 +2046,25 @@ NdbEventBuffer::insert_event(NdbEventOperationImpl* impl,
                              LinearSectionPtr *ptr,
                              Uint32 &oid_ref)
 {
-  NdbEventOperationImpl *dropped_ev_op = m_dropped_ev_op;
   DBUG_PRINT("info", ("gci{hi/lo}: %u/%u", data.gci_hi, data.gci_lo));
   do
   {
-    do
+    if (impl->m_stop_gci == ~Uint64(0))
     {
-      if (impl->m_node_bit_mask.get(0u))
+      oid_ref = impl->m_oid;
+      insertDataL(impl, &data, SubTableData::SignalLength, ptr);
+    }
+    NdbEventOperationImpl* blob_op = impl->theBlobOpList;
+    while (blob_op != NULL)
+    {
+      if (blob_op->m_stop_gci == ~Uint64(0))
       {
-        oid_ref = impl->m_oid;
-        insertDataL(impl, &data, SubTableData::SignalLength, ptr);
+        oid_ref = blob_op->m_oid;
+        insertDataL(blob_op, &data, SubTableData::SignalLength, ptr);
       }
-      NdbEventOperationImpl* blob_op = impl->theBlobOpList;
-      while (blob_op != NULL)
-      {
-        if (blob_op->m_node_bit_mask.get(0u))
-        {
-          oid_ref = blob_op->m_oid;
-          insertDataL(blob_op, &data, SubTableData::SignalLength, ptr);
-        }
-        blob_op = blob_op->m_next;
-      }
-    } while((impl = impl->m_next));
-    impl = dropped_ev_op;
-    dropped_ev_op = NULL;
-  } while (impl);
+      blob_op = blob_op->m_next;
+    }
+  } while((impl = impl->m_next));
 }
 
 bool
@@ -2050,48 +2090,19 @@ NdbEventBuffer::find_max_known_gci(Uint64 * res) const
 void
 NdbEventBuffer::report_node_connected(Uint32 node_id)
 {
-  NdbEventOperation* op= m_ndb->getEventOperation(0);
-  if (op == 0)
-  {
-    return;
-  }
-  
-  DBUG_ENTER("NdbEventBuffer::report_node_connected");
-  SubTableData data;
-  LinearSectionPtr ptr[3];
-  bzero(&data, sizeof(data));
-  bzero(ptr, sizeof(ptr));
-
-  data.tableId = ~0;
-  data.requestInfo = 0;
-  SubTableData::setOperation(data.requestInfo,
-			     NdbDictionary::Event::_TE_ACTIVE);
-  SubTableData::setReqNodeId(data.requestInfo, node_id);
-  SubTableData::setNdbdNodeId(data.requestInfo, node_id);
-  data.flags = SubTableData::LOG;
-
-  Uint64 gci = Uint64((m_latestGCI >> 32) + 1) << 32;
-  find_max_known_gci(&gci);
-
-  data.gci_hi = Uint32(gci >> 32);
-  data.gci_lo = Uint32(gci);
-
-  /**
-   * Insert this event for each operation
-   */
-  // no need to lock()/unlock(), receive thread calls this
-  insert_event(&op->m_impl, data, ptr, data.senderData);
-  DBUG_VOID_RETURN;
+  return;
 }
 
 void
-NdbEventBuffer::report_node_failure(Uint32 node_id)
+NdbEventBuffer::report_node_failure_completed(Uint32 node_id)
 {
+  m_alive_node_bit_mask.clear(node_id);
+
   NdbEventOperation* op= m_ndb->getEventOperation(0);
   if (op == 0)
     return;
 
-  DBUG_ENTER("NdbEventBuffer::report_node_failure");
+  DBUG_ENTER("NdbEventBuffer::report_node_failure_completed");
   SubTableData data;
   LinearSectionPtr ptr[3];
   bzero(&data, sizeof(data));
@@ -2116,20 +2127,17 @@ NdbEventBuffer::report_node_failure(Uint32 node_id)
    */
   // no need to lock()/unlock(), receive thread calls this
   insert_event(&op->m_impl, data, ptr, data.senderData);
-  DBUG_VOID_RETURN;
-}
 
-void
-NdbEventBuffer::completeClusterFailed()
-{
-  NdbEventOperation* op= m_ndb->getEventOperation(0);
-  if (op == 0)
-    return;
+  if (!m_alive_node_bit_mask.isclear())
+    DBUG_VOID_RETURN;
 
-  DBUG_ENTER("NdbEventBuffer::completeClusterFailed");
+  /*
+   * Cluster failure
+   */
 
+  DBUG_PRINT("info", ("Cluster failure"));
 
-  Uint64 gci = Uint64((m_latestGCI >> 32) + 1) << 32;
+  gci = Uint64((m_latestGCI >> 32) + 1) << 32;
   bool found = find_max_known_gci(&gci);
 
   Uint64 * array = m_known_gci.getBase();
@@ -2165,18 +2173,10 @@ NdbEventBuffer::completeClusterFailed()
   /**
    * Inject new event
    */
-  SubTableData data;
-  LinearSectionPtr ptr[3];
-  bzero(&data, sizeof(data));
-  bzero(ptr, sizeof(ptr));
-
   data.tableId = ~0;
   data.requestInfo = 0;
   SubTableData::setOperation(data.requestInfo,
 			     NdbDictionary::Event::_TE_CLUSTER_FAILURE);
-  data.flags = SubTableData::LOG;
-  data.gci_hi = Uint32(gci >> 32);
-  data.gci_lo = Uint32(gci);
 
   /**
    * Insert this event for each operation
@@ -2208,7 +2208,7 @@ NdbEventBuffer::completeClusterFailed()
   rep.gci_lo= gci & 0xFFFFFFFF;
   rep.gcp_complete_rep_count= cnt;
   rep.flags = 0;
-  execSUB_GCP_COMPLETE_REP(&rep, SubGcpCompleteRep::SignalLength);
+  execSUB_GCP_COMPLETE_REP(&rep, SubGcpCompleteRep::SignalLength, 1);
 
   DBUG_VOID_RETURN;
 }
@@ -2242,74 +2242,29 @@ NdbEventBuffer::insertDataL(NdbEventOperationImpl *op,
 
   if (!is_data_event)
   {
-    switch (operation)
+    if (operation == NdbDictionary::Event::_TE_CLUSTER_FAILURE)
+    {
+      /*
+        Mark event as stopping.  Subsequent dropEventOperation
+        will add the event to the dropped list for delete
+      */
+      op->m_stop_gci = gci;
+    }
+    else if (operation == NdbDictionary::Event::_TE_ACTIVE)
     {
-    case NdbDictionary::Event::_TE_NODE_FAILURE:
-      DBUG_ASSERT(op->m_node_bit_mask.get(0u) != 0);
-      op->m_node_bit_mask.clear(SubTableData::getNdbdNodeId(ri));
-      DBUG_PRINT("info",
-                 ("_TE_NODE_FAILURE: m_ref_count: %u for op: %p id: %u",
-                  op->m_ref_count, op, SubTableData::getNdbdNodeId(ri)));
-      break;
-    case NdbDictionary::Event::_TE_ACTIVE:
-      DBUG_ASSERT(op->m_node_bit_mask.get(0u) != 0);
-      op->m_node_bit_mask.set(SubTableData::getNdbdNodeId(ri));
       // internal event, do not relay to user
       DBUG_PRINT("info",
                  ("_TE_ACTIVE: m_ref_count: %u for op: %p id: %u",
                   op->m_ref_count, op, SubTableData::getNdbdNodeId(ri)));
       DBUG_RETURN_EVENT(0);
-      break;
-    case NdbDictionary::Event::_TE_CLUSTER_FAILURE:
-      DBUG_ASSERT(op->m_node_bit_mask.get(0u) != 0);
-      op->m_node_bit_mask.clear();
-      DBUG_ASSERT(op->m_ref_count > 0);
-      // remove kernel reference
-      // added in execute_nolock
-      op->m_ref_count--;
-      DBUG_PRINT("info", ("_TE_CLUSTER_FAILURE: m_ref_count: %u for op: %p",
-                          op->m_ref_count, op));
-      if (op->theMainOp)
-      {
-        DBUG_ASSERT(op->m_ref_count == 0);
-        DBUG_ASSERT(op->theMainOp->m_ref_count > 0);
-        // remove blob reference in main op
-        // added in execute_no_lock
-        op->theMainOp->m_ref_count--;
-        DBUG_PRINT("info", ("m_ref_count: %u for op: %p",
-                            op->theMainOp->m_ref_count, op->theMainOp));
-      }
-      break;
-    case NdbDictionary::Event::_TE_STOP:
-      DBUG_ASSERT(op->m_node_bit_mask.get(0u) != 0);
-      op->m_node_bit_mask.clear(0u);
-      op->m_node_bit_mask.clear(SubTableData::getNdbdNodeId(ri));
-      if (op->m_node_bit_mask.isclear())
-      {
-        DBUG_ASSERT(op->m_ref_count > 0);
-        // remove kernel reference
-        // added in execute_no_lock
-        op->m_ref_count--;
-        DBUG_PRINT("info", ("_TE_STOP: m_ref_count: %u for op: %p",
-                            op->m_ref_count, op));
-        if (op->theMainOp)
-        {
-          DBUG_ASSERT(op->m_ref_count == 0);
-          DBUG_ASSERT(op->theMainOp->m_ref_count > 0);
-          // remove blob reference in main op
-          // added in execute_no_lock
-          op->theMainOp->m_ref_count--;
-          DBUG_PRINT("info", ("m_ref_count: %u for op: %p",
-                              op->theMainOp->m_ref_count, op->theMainOp));
-        }
-      }
-      else
-      {
-        op->m_node_bit_mask.set(0u);
-      }
-      break;
-    default:
-      break;
+    }
+    else if (operation == NdbDictionary::Event::_TE_STOP)
+    {
+      // internal event, do not relay to user
+      DBUG_PRINT("info",
+                 ("_TE_STOP: m_ref_count: %u for op: %p id: %u",
+                  op->m_ref_count, op, SubTableData::getNdbdNodeId(ri)));
+      DBUG_RETURN_EVENT(0);
     }
   }
   
@@ -3095,10 +3050,6 @@ EventBufData_list::add_gci_op(Gci_op g)
 #ifndef DBUG_OFF
     i = m_gci_op_count;
 #endif
-    // add gci reference
-    // removed in deleteUsedOperations
-    g.op->m_ref_count++;
-    DBUG_PRINT("info", ("m_ref_count: %u for op: %p", g.op->m_ref_count, g.op));
     m_gci_op_list[m_gci_op_count++] = g;
   }
   DBUG_PRINT_EVENT("exit", ("m_gci_op_list[%u].event_types: %x", i, m_gci_op_list[i].event_types));
@@ -3202,12 +3153,23 @@ NdbEventBuffer::dropEventOperation(NdbEventOperation* tOp)
   // stop blob event ops
   if (op->theMainOp == NULL)
   {
+    Uint64 max_stop_gci = op->m_stop_gci;
     NdbEventOperationImpl* tBlobOp = op->theBlobOpList;
     while (tBlobOp != NULL)
     {
       tBlobOp->stop();
+      Uint64 stop_gci = tBlobOp->m_stop_gci;
+      if (stop_gci > max_stop_gci)
+        max_stop_gci = stop_gci;
       tBlobOp = tBlobOp->m_next;
     }
+    tBlobOp = op->theBlobOpList;
+    while (tBlobOp != NULL)
+    {
+      tBlobOp->m_stop_gci = max_stop_gci;
+      tBlobOp = tBlobOp->m_next;
+    }
+    op->m_stop_gci = max_stop_gci;
   }
 
   /**
@@ -3245,7 +3207,6 @@ NdbEventBuffer::dropEventOperation(NdbEventOperation* tOp)
   {
     NdbMutex_Unlock(m_mutex);
     DBUG_PRINT("info", ("deleting op: %p", op));
-    DBUG_ASSERT(op->m_node_bit_mask.isclear());
     delete op->m_facade;
   }
   else
@@ -3277,28 +3238,32 @@ NdbEventBuffer::reportStatus()
   else
     apply_gci= latest_gci;
 
-  if (100*(Uint64)m_free_data_sz < m_min_free_thresh*(Uint64)m_total_alloc &&
-      m_total_alloc > 1024*1024)
+  if (m_free_thresh)
   {
-    /* report less free buffer than m_free_thresh,
-       next report when more free than 2 * m_free_thresh
-    */
-    m_min_free_thresh= 0;
-    m_max_free_thresh= 2 * m_free_thresh;
-    goto send_report;
-  }
+    if (100*(Uint64)m_free_data_sz < m_min_free_thresh*(Uint64)m_total_alloc &&
+        m_total_alloc > 1024*1024)
+    {
+      /* report less free buffer than m_free_thresh,
+         next report when more free than 2 * m_free_thresh
+      */
+      m_min_free_thresh= 0;
+      m_max_free_thresh= 2 * m_free_thresh;
+      goto send_report;
+    }
   
-  if (100*(Uint64)m_free_data_sz > m_max_free_thresh*(Uint64)m_total_alloc &&
-      m_total_alloc > 1024*1024)
-  {
-    /* report more free than 2 * m_free_thresh
-       next report when less free than m_free_thresh
-    */
-    m_min_free_thresh= m_free_thresh;
-    m_max_free_thresh= 100;
-    goto send_report;
+    if (100*(Uint64)m_free_data_sz > m_max_free_thresh*(Uint64)m_total_alloc &&
+        m_total_alloc > 1024*1024)
+    {
+      /* report more free than 2 * m_free_thresh
+         next report when less free than m_free_thresh
+      */
+      m_min_free_thresh= m_free_thresh;
+      m_max_free_thresh= 100;
+      goto send_report;
+    }
   }
-  if (latest_gci-apply_gci >=  m_gci_slip_thresh)
+  if (m_gci_slip_thresh &&
+      (latest_gci-apply_gci >= m_gci_slip_thresh))
   {
     goto send_report;
   }
@@ -3314,7 +3279,7 @@ send_report:
   data[5]= apply_gci >> 32;
   data[6]= latest_gci & ~(Uint32)0;
   data[7]= latest_gci >> 32;
-  Ndb_internal::send_event_report(m_ndb, data,8);
+  Ndb_internal::send_event_report(true, m_ndb, data,8);
 #ifdef VM_TRACE
   assert(m_total_alloc >= m_free_data_sz);
 #endif
