@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008, 2009 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
 #include <NdbOperation.hpp>
@@ -26,7 +29,7 @@
 #include <NdbBlob.hpp>
 
 #include <Interpreter.hpp>
-
+#include <NdbInterpretedCode.hpp>
 #include <AttributeHeader.hpp>
 #include <signaldata/TcKeyReq.hpp>
 
@@ -387,6 +390,7 @@ NdbOperation::getValue_impl(const NdbColumnImpl* tAttrInfo, char* aValue)
     m_no_disk_flag &= (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
     if (theStatus != GetValue) {
       if (theStatus == UseNdbRecord)
+        /* This path for extra GetValues for NdbRecord */
         return getValue_NdbRecord(tAttrInfo, aValue);
       if (theInterpretIndicator == 1) {
 	if (theStatus == FinalGetValue) {
@@ -441,6 +445,9 @@ NdbRecAttr*
 NdbOperation::getValue_NdbRecord(const NdbColumnImpl* tAttrInfo, char* aValue)
 {
   NdbRecAttr* tRecAttr;
+
+  m_no_disk_flag &= (tAttrInfo->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+
   /*
     For getValue with NdbRecord operations, we just allocate the NdbRecAttr,
     the signal data will be constructed later.
@@ -471,13 +478,13 @@ NdbOperation::setValue( const NdbColumnImpl* tAttrInfo,
 {
   DBUG_ENTER("NdbOperation::setValue");
   DBUG_PRINT("enter", ("col: %s  op:%d  val: 0x%lx",
-                       tAttrInfo->m_name.c_str(), theOperationType,
-                       (long) aValuePassed));
+                       tAttrInfo ? tAttrInfo->m_name.c_str() : "NULL",
+                       theOperationType, (long) aValuePassed));
 
   int tReturnCode;
   Uint32 tAttrId;
   Uint32 tData;
-  Uint32 tempData[2000];
+  Uint32 tempData[ NDB_MAX_TUPLE_SIZE_IN_WORDS ];
   OperationType tOpType = theOperationType;
   OperationStatus tStatus = theStatus;
 
@@ -571,20 +578,24 @@ NdbOperation::setValue( const NdbColumnImpl* tAttrInfo,
     }//if
   }//if
   
-  Uint32 len;
-  if (! tAttrInfo->get_var_length(aValue, len)) {
+  Uint32 sigLen = 0;  // Length of significant part
+  Uint32 sendLen = 0; // Length of data to send.  Normally sigLen == sendLen
+
+  if (! tAttrInfo->get_var_length_bug39645(aValue, sigLen, sendLen)) {
     setErrorCodeAbort(4209);
     DBUG_RETURN(-1);
   }
-
-  const Uint32 sizeInBytes = len;
+  
+  const Uint32 sizeInBytes = sendLen;
   const Uint32 bitsInLastWord = 8 * (sizeInBytes & 3) ;
   
   const int attributeSize = sizeInBytes;
   const int slack = sizeInBytes & 3;
   
-  if (((UintPtr)aValue & 3) != 0 || (slack != 0)){
-    memcpy(&tempData[0], aValue, attributeSize);
+  if (((UintPtr)aValue & 3) != 0 || // Alignment 
+      (slack != 0) ||               // Slack bits
+      (sigLen != sendLen)){         // Bug39645
+    memcpy(&tempData[0], aValue, sigLen);
     aValue = (char*)&tempData[0];
     if(slack != 0) {
       char * tmp = (char*)&tempData[0];
@@ -630,19 +641,10 @@ NdbOperation::setAnyValue(Uint32 any_value)
 
   if (theStatus == UseNdbRecord)
   {
-    switch(tOpType)
-    {
-      case InsertRequest:
-      case WriteRequest:
-      case UpdateRequest:
-      case DeleteRequest:
-        m_any_value= any_value;
-        m_use_any_value= 1;
-        return 0;
-      default:
-        setErrorCodeAbort(4504);
-        return -1;
-    }
+    /* Method not allowed for NdbRecord, use OperationOptions or 
+       ScanOptions structure instead */
+    setErrorCodeAbort(4515);
+    return -1;
   }
 
   const NdbColumnImpl* impl =
@@ -665,7 +667,16 @@ NdbOperation::setAnyValue(Uint32 any_value)
   return -1;
 }
 
+int
+NdbOperation::setOptimize(Uint32 options)
+{
+  return setValue(&NdbColumnImpl::getImpl(*NdbDictionary::Column::OPTIMIZE),
+                  (const char*)&options);
+}
 
+/* Non-const variant of getBlobHandle - can return existing blob
+ * handles, or create new ones for non-NdbRecord operations 
+ */
 NdbBlob*
 NdbOperation::getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* tAttrInfo)
 {
@@ -679,16 +690,39 @@ NdbOperation::getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* tAttrInfo
   }
 
   /*
-    For NdbRecord operation, we only fetch existing blob handles here,
-    creation must be done by requesting the blob in the NdbRecord and
-    mask when creating the operation.
-  */
+   * For NdbRecord PK, unique index and scan operations, we only fetch existing 
+   * blob handles here, creation must be done by requesting the blob in the 
+   * NdbRecord and mask when creating the operation.
+   * For NdbRecAttr PK, IK and scan operations, we allow Blob handles
+   * to be created here.  Note that NdbRecAttr PK and unique index ops are handled
+   * differently to NdbRecAttr scan operations.
+   */
   if (m_attribute_record)
   {
     setErrorCodeAbort(4288);
     return NULL;
   }
 
+  /* Check key fully defined for key operations */
+  switch (theStatus)
+  {
+  case TupleKeyDefined:
+  case GetValue:
+  case SetValue:
+  case FinalGetValue:
+  case ExecInterpretedValue:
+  case SetValueInterpreted:
+    /* All ok states to create a Blob Handle in */
+    break;
+  default:
+  {
+    /* Unexpected state to be obtaining Blob handle */
+    /* Invalid usage of blob attribute */
+    setErrorCodeAbort(4264);
+    return NULL;
+  }
+  }
+
   tBlob = theNdb->getNdbBlob();
   if (tBlob == NULL)
     return NULL;
@@ -705,6 +739,25 @@ NdbOperation::getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* tAttrInfo
   return tBlob;
 }
 
+/* const variant of getBlobHandle - only returns existing blob handles */
+NdbBlob*
+NdbOperation::getBlobHandle(NdbTransaction* aCon, const NdbColumnImpl* tAttrInfo) const
+{
+  NdbBlob* tBlob = theBlobList;
+  while (tBlob != NULL) {
+    if (tBlob->theColumn == tAttrInfo)
+      return tBlob;
+    tBlob = tBlob->theNext;
+  }
+
+  /*
+    Const method - cannot create a new BLOB handle, NdbRecord
+    or NdbRecAttr
+  */
+  setErrorCodeAbort(4288);
+  return NULL;
+}
+
 /*
   This is used to set up a blob handle for an NdbRecord operation.
 
@@ -765,12 +818,12 @@ NdbOperation::linkInBlobHandle(NdbTransaction *aCon,
 }
 
 /*
-  Setup blob handles for an NdbRecord operation.
-
-  Create blob handles for all requested blob columns.
-
-  For read request, store the pointers to blob handles in the row.
-*/
+ * Setup blob handles for an NdbRecord operation.
+ *
+ * Create blob handles for all requested blob columns.
+ *
+ * For read request, store the pointers to blob handles in the row.
+ */
 int
 NdbOperation::getBlobHandlesNdbRecord(NdbTransaction* aCon)
 {
@@ -797,9 +850,9 @@ NdbOperation::getBlobHandlesNdbRecord(NdbTransaction* aCon)
     if (theOperationType == ReadRequest || theOperationType == ReadExclusive)
     {
       /*
-        For read request, it is safe to cast away const-ness for the
-        m_attribute_row.
-      */
+       * For read request, it is safe to cast away const-ness for the
+       * m_attribute_row.
+       */
       memcpy((char *)&m_attribute_row[col->offset], &bh, sizeof(bh));
     }
   }
@@ -810,9 +863,12 @@ NdbOperation::getBlobHandlesNdbRecord(NdbTransaction* aCon)
 /*
   For a delete, we need to create blob handles for all table blob columns,
   so that we can be sure to delete all blob parts for the row.
+  If checkReadset is true, we also check that the caller is not asking to 
+  read any blobs as part of the delete.
 */
 int
-NdbOperation::getBlobHandlesDelete(NdbTransaction* aCon)
+NdbOperation::getBlobHandlesNdbRecordDelete(NdbTransaction* aCon,
+                                            bool checkReadSet)
 {
   NdbBlob *lastBlob= NULL;
 
@@ -825,6 +881,15 @@ NdbOperation::getBlobHandlesDelete(NdbTransaction* aCon)
     if (!c->getBlobType())
       continue;
 
+    if (checkReadSet &&
+        (BitmaskImpl::get((NDB_MAX_ATTRIBUTES_IN_TABLE+31)>>5,
+                          m_read_mask, c->m_attrId)))
+    {
+      /* Blobs are not allowed in NdbRecord delete result record */
+      setErrorCodeAbort(4511);
+      return -1;
+    }
+
     NdbBlob *bh= linkInBlobHandle(aCon, c, lastBlob);
     if (bh == NULL)
       return -1;
@@ -853,7 +918,9 @@ NdbOperation::setVarValue(const NdbColumnImpl* tAttrInfo,
   DBUG_PRINT("info", ("aLen=%u", (Uint32)aLen));
 
   // wl3717_todo not optimal..
-  Uint64 buf[2048];
+  const Uint32 MaxTupleSizeInLongWords= (NDB_MAX_TUPLE_SIZE + 7)/ 8;
+  Uint64 buf[ MaxTupleSizeInLongWords ];
+  assert( aLen < (NDB_MAX_TUPLE_SIZE - 2) );
   unsigned char* p = (unsigned char*)buf;
   p[0] = (aLen & 0xff);
   p[1] = (aLen >> 8);
@@ -989,6 +1056,14 @@ NdbOperation::getAbortOption() const
 int
 NdbOperation::setAbortOption(AbortOption ao)
 {
+  if (theStatus == UseNdbRecord)
+  {
+    /* Method not allowed for NdbRecord, use OperationOptions or 
+       ScanOptions structure instead */
+    setErrorCodeAbort(4515);
+    return -1;
+  }
+
   switch(ao)
   {
     case AO_IgnoreError:
@@ -999,3 +1074,250 @@ NdbOperation::setAbortOption(AbortOption ao)
       return -1;
   }
 }
+
+
+/*
+ * handleOperationOptions
+ * static member for setting operation options
+ * Called when defining operations, from NdbTransaction and
+ * NdbScanOperation
+ */
+int
+NdbOperation::handleOperationOptions (const OperationType type,
+                                      const OperationOptions *opts,
+                                      const Uint32 sizeOfOptions,
+                                      NdbOperation *op)
+{
+  /* Check options size for versioning... */
+  if (unlikely((sizeOfOptions != 0) && 
+               (sizeOfOptions != sizeof(OperationOptions))))
+  {
+    // Handle different sized OperationOptions
+    // Probably smaller is old version, larger is new version.
+    
+    // No other versions currently supported
+    // Invalid or unsupported OperationOptions structure
+    return 4297;
+  }
+
+  bool isScanTakeoverOp = (op->m_key_record == NULL); 
+  
+  if (opts->optionsPresent & OperationOptions::OO_ABORTOPTION)
+  {
+    /* User defined operation abortoption : Allowed for 
+     * any operation 
+     */
+    switch (opts->abortOption)
+    {
+    case AO_IgnoreError:
+    case AbortOnError:
+    {
+      op->m_abortOption=opts->abortOption;
+      break;
+    }
+    default:
+      // Non-specific abortoption
+      // Invalid AbortOption
+      return 4296;
+    }
+  } 
+
+  if ((opts->optionsPresent & OperationOptions::OO_GETVALUE) &&
+      (opts->numExtraGetValues > 0))
+  {
+    if (opts->extraGetValues == NULL)
+    {
+      // Incorrect combination of OperationOptions optionsPresent, 
+      // extraGet/SetValues ptr and numExtraGet/SetValues
+      return 4512;
+    }
+
+    // Only certain operation types allow extra GetValues
+    // Update could be made to support it in future
+    if (type == ReadRequest ||
+        type == ReadExclusive ||
+        type == DeleteRequest)
+    {
+      // Could be readTuple(), or lockCurrentTuple().
+      // We perform old-school NdbRecAttr reads on
+      // these values.
+      for (unsigned int i=0; i < opts->numExtraGetValues; i++)
+      {
+        GetValueSpec *pvalSpec 
+          = &(opts->extraGetValues[i]);
+
+        pvalSpec->recAttr=NULL;
+
+        if (pvalSpec->column == NULL)
+        {
+          // Column is NULL in Get/SetValueSpec structure
+          return 4295;
+        }
+
+        NdbRecAttr *pra=
+          op->getValue_NdbRecord(&NdbColumnImpl::getImpl(*pvalSpec->column),
+                                 (char *) pvalSpec->appStorage);
+        
+        if (pra == NULL)
+        {
+          return -1;
+        }
+
+        pvalSpec->recAttr = pra;
+      }
+    }
+    else
+    {
+      // Bad operation type for GetValue
+      switch (type)
+      {
+      case WriteRequest : 
+      case UpdateRequest :
+      {
+        return 4502;
+        // GetValue not allowed in Update operation
+      }
+      case InsertRequest :
+      {
+        return 4503;
+        // GetValue not allowed in Insert operation
+      }
+      default :
+        return 4118;
+        // Parameter error in API call
+      }
+    }
+  }
+
+  if ((opts->optionsPresent & OperationOptions::OO_SETVALUE) &&
+      (opts->numExtraSetValues > 0))
+  {
+    if (opts->extraSetValues == NULL)
+    {
+      // Incorrect combination of OperationOptions optionsPresent, 
+      // extraGet/SetValues ptr and numExtraGet/SetValues
+      return 4512;
+    }
+
+    if ((type == InsertRequest) ||
+        (type == UpdateRequest) ||
+        (type == WriteRequest))
+    {
+      /* Could be insert/update/writeTuple() or 
+       * updateCurrentTuple()
+       */
+      // Validate SetValuesSpec
+      for (Uint32 i=0; i< opts->numExtraSetValues; i++)
+      {
+        const NdbDictionary::Column *pcol=opts->extraSetValues[i].column;
+        const void *pvalue=opts->extraSetValues[i].value;
+
+        if (pcol == NULL)
+        {
+          // Column is NULL in Get/SetValueSpec structure
+          return 4295;
+        }
+
+        if (type == UpdateRequest && pcol->getPrimaryKey())
+        {
+          // It is not possible to update a primary key column.
+          // It can be set like this for insert and write (but it
+          // still needs to be included in the key NdbRecord and row).
+          return 4202;
+        }
+
+        if (pvalue == NULL)
+        {
+          if (!pcol->getNullable())
+          {
+            // Trying to set a NOT NULL attribute to NULL
+            return 4203;
+          }
+        }
+          
+        NdbDictionary::Column::Type colType=pcol->getType();
+          
+        if ((colType == NdbDictionary::Column::Blob) ||
+            (colType == NdbDictionary::Column::Text))
+        {
+          // Invalid usage of blob attribute
+          return 4264;
+        }          
+      }
+
+      // Store details of extra set values for later
+      op->m_extraSetValues = opts->extraSetValues;
+      op->m_numExtraSetValues = opts->numExtraSetValues;
+    }
+    else
+    {
+      // Set value and Read/Delete etc is incompatible
+      return 4204;
+    }
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_PARTITION_ID)
+  {
+    /* Should not have any blobs defined at this stage */
+    assert(op->theBlobList == NULL);
+
+    /* Not allowed for scan takeover ops */
+    if (unlikely(isScanTakeoverOp))
+    {
+      return 4510;
+      /* User-specified partition id not allowed for scan 
+       * takeover operation 
+       */
+    }
+    op->theDistributionKey=opts->partitionId;
+    op->theDistrKeyIndicator_= 1;       
+  }
+    
+  if (opts->optionsPresent & OperationOptions::OO_INTERPRETED)
+  {
+    /* Check the operation type is valid */
+    if (! ((type == ReadRequest)   ||
+           (type == ReadExclusive) ||
+           (type == UpdateRequest) ||
+           (type == DeleteRequest)))
+      /* NdbInterpretedCode not supported for operation type */
+      return 4539;
+    
+    /* Check the program's for the same table as the
+     * operation, within a major version number
+     * Perhaps NdbInterpretedCode should not contain the table
+     */
+    const NdbDictionary::Table* codeTable= opts->interpretedCode->getTable();
+    if (codeTable != NULL)
+    {
+      NdbTableImpl* impl= &NdbTableImpl::getImpl(*codeTable);
+      
+      if ((impl->m_id != (int) op->m_attribute_record->tableId) ||
+          (table_version_major(impl->m_version) != 
+           table_version_major(op->m_attribute_record->tableVersion)))
+        return 4524; // NdbInterpretedCode is for different table`
+    }
+
+    /* Check the program's finalised */
+    if ((opts->interpretedCode->m_flags & 
+         NdbInterpretedCode::Finalised) == 0)
+      return 4519; // NdbInterpretedCode::finalise() not called.
+
+    op->m_interpreted_code = opts->interpretedCode;
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_ANYVALUE)
+  {
+    /* Any operation can have an ANYVALUE set */
+    op->m_any_value = opts->anyValue;
+    op->m_use_any_value = 1;
+  }
+
+  if (opts->optionsPresent & OperationOptions::OO_CUSTOMDATA)
+  {
+    /* Set the operation's customData ptr */
+    op->m_customData = opts->customData;
+  }
+
+  return 0;
+}
