@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008, 2009 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 /******************************************************************************
@@ -58,11 +61,14 @@ NdbOperation::equal_impl(const NdbColumnImpl* tAttrInfo,
 {
   DBUG_ENTER("NdbOperation::equal_impl");
   DBUG_PRINT("enter", ("col: %s  op: %d  val: 0x%lx",
-                       tAttrInfo->m_name.c_str(), theOperationType,
+                       (tAttrInfo == NULL) ? "NULL" :
+                       tAttrInfo->m_name.c_str(), 
+                       theOperationType,
                        (long) aValuePassed));
   
   const char* aValue = aValuePassed;
-  Uint64 tempData[512];
+  const Uint32 MaxKeyLenInLongWords= (NDB_MAX_KEY_SIZE + 7)/8;
+  Uint64 tempData[ MaxKeyLenInLongWords ];
 
   if ((theStatus == OperationDefined) &&
       (aValue != NULL) &&
@@ -465,9 +471,10 @@ LastWordLabel:
 void
 NdbOperation::reorderKEYINFO()
 {
-  Uint32 data[4000];
-  Uint32 size = 4000;
-  getKeyFromTCREQ(data, size);
+  Uint32 data[ NDB_MAX_KEYSIZE_IN_WORDS ];
+  Uint32 size = NDB_MAX_KEYSIZE_IN_WORDS;
+  int rc = getKeyFromTCREQ(data, size);
+  assert(rc == 0);
   Uint32 pos = 1;
   Uint32 k;
   for (k = 0; k < m_accessTable->m_noOfKeys; k++) {
@@ -480,7 +487,8 @@ NdbOperation::reorderKEYINFO()
           if (theTupleKeyDefined[j][0] == i) {
             Uint32 off = theTupleKeyDefined[j][1] - 1;
             Uint32 len = theTupleKeyDefined[j][2];
-            assert(off < 4000 && off + len <= 4000);
+            assert(off < NDB_MAX_KEYSIZE_IN_WORDS && 
+                   off + len <= NDB_MAX_KEYSIZE_IN_WORDS);
             int ret = insertKEYINFO((char*)&data[off], pos, len);
             assert(ret == 0);
             pos += len;
@@ -498,7 +506,10 @@ NdbOperation::reorderKEYINFO()
 int
 NdbOperation::getKeyFromTCREQ(Uint32* data, Uint32 & size)
 {
-  assert(size >= theTupKeyLen && theTupKeyLen > 0);
+  /* Check that we can correctly return a valid key */
+  if ((size < theTupKeyLen) || (theTupKeyLen == 0))
+    return -1;
+  
   size = theTupKeyLen;
   unsigned pos = 0;
   while (pos < 8 && pos < size) {
@@ -508,82 +519,39 @@ NdbOperation::getKeyFromTCREQ(Uint32* data, Uint32 & size)
   NdbApiSignal* tSignal = theTCREQ->next();
   unsigned n = 0;
   while (pos < size) {
-    if (n == 20) {
+    if (n == KeyInfo::DataLength) {
       tSignal = tSignal->next();
       n = 0;
     }
-    data[pos++] = tSignal->getDataPtrSend()[3 + n++];
+    data[pos++] = 
+      tSignal->getDataPtrSend()[KeyInfo::HeaderLength + n++];
   }
   return 0;
 }
 
-int
-NdbOperation::handle_distribution_key(const NdbColumnImpl* tAttrInfo,
-                                      const Uint64* value, Uint32 len)
-{
-  DBUG_ENTER("NdbOperation::handle_distribution_key");
-
-  if (theDistrKeyIndicator_ == 1)
-    DBUG_RETURN(0);
-
-  if (theNoOfTupKeyLeft > 0 || m_accessTable->m_noOfDistributionKeys > 1)
-    DBUG_RETURN(0);
-  
-  DBUG_DUMP("value", (const uchar*)value, len << 2);
-
-  if(m_accessTable->m_noOfDistributionKeys == 1)
-  {
-    Ndb::Key_part_ptr ptrs[2];
-    ptrs[0].ptr = value;
-    ptrs[0].len = len;
-    ptrs[1].ptr = 0;
-    
-    Uint64 tmp[1000];
-    Uint32 hashValue;
-    int ret = Ndb::computeHash(&hashValue, 
-                               m_currentTable,
-                               ptrs, tmp, sizeof(tmp));
-    
-    if (ret == 0)
-    {
-      setPartitionId(m_currentTable->getPartitionId(hashValue));
-    }
-#ifdef VM_TRACE
-    else
-    {
-      ndbout << "Err: " << ret << endl;
-      assert(false);
-    }
-#endif
-  }
-  DBUG_RETURN(0);
-}
-
 void
 NdbOperation::setPartitionId(Uint32 value)
 {
+  if (theStatus == UseNdbRecord)
+  {
+    /* Method not allowed for NdbRecord, use OperationOptions or 
+       ScanOptions structure instead */
+    setErrorCodeAbort(4515);
+    return; // TODO : Consider adding int rc for error
+  }
+
+  /* Todo - 6.4
+   * Should only be allowed for user-defined partitioned tables
+   * Add an #ifdef STRICT_SET_PARTITION_ID
+   * Cannot be on at least until Blobs stop setting partition
+   * id for parts and MySQLD stops setting it for Alter table.
+   * operations.
+   */
+
   theDistributionKey = value;
   theDistrKeyIndicator_ = 1;
   DBUG_PRINT("info", ("NdbOperation::setPartitionId: %u",
                        theDistributionKey));
-  /*
-    NdbBlob needs the distribution key to set it correctly on the injected
-    operations.
-    Unfortunately, NdbBlob currently requires that distribution key be set
-    _before_ calling getBlobHandle() (and not changed afterwards). This is
-    impossible for NdbRecord operations, which do getBlobHandle() before the
-    operation is even returned to the caller.
-
-    This hack allows things to work in a few more cases, however it is still
-    necessary to set the distribution key before doing any blob operations
-    (and not change it after).
-  */
-  NdbBlob *blob = theBlobList;
-  while (blob != NULL)
-  {
-    blob->thePartitionId = value;
-    blob= blob->next();
-  }
 }
 
 Uint32
