@@ -24,6 +24,7 @@
 #include "ScanFunctions.hpp"
 #include <random.h>
 #include <signaldata/DumpStateOrd.hpp>
+#include <NdbConfig.hpp>
 
 const NdbDictionary::Table *
 getTable(Ndb* pNdb, int i){
@@ -440,6 +441,47 @@ int runScanReadError(NDBT_Context* ctx, NDBT_Step* step){
   return result;
 }
 
+int runScanReadExhaust(NDBT_Context* ctx, NDBT_Step* step)
+{
+  int result = NDBT_OK;
+  int loops = ctx->getNumLoops();
+  int records = ctx->getNumRecords();
+  int parallelism = 240; // Max parallelism
+  int error = 8093;
+  NdbRestarter restarter;
+  
+  /* First take a TC resource snapshot */
+  int savesnapshot= DumpStateOrd::TcResourceSnapshot;
+  int checksnapshot= DumpStateOrd::TcResourceCheckLeak;
+  
+  restarter.dumpStateAllNodes(&savesnapshot, 1);
+  int i = 0;
+  HugoTransactions hugoTrans(*ctx->getTab());
+  hugoTrans.setRetryMax(1);
+  while (i<loops && !ctx->isTestStopped()) {
+    g_info << i << ": ";
+    
+    ndbout << "insertErrorInAllNodes("<<error<<")"<<endl;
+    if (restarter.insertErrorInAllNodes(error) != 0){
+      ndbout << "Could not insert error in all nodes "<<endl;
+      return NDBT_FAILED;
+    }
+    
+    if (hugoTrans.scanReadRecords(GETNDB(step), records, 0, parallelism) == 0)
+    {
+      /* Expect error 291 */
+      result = NDBT_FAILED;
+      break;
+    }
+    i++;
+  }
+  
+  restarter.insertErrorInAllNodes(0);
+
+  restarter.dumpStateAllNodes(&checksnapshot, 1);
+  return result;
+}
+
 int
 runInsertError(NDBT_Context* ctx, NDBT_Step* step){
   int error = ctx->getProperty("ErrorCode");
@@ -1334,6 +1376,220 @@ finalizeBug42559(NDBT_Context* ctx, NDBT_Step* step){
   return NDBT_OK;
 }
 
+int takeResourceSnapshot(NDBT_Context* ctx, NDBT_Step* step)
+{
+  NdbRestarter restarter;
+  
+  int checksnapshot = DumpStateOrd::TcResourceSnapshot;
+  restarter.dumpStateAllNodes(&checksnapshot, 1);
+
+  /* TODO : Check other block's resources? */
+  return NDBT_OK;
+}
+
+int runScanReadIndexWithBounds(NDBT_Context* ctx, NDBT_Step* step){
+  int loops = ctx->getNumLoops();
+  int records = ctx->getNumRecords();
+  int numRanges = ctx->getProperty("NumRanges", 1);
+  int maxRunSecs = ctx->getProperty("MaxRunSecs", 60);
+  int maxRetries = ctx->getProperty("MaxRetries", 1000000);
+  
+  const NdbDictionary::Index * pIdx = 
+    GETNDB(step)->getDictionary()->getIndex(orderedPkIdxName, 
+					    ctx->getTab()->getName());
+
+  int i = 0;
+  HugoCalculator calc(*ctx->getTab());
+  NDBT_ResultRow row (*ctx->getTab());
+  Ndb* ndb = GETNDB(step);
+
+  Uint64 start = NdbTick_CurrentMillisecond();
+  Uint64 end = start + (1000*maxRunSecs);
+  int retries = 0;
+
+  /* Here we run an ordered index scan, with a bound.
+   * There are numRanges sub-scans with the same bound
+   * This is done to use up some KeyInfo, and expose bugs in that area
+   * If we run many of these in parallel we may exhaust the available KeyInfo storage,
+   * which may expose some bugs.
+   */
+  while (pIdx &&
+         i<loops && 
+         !ctx->isTestStopped() &&
+         NdbTick_CurrentMillisecond() < end) {
+    g_info << "Step " << step->getStepNo() 
+           << "Loop : " << i << ": ";
+    
+    /* Use specific-partition variant of startTransaction to ensure a single
+     * TC node is used
+     */
+    NdbTransaction* trans = ndb->startTransaction(ctx->getTab(), Uint32(0));
+    if (trans == NULL)
+    {
+      g_err << "Transaction start failed " << ndb->getNdbError() << endl;
+      return NDBT_FAILED;
+    }
+
+    NdbIndexScanOperation* iso = trans->getNdbIndexScanOperation(pIdx->getName(), 
+                                                                 ctx->getTab()->getName());
+    if (iso == NULL)
+    {
+      g_err << "Error obtaining IndexScanOperation : " << trans->getNdbError() << endl;
+      trans->close();
+      return NDBT_FAILED;
+    }
+
+    if (iso->readTuples(NdbOperation::LM_CommittedRead, 
+                        (NdbScanOperation::SF_OrderBy |
+                         NdbScanOperation::SF_ReadRangeNo |
+                         NdbScanOperation::SF_MultiRange), 
+                        0) != 0)
+    {
+      g_err << "Error calling readTuples : " << iso->getNdbError() << endl;
+      trans->close();
+      return NDBT_FAILED;
+    }
+
+    for (int range = 0; range < numRanges; range++)
+    {
+      /* Now define a bound... */
+      for (Uint32 k=0; k<pIdx->getNoOfColumns(); k++)
+      {
+        const NdbDictionary::Column* idxCol = pIdx->getColumn(k);
+        const char* colName = idxCol->getName();
+        /* Lower bound of <= NULL should return all rows */
+        if (iso->setBound(colName, NdbIndexScanOperation::BoundLE, NULL) != 0)
+        {
+          g_err << "Error setting bound for column %s. " 
+                << iso->getNdbError() << endl;
+          trans->close();
+          return NDBT_FAILED;
+        }
+      }
+      
+      if (iso->end_of_bound(range) != 0)
+      {
+        g_err << "Error closing range " << range << endl;
+        g_err << iso->getNdbError() << endl;
+        return NDBT_FAILED;
+      }
+    }
+    
+    const NdbDictionary::Table& tab= *ctx->getTab();
+
+    /* Now request all columns in result projection */
+    for (int a=0; a<tab.getNoOfColumns(); a++){
+      if((row.attributeStore(a) = 
+	  iso->getValue(tab.getColumn(a)->getName())) == 0) {
+	g_err << "Error defining read value " << trans->getNdbError() << endl;
+        trans->close();
+	return NDBT_FAILED;
+      }
+    }
+          
+    /* Ready to go... */
+    trans->execute(NoCommit, AbortOnError);
+
+    if (trans->getNdbError().code != 0)
+    {
+      if (trans->getNdbError().code == 218)
+      {
+        /* Out of KeyInfo buffers in TC - that's ok, let's try again */
+        trans->close();
+        if (retries++ < maxRetries)
+        {
+          g_err << "Step " << step->getStepNo()
+                << " TC out of Keyinfo buffers (218) - retrying" << endl;
+          continue;
+        }
+      }
+
+      g_err << "Error on execution : " << trans->getNdbError() << endl;
+      trans->close();
+      return NDBT_FAILED;
+    }
+    
+    int eof;
+    int rows = 0;
+    while ((eof = iso->nextResult(true)) == 0)
+    {
+      rows++;
+      if (calc.verifyRowValues(&row) != 0)
+      {
+        g_err << "Verification failed." << endl;
+        trans->close();
+        return NDBT_FAILED;
+      }
+
+#ifdef BUG_14388257_FIXED
+      int rangeNum = (rows -1) / records;
+      if (iso->get_range_no() != rangeNum)
+      {
+        g_err << "Expected row " << rows 
+              << " to be in range " << rangeNum
+              << " but it reports range num " << iso->get_range_no()
+              << " : " << row
+              << endl;
+        return NDBT_FAILED;
+      }
+#endif
+
+      //g_err << row << endl;
+    }
+
+    if (eof != 1)
+    {
+      g_err << "nextResult() returned " << eof << endl;
+      g_err << "Scan error : " << iso->getNdbError() << endl;
+
+      if (iso->getNdbError().status == NdbError::TemporaryError)
+      {
+        if (retries++ < maxRetries)
+        {
+          g_err << "Step "
+                << step->getStepNo()
+                << "  Temporary, retrying on iteration " 
+                << i << " rows so far : " << rows << endl;
+          trans->close();
+          NdbSleep_MilliSleep(2500);
+          continue;
+        }
+      }
+
+      trans->close();
+      return NDBT_FAILED;
+    }
+    
+    g_err << "Read " << rows << " rows." << endl;
+    
+    if (records != 0 && rows != (numRanges * records))
+    {
+      g_err << "Expected " << records << " rows"
+            << ", read " << rows << endl;
+#ifdef BUG_14388257_FIXED
+      trans->close();
+      assert(false);
+      return NDBT_FAILED;
+#endif
+    }
+
+    trans->close();
+    i++;
+  }
+  return NDBT_OK;
+}
+
+int checkResourceSnapshot(NDBT_Context* ctx, NDBT_Step* step)
+{
+  NdbRestarter restarter;
+  
+  int checksnapshot = DumpStateOrd::TcResourceCheckLeak;
+  restarter.dumpStateAllNodes(&checksnapshot, 1);
+
+  /* TODO : Check other block's resources? */
+  return NDBT_OK;
+}
+
 
 int
 runBug54945(NDBT_Context* ctx, NDBT_Step* step)
@@ -1490,6 +1746,112 @@ runCloseRefresh(NDBT_Context* ctx, NDBT_Step* step)
   return NDBT_OK;
 }
 
+// An 'assert' that is always executed, so that 'cond' may have side effects.
+#define require(cond) \
+  do \
+  { \
+    if (!(cond)) \
+    { \
+      g_err << "require(" << #cond << ") failed at " << __FILE__ << ":" \
+        << __LINE__ << endl; \
+      abort(); \
+    } \
+  }while(false)
+
+/**
+ * This is a regression test for bug #11748194 "TRANSACTION OBJECT CREATED 
+ * AND UNRELEASED BY EXTRA CALL TO NEXTRESULT()".
+ * If a transaction made an extra call to nextResult() after getting
+ * end-of-scan from nextResult(), the API would leak transaction objects. 
+ */
+static int
+runExtraNextResult(NDBT_Context* ctx, NDBT_Step* step)
+{
+  const NdbDictionary::Table * pTab = ctx->getTab();
+  // Fill table with 10 rows.
+  HugoTransactions hugoTrans(*pTab);
+  Ndb* const ndb = GETNDB(step);
+  hugoTrans.loadTable(ndb, 10);
+  // Read MaxNoOfConcurrentTransactions configuration value.
+  Uint32 maxTrans = 0;
+  NdbConfig conf(ndb->getNodeId());
+  require(conf.getProperty(conf.getMasterNodeId(),
+                           NODE_TYPE_DB,
+                           CFG_DB_NO_TRANSACTIONS,
+                           &maxTrans));
+  require(maxTrans > 0);
+  
+  /**
+   * The bug causes each scan to leak one object.
+   */
+  int result = NDBT_OK;
+  Uint32 i = 0;
+  while (i < maxTrans+1)
+  {
+    NdbTransaction* const trans = ndb->startTransaction();
+    if (trans == NULL)
+    {
+      g_err << "ndb->startTransaction() gave unexpected error : " 
+            << ndb->getNdbError() << " in the " << i << "th iteration." <<endl;
+      return NDBT_FAILED;
+    }
+    
+    // Do a random numer of scans in this transaction.
+    const int scanCount = rand()%4;
+    for (int j=0; j < scanCount; j++)
+    {
+      NdbScanOperation* const scan = trans->getNdbScanOperation(pTab);
+      if (scan == NULL)
+      {
+        g_err << "trans->getNdbScanOperation() gave unexpected error : " 
+              << trans->getNdbError() << " in the " << i
+              << "th iteration." <<endl;
+        return NDBT_FAILED;
+      }
+      
+      require(scan->readTuples(NdbOperation::LM_CommittedRead) == 0);
+      require(scan->getValue(0u) != 0);
+      require(trans->execute(NoCommit) == 0);
+      
+      // Scan table until end.
+      int scanResult;
+      do
+      {
+        // Fetch new batch.
+        scanResult = scan->nextResult(true);
+        while (scanResult == 0)
+        {
+          // Iterate over batch.
+          scanResult = scan->nextResult(false);
+        }
+      } while (scanResult == 0 || scanResult == 2);
+
+      /** 
+       * Do extra nextResult. This is the application error that triggers the 
+       * bug.
+       */
+      scanResult = scan->nextResult(true);
+      require(scanResult < 0);
+      // Here we got the undefined error code -1. So check for that too.
+      // if (trans->getNdbError().code != 4120
+      if (scan->getNdbError().code != 4120
+          && result == NDBT_OK)
+      {
+        g_err << "scan->nextResult() gave unexpected error : " 
+              << scan->getNdbError() << " in the " << i << "th iteration." 
+              << endl;
+        result = NDBT_FAILED;
+      }
+      i++;
+    }
+    ndb->closeTransaction(trans);
+  } // while (i < maxTrans+1)
+
+  // Delete table rows.
+  require(UtilTransactions(*ctx->getTab()).clearTable(ndb) == 0);
+  return result;
+}
+
 NDBT_TESTSUITE(testScan);
 TESTCASE("ScanRead", 
 	 "Verify scan requirement: It should be possible "\
@@ -2024,6 +2386,59 @@ TESTCASE("Bug54945", "")
 {
   INITIALIZER(runBug54945);
 }
+TESTCASE("ScanFragRecExhaust", 
+         "Test behaviour when TC scan frag recs exhausted")
+{
+  INITIALIZER(runLoadTable);
+  INITIALIZER(runScanReadExhaust);
+  FINALIZER(runClearTable);
+}
+TESTCASE("extraNextResultBug11748194",
+         "Regression test for bug #11748194")
+{
+  INITIALIZER(runExtraNextResult);
+}
+TESTCASE("ScanRealKeyInfoExhaust",
+         "Test behaviour when TC keyinfo buffers exhausted 4real")
+{
+  /* 55 threads, each setting 200 ranges in their keyinfo
+   * For the lightest single column PK case, each range should
+   * use 2 words, 200 ranges = 400 words per scan thread = 
+   * 400/4 = 100 Databuffers used.
+   * 55 threads = 55*100 = 5500 Databuffers which is >
+   * the 4000 statically allocated in 6.3
+   */
+  TC_PROPERTY("NumRanges", 200);
+  TC_PROPERTY("MaxRunSecs", 120);
+  INITIALIZER(createOrderedPkIndex);
+  INITIALIZER(runLoadTable);
+  INITIALIZER(takeResourceSnapshot);
+  STEPS(runScanReadIndexWithBounds,55);
+  FINALIZER(checkResourceSnapshot);
+  FINALIZER(createOrderedPkIndex_Drop);
+  FINALIZER(runClearTable);
+}
+TESTCASE("ScanKeyInfoExhaust",
+         "Test behaviour when TC keyinfo buffers exhausted with error insert")
+{
+  /* Use error insert 8094 to cause keyinfo exhaustion, then run a single scan
+   * with keyinfo to hit the error path
+   */
+  TC_PROPERTY("MaxRunSecs", 10);
+  INITIALIZER(createOrderedPkIndex);
+  INITIALIZER(runLoadTable);
+  INITIALIZER(takeResourceSnapshot);
+  TC_PROPERTY("ErrorCode", 8094);
+  INITIALIZER(runInsertError);
+  STEP(runScanReadIndexWithBounds);
+  FINALIZER(checkResourceSnapshot);
+  FINALIZER(runInsertError);
+  FINALIZER(createOrderedPkIndex_Drop);
+  FINALIZER(runClearTable);
+}
+  
+
+
 NDBT_TESTSUITE_END(testScan);
 
 int main(int argc, const char** argv){
